Require Export scase.
Require Import Arith.
Require Import Coq.Arith.Compare.
Require Import Coq.Lists.List.
Require Import Coq.Logic.FunctionalExtensionality.
Require Import Coq.Relations.Operators_Properties.
Require Import Coq.Relations.Relation_Definitions.
Require Import Coq.Relations.Relation_Operators.
Require Import Nat.
Require Import Omega.

Import ListNotations.

Section Coterm.

  (****************************************************************************
   ******************************** Definition ********************************
   ****************************************************************************)

  Variable Label : Set.

  Inductive clos_refl_trans_1n_length (A : Type) (R : relation A) (x : A) :
    A -> nat -> Prop :=
  | rt1n_length_refl : clos_refl_trans_1n_length A R x x 0
  | rt1n_length_trans : forall (y z : A) (n : nat),
      R x y ->
      clos_refl_trans_1n_length A R y z n ->
      clos_refl_trans_1n_length A R x z (S n).

  Lemma clos_rt1n_length :
    forall A R x y, clos_refl_trans_1n A R x y ->
                    exists n, clos_refl_trans_1n_length A R x y n.
  Proof.
    intros.
    induction H.
    exists 0.
    apply rt1n_length_refl.

    inversion IHclos_refl_trans_1n.
    exists (S x0).
    apply rt1n_length_trans with (y := y); auto.
  Qed.

  Lemma clos_length_rt1n :
    forall A R x y n, clos_refl_trans_1n_length A R x y n ->
                      clos_refl_trans_1n A R x y.
  Proof.
    intros.
    induction H.
    apply rt1n_refl.
    apply rt1n_trans with (y := y); auto.
  Qed.

  Set Implicit Arguments.

  Fixpoint repeat {A} (f : A -> A) (x : A) (n : nat) : A :=
    match n with
      | O => x
      | S n' => repeat f (f x) n'
    end.

  Definition var := nat.
  Definition tyvar := nat.

  Inductive const : Type :=
  | c_bool : bool -> const
  | c_nat  : nat  -> const.

  Inductive bty : Type :=
  | bty_bool : bty
  | bty_nat  : bty.

  Inductive term : Type :=
  | term_var    : var -> term
  | term_val    : val -> term
  | term_app    : term -> term -> term
  | term_tapp   : term -> ty -> term
  | term_wcheck : ty -> term -> term -> Label -> term
  | term_acheck : ty -> term -> term -> val -> Label -> term
  | term_blame  : Label -> term
  with val : Type :=
  | val_const   : const -> val
  | val_fix     : ty -> ty -> term -> val
  | val_tabs    : term -> val
  | val_cast    : ty -> ty -> Label -> val
  with ty : Type :=
  | ty_base     : bty -> ty
  | ty_var      : tyvar -> ty
  | ty_fun      : ty -> ty -> ty
  | ty_forall   : ty -> ty
  | ty_refine   : ty -> term -> ty.

  Notation "! v" := (term_val v) (at level 0).
  Infix "@" := term_app (at level 55, left associativity).
  Infix "@@" := term_tapp (at level 55, left associativity).

  Variable ty_of_const : const -> ty.

  Definition Bool : ty := ty_base bty_bool.
  Definition tt : val  := val_const (c_bool true).
  Definition ff : val  := val_const (c_bool false).

  Scheme term_mind := Induction for term Sort Type
  with val_mind := Induction for val Sort Type
  with ty_mind := Induction for ty Sort Type.

  Lemma term_val_ty_mind :
    forall (Pterm : term -> Prop) (Pval : val -> Prop) (Pty : ty -> Prop),
      (forall v : var, Pterm (term_var v)) ->
      (forall v : val, Pval v -> Pterm ! (v)) ->
      (forall t : term, Pterm t ->
        forall t0 : term, Pterm t0 -> Pterm (t @ t0)) ->
      (forall t : term, Pterm t -> forall t0 : ty, Pty t0 -> Pterm (t @@ t0)) ->
      (forall t : ty, Pty t ->
        forall t0 : term, Pterm t0 ->
          forall t1 : term, Pterm t1 ->
            forall l : Label, Pterm (term_wcheck t t0 t1 l)) ->
      (forall t : ty, Pty t ->
        forall t0 : term, Pterm t0 ->
          forall t1 : term, Pterm t1 ->
            forall v : val, Pval v ->
              forall l : Label, Pterm (term_acheck t t0 t1 v l)) ->
      (forall l : Label, Pterm (term_blame l)) ->
      (forall c : const, Pval (val_const c)) ->
      (forall t : ty, Pty t ->
        forall t0 : ty, Pty t0 ->
          forall t1 : term, Pterm t1 -> Pval (val_fix t t0 t1)) ->
      (forall t : term, Pterm t -> Pval (val_tabs t)) ->
      (forall t : ty, Pty t ->
        forall t0 : ty, Pty t0 ->
          forall l : Label, Pval (val_cast t t0 l)) ->
      (forall b : bty, Pty (ty_base b)) ->
      (forall t : tyvar, Pty (ty_var t)) ->
      (forall t : ty, Pty t ->
        forall t0 : ty, Pty t0 -> Pty (ty_fun t t0)) ->
      (forall t : ty, Pty t -> Pty (ty_forall t)) ->
      (forall t : ty, Pty t ->
        forall t0 : term, Pterm t0 -> Pty (ty_refine t t0)) ->
      (forall t : term, Pterm t) /\
      (forall v : val, Pval v) /\
      (forall T : ty, Pty T).
  Proof.
    intros.
    split.
    apply (term_mind Pterm Pval Pty); assumption.
    split.
    apply (val_mind Pterm Pval Pty); assumption.
    apply (ty_mind Pterm Pval Pty); assumption.
  Qed.

  (* Metrics *)
  Fixpoint size_term (e : term) :=
    match e with
    | term_var _ | term_blame _ => 1
    | term_val v => size_val v
    | term_app e1 e2 => 1 + size_term e1 + size_term e2
    | term_tapp e1 T2 => 1 + size_term e1 + size_ty T2
    | term_wcheck T e1 e2 _ => 1 + size_ty T + size_term e1 + size_term e2
    | term_acheck T e1 e2 v _ =>
      1 + size_ty T + size_term e1 + size_term e2 + size_val v
    end
  with size_val (v : val) :=
     match v with
     | val_const _ => 1
     | val_fix T1 T2 e => 1 + size_ty T1 + size_ty T2 + size_term e
     | val_tabs e => 1 + size_term e
     | val_cast T1 T2 _ => 1 + size_ty T1 + size_ty T2
     end
  with size_ty (T : ty) :=
     match T with
     | ty_base _ | ty_var _ => 1
     | ty_fun T1 T2 => 1 + size_ty T1 + size_ty T2
     | ty_forall T' => 1 + size_ty T'
     | ty_refine T' e' => 1 + size_ty T' + size_term e'
     end.

  Definition size_terms (es : list term) :=
    fold_left (fun x y => x + y) (map (fun e => size_term e) es) 0.

  Definition size_vals (vs : list val) :=
    fold_left (fun x y => x + y) (map (fun v => size_val v) vs) 0.

  (* Substitution *)

  Fixpoint shift (e : term)  (n : nat) : term :=
    match e with
      | term_var x => term_var (if le_gt_dec n x then 1 + x else x)
      | term_val v => term_val (shift_val v n)
      | term_app e1 e2 => term_app (shift e1 n) (shift e2 n)
      | term_tapp e1 T2 => term_tapp (shift e1 n) (shift_ty T2 n)
      | term_wcheck T1 e1 e2 l => term_wcheck (shift_ty T1 n)
                                              (shift e1 (1 + n))
                                              (shift e2 n)
                                              l
      | term_acheck T1 e1 e2 v l => term_acheck (shift_ty T1 n)
                                                (shift e1 (1 + n))
                                                (shift e2 n)
                                                (shift_val v n)
                                                l
      | term_blame _ => e
    end
  with shift_val (v : val) (n : nat) : val :=
    match v with
      | val_const _ => v
      | val_fix T1 T2 e => val_fix (shift_ty T1 n)
                                   (shift_ty T2 (1 + n))
                                   (shift e (2 + n))
      | val_tabs e => val_tabs (shift e n)
      | val_cast T1 T2 l => val_cast (shift_ty T1 n) (shift_ty T2 n) l
    end
  with shift_ty (T : ty) (n : nat) : ty :=
    match T with
      | ty_base _ | ty_var _ => T
      | ty_fun T1 T2 => ty_fun (shift_ty T1 n) (shift_ty T2 (1 + n))
      | ty_forall T' => ty_forall (shift_ty T' n)
      | ty_refine T' e' => ty_refine (shift_ty T' n) (shift e' (1 + n))
    end.

  Definition shift_n (e : term) (n : nat) : term :=
    repeat (fun e => shift e 0) e n.

  Definition shift_ty_n (T : ty) (n : nat) : ty :=
    repeat (fun T => shift_ty T 0) T n.

  Definition shift_subst (s : nat * list term) :=
    (S (fst s), (map (fun e => shift e 0) (snd s))).

  Definition shift_tysubst (s : nat * list ty) :=
    (fst s, (map (fun T => shift_ty T 0) (snd s))).

  Fixpoint tyshift (e : term)  (n : nat) : term :=
    match e with
    | term_var _ | term_blame _ => e
    | term_val v => term_val (tyshift_val v n)
    | term_app e1 e2 => term_app (tyshift e1 n) (tyshift e2 n)
    | term_tapp e1 T2 => term_tapp (tyshift e1 n) (tyshift_ty T2 n)
    | term_wcheck T1 e1 e2 l => term_wcheck (tyshift_ty T1 n)
                                            (tyshift e1 n)
                                            (tyshift e2 n)
                                            l
    | term_acheck T1 e1 e2 v l => term_acheck (tyshift_ty T1 n)
                                              (tyshift e1 n)
                                              (tyshift e2 n)
                                              (tyshift_val v n)
                                              l
    end
  with tyshift_val (v : val) (n : nat) : val :=
    match v with
    | val_const _ => v
    | val_fix T1 T2 e => val_fix (tyshift_ty T1 n)
                                 (tyshift_ty T2 n)
                                 (tyshift e n)
    | val_tabs e => val_tabs (tyshift e (1 + n))
    | val_cast T1 T2 l => val_cast (tyshift_ty T1 n) (tyshift_ty T2 n) l
    end
  with tyshift_ty (T : ty) (n : nat) : ty :=
    match T with
    | ty_base _ => T
    | ty_var a => ty_var (if le_gt_dec n a then 1 + a else a)
    | ty_fun T1 T2 => ty_fun (tyshift_ty T1 n) (tyshift_ty T2 n)
    | ty_forall T' => ty_forall (tyshift_ty T' (1 + n))
    | ty_refine T' e' => ty_refine (tyshift_ty T' n) (tyshift e' n)
    end.

  Definition tyshift_n (e : term) (n : nat) : term :=
    repeat (fun e => tyshift e 0) e n.

  Definition tyshift_tysubst (s : nat * list ty) :=
    (S (fst s), map (fun T => tyshift_ty T 0) (snd s)).

  Definition tyshift_subst (s : nat * list term) :=
    (fst s, map (fun e => tyshift e 0) (snd s)).

  Fixpoint subst (e : term) (s : nat * list term) : term :=
    match e with
    | term_var x => let n := fst s in
                    let l := snd s in
                    if lt_dec x n then term_var x
                    else nth (x - n) l (term_var (x - (length l)))
    | term_val v => term_val (subst_val v s)
    | term_app e1 e2 => term_app (subst e1 s) (subst e2 s)
    | term_tapp e1 T2 => term_tapp (subst e1 s) (subst_ty T2 s)
    | term_wcheck T1 e1 e2 l => term_wcheck (subst_ty T1 s)
                                            (subst e1 (shift_subst s))
                                            (subst e2 s)
                                            l
    | term_acheck T1 e1 e2 v l => term_acheck (subst_ty T1 s)
                                              (subst e1 (shift_subst s))
                                              (subst e2 s)
                                              (subst_val v s)
                                              l
    | term_blame _ => e
    end
  with subst_val (v : val) (s : nat * list term) : val :=
    match v with
    | val_const _ => v
    | val_fix T1 T2 e => val_fix (subst_ty T1 s)
                                 (subst_ty T2 (shift_subst s))
                                 (subst e (shift_subst (shift_subst s)))
    | val_tabs e => val_tabs (subst e (tyshift_subst s))
    | val_cast T1 T2 l => val_cast (subst_ty T1 s)
                                   (subst_ty T2 s)
                                   l
    end
  with subst_ty (T : ty) (s : nat * list term) : ty :=
    match T with
    | ty_base _ | ty_var _ => T
    | ty_fun T1 T2 => ty_fun (subst_ty T1 s) (subst_ty T2 (shift_subst s))
    | ty_forall T' => ty_forall (subst_ty T' (tyshift_subst s))
    | ty_refine T' e' => ty_refine (subst_ty T' s) (subst e' (shift_subst s))
  end.


  Fixpoint tysubst (e : term) (s : nat * list ty) : term :=
    match e with
    | term_var _ | term_blame _ => e
    | term_val v => term_val (tysubst_val v s)
    | term_app e1 e2 => term_app (tysubst e1 s) (tysubst e2 s)
    | term_tapp e1 T2 => term_tapp (tysubst e1 s) (tysubst_ty T2 s)
    | term_wcheck T1 e1 e2 l => term_wcheck (tysubst_ty T1 s)
                                            (tysubst e1 (shift_tysubst s))
                                            (tysubst e2 s)
                                            l
    | term_acheck T1 e1 e2 v l => term_acheck (tysubst_ty T1 s)
                                              (tysubst e1 (shift_tysubst s))
                                              (tysubst e2 s)
                                              (tysubst_val v s)
                                              l
    end
  with tysubst_val (v : val) (s : nat * list ty) : val :=
    match v with
    | val_const _ => v
    | val_fix T1 T2 e => val_fix (tysubst_ty T1 s)
                                 (tysubst_ty T2 (shift_tysubst s))
                                 (tysubst e (shift_tysubst (shift_tysubst s)))
    | val_tabs e => val_tabs (tysubst e (tyshift_tysubst s))
    | val_cast T1 T2 l => val_cast (tysubst_ty T1 s)
                                   (tysubst_ty T2 s)
                                   l
    end
  with tysubst_ty (T : ty) (s : nat * list ty) : ty :=
    match T with
    | ty_base _ => T
    | ty_var a => let n := fst s in
                  let l := snd s in
                  if lt_dec a n then ty_var a
                  else nth (a - n) l (ty_var (a - length l))
    | ty_fun T1 T2 => ty_fun (tysubst_ty T1 s) (tysubst_ty T2 (shift_tysubst s))
    | ty_forall T' => ty_forall (tysubst_ty T' (tyshift_tysubst s))
    | ty_refine T' e' => ty_refine (tysubst_ty T' s)
                                   (tysubst e' (shift_tysubst s))
    end.

  Definition shift_substO l := shift_subst (O, l).
  Definition tyshift_tysubstO l := tyshift_tysubst (O, l).

  Definition substO x l := subst x (O, l).
  Definition subst_valO x l := subst_val x (O, l).
  Definition subst_tyO x l := subst_ty x (O, l).
  Definition tysubstO x l := tysubst x (O, l).
  Definition tysubst_valO x l := tysubst_val x (O, l).
  Definition tysubst_tyO x l := tysubst_ty x (O, l).

  (* Closed *)
  Definition closed_term e :=
    forall n e', subst e (n, [e']) = e.

  Section subst_test.

    Example subst_fix :
      subst_val
        (val_fix
           (ty_refine Bool (term_app (term_var 0) (term_var 1)))
           (ty_refine Bool (term_app (term_var 0) (term_app (term_var 1) (term_var 2))))
           (term_app (term_var 0) (term_app (term_var 1) (term_var 2))))
        (0, [term_app (term_var 0) (term_var 1)])
      =
      (val_fix
         (ty_refine Bool (term_app (term_var 0) (term_app (term_var 1) (term_var 2))))
         (ty_refine Bool (term_app (term_var 0) (term_app (term_var 1) (term_app (term_var 2) (term_var 3)))))
         (term_app (term_var 0) (term_app (term_var 1) (term_app (term_var 2) (term_var 3))))).
    Proof.
      simpl. reflexivity.
    Qed.

    Example tysubst_fix :
      tysubst_valO
        (val_fix
           (ty_refine Bool (term_app (term_var 0) (term_var 1)))
           (ty_refine Bool (term_app (term_var 0) (term_app (term_var 1) (term_var 2))))
           (term_app (term_var 0) (term_app (term_var 1) (term_var 2))))
        [ty_var 2]
      =
      (val_fix
           (ty_refine Bool (term_app (term_var 0) (term_var 1)))
           (ty_refine Bool (term_app (term_var 0) (term_app (term_var 1) (term_var 2))))
           (term_app (term_var 0) (term_app (term_var 1) (term_var 2)))).
    Proof.
      simpl. reflexivity.
    Qed.

    Example subst_empty :
      subst_tyO (ty_refine (ty_base bty_bool) (term_var 1)) [] =
      ty_refine (ty_base bty_bool) (term_var 1).
    Proof.
      reflexivity.
    Qed.
  End subst_test.


  Inductive ectx : Type :=
  | ectx_hole   : ectx
  | ectx_app1   : ectx -> term -> ectx
  | ectx_app2   : val -> ectx -> ectx
  | ectx_tapp   : ectx -> ty -> ectx
  | ectx_wcheck : ty -> term -> ectx -> Label -> ectx
  | ectx_acheck : ty -> term -> ectx -> val -> Label -> ectx.

  Fixpoint size_ectx (E : ectx) :=
    match E with
    | ectx_hole => 0
    | ectx_app1 E1 e2 => 1 + size_ectx E1 + size_term e2
    | ectx_app2 v1 E2 => 1 + size_val v1 + size_ectx E2
    | ectx_tapp E' T => 1 + size_ectx E' + size_ty T
    | ectx_wcheck T e1 E2 _ => 1 + size_ty T + size_term e1 + size_ectx E2
    | ectx_acheck T e1 E2 v _ =>
      1 + size_ty T + size_term e1 + size_ectx E2 + size_val v
    end.

  Fixpoint ectx_plug (E : ectx) (e : term) : term :=
    match E with
      | ectx_hole                   => e
      | ectx_app1 E' e2'            => term_app (ectx_plug E' e) e2'
      | ectx_app2 v1' E'            => term_app !v1' (ectx_plug E' e)
      | ectx_tapp E' T'             => term_tapp (ectx_plug E' e) T'
      | ectx_wcheck T' e1' E' l'    => term_wcheck T' e1' (ectx_plug E' e) l'
      | ectx_acheck T' e1' E' v' l' => term_acheck T' e1' (ectx_plug E' e) v' l'
    end.

  Notation "E @[ e ]" := (ectx_plug E e) (at level 55).

  Fixpoint ectx_compose (E1 E2 : ectx) : ectx :=
    match E1 with
    | ectx_hole => E2
    | ectx_app1 E1' e2' => ectx_app1 (ectx_compose E1' E2) e2'
    | ectx_app2 v1' E2' => ectx_app2 v1' (ectx_compose E2' E2)
    | ectx_tapp E' T' => ectx_tapp (ectx_compose E' E2) T'
    | ectx_wcheck T' e' E' l => ectx_wcheck T' e' (ectx_compose E' E2) l
    | ectx_acheck T' e' E' v' l => ectx_acheck T' e' (ectx_compose E' E2) v' l
    end.

  Notation "E1 ^^ E2" := (ectx_compose E1 E2) (at level 55).

  Fixpoint subst_ectx (E : ectx) (s : nat * list term) :=
    match E with
    | ectx_hole => ectx_hole
    | ectx_app1 E1' e2' => ectx_app1 (subst_ectx E1' s) (subst e2' s)
    | ectx_app2 v1' E2' => ectx_app2 (subst_val v1' s) (subst_ectx E2' s)
    | ectx_tapp E' T' => ectx_tapp (subst_ectx E' s) (subst_ty T' s)
    | ectx_wcheck T' e' E' l => ectx_wcheck (subst_ty T' s)
                                            (subst e' (shift_subst s))
                                            (subst_ectx E' s)
                                            l
    | ectx_acheck T' e' E' v' l => ectx_acheck (subst_ty T' s)
                                               (subst e' (shift_subst s))
                                               (subst_ectx E' s)
                                               (subst_val v' s)
                                               l
    end.

  Definition subst_ectxO x l := subst_ectx x (O, l).

  Definition closed_ectx E :=
    forall n e', subst_ectx E (n, [e']) = E.

  Inductive red : term -> term -> Prop :=
  | red_beta : forall (T1 T2 : ty) (e1 : term) (v2 : val),
                 let fix_v := !(val_fix T1 T2 e1) in
                 let e := fix_v @ !v2 in
                 red e (substO e1 [!v2; fix_v])
  | red_tbeta : forall (e : term) (T : ty),
                  let e' := !(val_tabs e) @@ T in
                  red e' (tysubstO e [T])
  | red_cast_base : forall (B : bty) (l : Label) (v : val),
                      red (!(val_cast (ty_base B) (ty_base B) l) @ !v) !v
  | red_cast_fun : forall (T11 T21 T12 T22 : ty)
                          (l : Label) (v : val),
                     let T11' := shift_ty_n T11 2 in
                     let T12' := shift_ty (shift_ty T12 1) 0 in
                     let T21' := shift_ty_n T21 2 in
                     let T22' := shift_ty (shift_ty (shift_ty T22 1) 1) 1 in
                     let T12'' := shift_ty (shift_ty_n T12 2) 3 in
                     let v' := shift_n !v 4 in
                     let e :=
                         !(val_cast (ty_fun T11 T12) (ty_fun T21 T22) l) @ !v in
                     let domain_cast := !(val_cast T21' T11' l) in
                     let codomain_cast := !(val_cast T12'' T22' l) in
                     red e
                         !(val_fix
                             T11 T12
                             (!(val_fix
                                  T21' T12'
                                  (codomain_cast @ (v' @ (term_var 0))))
                               @ (domain_cast @ (term_var 0))))
  | red_cast_forall : forall (T1 T2 : ty) (l : Label) (v : val),
                        let cast :=
                            !(val_cast T1 T2 l) in
                        let body := (tyshift !v 0) @@ (ty_var 0) in
                        let e :=
                            !(val_cast (ty_forall T1) (ty_forall T2) l) @ !v in
                        red e !(val_tabs (cast @ body))
  | red_forget : forall (T1 T2 : ty) (e2 : term) (l  : Label) (v  : val),
                   let e := !(val_cast T1 (ty_refine T2 e2) l) @ !v in
                   red e (!(val_cast T1 T2 l) @ !v)
  | red_precheck : forall (T1 : ty) (e1 : term) (T2 : ty) (l  : Label) (v  : val),
                     (forall T e, T2 <> ty_refine T e) ->
                     let e := !(val_cast (ty_refine T1 e1) T2 l) @ !v in
                     red e
                         (term_wcheck T1 e1 (!(val_cast T1 T2 l) @ !v) l)
  | red_check : forall (T : ty) (e : term) (v : val) (l : Label),
                  let e' := term_wcheck T e !v l in
                  red e' (term_acheck T e (subst e (O, [!v])) v l)
  | red_ok : forall (T : ty) (e : term) (v : val) (l : Label),
               let e := term_acheck T e !tt v l in
               red e !v
  | red_fail : forall (T : ty) (e : term) (v : val) (l : Label),
                 let e := term_acheck T e !ff v l in
                 red e (term_blame l).

  Infix "~~>"  := red  (at level 100, right associativity).

  Inductive eval : term -> term -> Prop :=
  | eval_red : forall (e1 e2 : term) (E : ectx),
                 (e1 ~~> e2) ->
                 eval (ectx_plug E e1) (ectx_plug E e2)
  | eval_blame : forall (E : ectx) (l : Label),
                   E <> ectx_hole ->
                   eval (ectx_plug E (term_blame l)) (term_blame l).

  Infix "-->"  := eval (at level 100, right associativity).
  Infix "-->*" := (clos_refl_trans term eval) (at level 100, right associativity).

  Inductive Gbinding : Type :=
  | Gbinding_var : ty -> Gbinding
  | Gbinding_tyvar : Gbinding.

  Definition tctx := list Gbinding.

  Definition shift_G (G : tctx) : tctx :=
    map (fun b => match b with
                    | Gbinding_var T => Gbinding_var (shift_ty T 0)
                    | Gbinding_tyvar => Gbinding_tyvar
        end)
        G.
  Definition tyshift_G (G : tctx) : tctx :=
    map (fun b => match b with
                    | Gbinding_var T => Gbinding_var (tyshift_ty T 0)
                    | Gbinding_tyvar => Gbinding_tyvar
                  end)
        G.

  Definition G_vars (G : tctx) : list ty :=
    fold_right (fun b l => match b with
                             | Gbinding_var T => cons T l
                             | Gbinding_tyvar => l
                           end)
               nil G.

  Definition G_tyvars_length (G : tctx) : nat :=
    (length G) - (length (G_vars G)).

  Inductive compat_ty : ty -> ty -> Prop :=
  | compat_ty_base : forall (B : bty), compat_ty (ty_base B) (ty_base B)
  | compat_ty_tyvar : forall (a : tyvar), compat_ty (ty_var a) (ty_var a)
  | compat_ty_fun : forall (T1 T1' : ty) (T2 T2' : ty),
                        compat_ty T1 T1' ->
                        compat_ty T2 T2' ->
                        compat_ty (ty_fun T1 T2) (ty_fun T1' T2')
  | compat_ty_forall : forall (T T' : ty),
                         compat_ty T T' ->
                         compat_ty (ty_forall T) (ty_forall T')
  | compat_ty_refineL : forall (T T' : ty) (e : term),
                            compat_ty T T' ->
                            compat_ty (ty_refine T e) T'
  | compat_ty_refineR : forall (T T' : ty) (e : term),
                            compat_ty T T' ->
                            compat_ty T (ty_refine T' e).

  Inductive ty_CSR : ty -> ty -> Prop :=
  | ty_CSR_cnstr : forall (T : ty) (e1 e2 : term),
                     (e1 --> e2) ->
                     ty_CSR (subst_tyO T [e1]) (subst_tyO T [e2]).

  Inductive clos_sym_trans (A : Type) (R : A -> A -> Prop) : A -> A -> Prop :=
  | st_step  : forall (x y : A), R x y -> clos_sym_trans R x y
  | st_sym   : forall (x y : A), clos_sym_trans R x y -> clos_sym_trans R y x
  | st_trans : forall (x y z : A), clos_sym_trans R x y ->
                                   clos_sym_trans R y z ->
                                   clos_sym_trans R x z.

  Infix "==" := (clos_sym_trans ty_CSR) (at level 100, right associativity).

  Inductive wellformed_G : tctx -> Prop :=
  | wellformed_empty      : wellformed_G nil
  | wellformed_extend_var : forall (G : tctx) (T : ty),
                              wellformed_ty G T ->
                              wellformed_G G ->
                              wellformed_G (cons (Gbinding_var T) (shift_G G))
  | wellformed_extend_tyvar : forall (G : tctx),
                                wellformed_G G ->
                                wellformed_G (cons Gbinding_tyvar (tyshift_G G))
  with wellformed_ty : tctx -> ty -> Prop :=
  | wellformed_ty_base : forall (G : tctx) (B : bty),
                           wellformed_G G ->
                           wellformed_ty G (ty_base B)
  | wellformed_ty_var : forall (G : tctx) (a : tyvar),
                          wellformed_G G ->
                          a < G_tyvars_length G ->
                          wellformed_ty G (ty_var a)
  | wellformed_ty_fun : forall (G : tctx) (T1 T2 : ty),
                          wellformed_ty G T1 ->
                          wellformed_ty (cons (Gbinding_var T1) (shift_G G)) T2 ->
                          wellformed_ty G (ty_fun T1 T2)
  | wellformed_ty_forall : forall (G : tctx) (T : ty),
                             wellformed_ty (cons Gbinding_tyvar (tyshift_G G)) T ->
                             wellformed_ty G (ty_forall T)
  | wellformed_ty_refine : forall (G : tctx) (T : ty) (e : term),
                             wellformed_ty G T ->
                             typing
                               (cons (Gbinding_var T) (shift_G G))
                               e Bool ->
                             wellformed_ty G (ty_refine T e)
  with typing : tctx -> term -> ty -> Prop :=
  | typing_var : forall (G : tctx) (x : var) (T : ty) l l',
                   wellformed_G G ->
                   G_vars G = l ++ (T :: l') ->
                   x = length l ->
                   typing G (term_var x) T
  | typing_const : forall (G : tctx) (c : const),
                     wellformed_G G ->
                     typing G !(val_const c) (ty_of_const c)
  | typing_fix : forall (G : tctx) (T1 T2 : ty) (e : term),
                   let G'  := Gbinding_var (ty_fun T1 T2) :: (shift_G G) in
                   let G'' := Gbinding_var T1 :: (shift_G G') in
                   typing G'' e (shift_ty_n T2 1) ->
                   typing G !(val_fix T1 T2 e) (ty_fun T1 T2)
  | typing_cast : forall (G : tctx) (T1 T2 : ty) (l : Label),
                    wellformed_ty G T1 ->
                    wellformed_ty G T2 ->
                    compat_ty T1 T2 ->
                    typing G !(val_cast T1 T2 l) (ty_fun T2 (shift_ty_n T1 1))
  | typing_app : forall (G : tctx) (e1 e2 : term) (T1 T2 : ty),
                   typing G e1 (ty_fun T1 T2) ->
                   typing G e2 T1 ->
                   typing G (term_app e1 e2) (subst_tyO T2 [e2])
  | typing_tabs : forall (G : tctx) (e : term) (T : ty),
                    typing (Gbinding_tyvar :: (tyshift_G G)) e T ->
                    typing G !(val_tabs e) (ty_forall T)
  | typing_tapp : forall (G : tctx) (e1 : term) (T2 T : ty),
                    typing G e1 (ty_forall T) ->
                    wellformed_ty G T2 ->
                    typing G (term_tapp e1 T2) (tysubst_tyO T [T2])
  | typing_wcheck : forall (G : tctx) (T : ty) (e1 e2 : term) (l : Label),
                      wellformed_G G ->
                      wellformed_ty nil (ty_refine T e1) ->
                      typing nil e2 T ->
                      typing G (term_wcheck T e1 e2 l) (ty_refine T e1)
  | typing_acheck : forall (G : tctx)
                           (T : ty) (e1 e2 : term) (v : val) (l : Label),
                      wellformed_G G ->
                      wellformed_ty nil (ty_refine T e1) ->
                      typing nil e2 Bool ->
                      typing nil !v T ->
                      (substO e1 [!v] -->* e2) ->
                      typing G (term_acheck T e1 e2 v l) (ty_refine T e1)
  | typing_blame : forall (G : tctx) (T : ty) (l : Label),
                     wellformed_G G ->
                     wellformed_ty nil T ->
                     typing G (term_blame l) T
  | typing_conv : forall (G : tctx) (e : term) (T1 T2 : ty),
                    wellformed_G G ->
                    typing nil e T1 ->
                    (T1 == T2) ->
                    wellformed_ty nil T2 ->
                    typing G e T2
  | typing_forget : forall (G : tctx) (v : val) (T : ty) (e : term),
                      wellformed_G G ->
                      typing nil !v (ty_refine T e) ->
                      typing G !v T
  | typing_exact : forall (G : tctx) (v : val) (T : ty) (e : term),
                     wellformed_G G ->
                     typing nil !v T ->
                     wellformed_ty nil (ty_refine T e) ->
                     (substO e [!v] -->* !tt) ->
                     typing G !v (ty_refine T e).


  (****************************************************************************
   *********************************** Proof **********************************
   ****************************************************************************)

  Unset Implicit Arguments.

  Fixpoint range (n : nat) : list nat :=
    match n with
    | O => nil
    | S n' => cons n (range n')
    end.

  Lemma lemm_subst_var :
    forall n e e',
      substO (term_var n) [e] = e' ->
      e' = e \/ e' = term_var (n-1).
  Proof.
    intros.
    induction n.

    compute in H.
    apply or_introl.
    rewrite <- H.
    reflexivity.

    unfold substO in H.
    simpl in H.
    assert (e' = term_var n).
      destruct n. rewrite <- H. reflexivity. rewrite <- H. reflexivity.

    apply or_intror.
    rewrite -> H0.
    assert (S n - 1 = n). omega.
    auto.
  Qed.

  Lemma lemm_subst_val :
    forall v s e, substO !v s = e -> exists v', e = !v'.
  Proof.
    intros.
    destruct v.

    Case "v = c".
    exists (val_const c).
    assert (!(val_const c) = e).  apply H.
    rewrite H0. reflexivity.

    Case "v = fix".
    exists (val_fix (subst_tyO t s)
                      (subst_ty t0 (shift_substO s))
                      (subst t1 (shift_subst (shift_substO s)))).
    assert (!(val_fix (subst_tyO t s)
                      (subst_ty t0 (shift_substO s))
                      (subst t1 (shift_subst (shift_substO s)))) = e).
      rewrite <- H.
      reflexivity.
    rewrite H0.
    reflexivity.

    Case "v = tabs".
    exists (val_tabs (subst t (tyshift_subst (0, s)))).
    rewrite <- H.
    reflexivity.

    Case "v = cast".
    exists (val_cast (subst_tyO t s) (subst_tyO t0 s) l).
    assert (!(val_cast (subst_tyO t s) (subst_tyO t0 s) l) = e).
      rewrite <- H.
      reflexivity.
    rewrite H0.
    reflexivity.
  Qed.

  Lemma lemm_left_red_not_blame :
    forall e1 e2 l, (e1 ~~> e2) -> e1 = term_blame l -> False.
  Proof.
    unfold not.
    intros.
    inversion H; try rewrite -> H0 in H1; try inversion H1.

    rewrite -> H0 in H2.
    inversion H2.
  Qed.

  Lemma lemm_val_size :
    forall v, size_val v = size_term !v.
  Proof.
    intros.
    reflexivity.
  Qed.

  Lemma lemm_ectx_plug_size :
    forall e E, size_term(E@[e]) = size_term(e) + size_ectx(E).
  Proof.
    intros.
    induction E.

    Case "E = []".
    intros.
    unfold ectx_plug.
    unfold size_ectx.
    rewrite <- plus_n_O.
    reflexivity.

    Case "E = app1".
    assert (size_term (ectx_app1 E t @[e]) =
            1 + (size_term (E@[e])) + size_term t).
      reflexivity.
    rewrite -> H.
    assert (size_ectx (ectx_app1 E t) = 1 + size_ectx E + size_term t).
    reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    omega.

    Case "E = app2".
    assert (size_term (ectx_app2 v E @[e]) =
            1 + size_term !v + size_term (E@[e])).
      reflexivity.
    rewrite -> H.
    assert (size_ectx (ectx_app2 v E) = 1 + size_val v + size_ectx E).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    rewrite lemm_val_size.
    omega.

    Case "E = tapp".
    assert (size_term (ectx_tapp E t @[e]) = 1 + size_term (E@[e]) + size_ty t).
      reflexivity.
    rewrite -> H.
    assert (size_ectx (ectx_tapp E t) = 1 + size_ectx E + size_ty t).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    omega.

    Case "E = wcheck".
    assert (size_term (ectx_wcheck t t0 E l @[e]) =
            1 + size_ty t + size_term t0 + size_term (E@[e])).
      reflexivity.
    rewrite -> H.
    assert (size_ectx (ectx_wcheck t t0 E l) =
            1 + size_ty t + size_term t0 + size_ectx E).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    repeat (rewrite plus_assoc).
    omega.

    Case "E = acheck".
    assert (size_term (ectx_acheck t t0 E v l @[e]) =
            1 + size_ty t + size_term t0 + size_term (E@[e]) + size_val v).
      reflexivity.
    rewrite -> H.
    assert (size_ectx (ectx_acheck t t0 E v l) =
            1 + size_ty t + size_term t0 + size_ectx E + size_val v).
      reflexivity.
    rewrite -> H0.
    omega.
  Qed.

  Lemma lemm_ectx_size_pos :
    forall E, E <> ectx_hole -> 0 < size_ectx(E).
  Proof.
    intros.
    destruct E;
      try (contradiction H; reflexivity);
      try (apply neq_0_lt;
           unfold not;
           intros;
           inversion H0).
  Qed.

  Lemma lemm_left_eval_not_blame :
    forall e1 e2 l, (e1 --> e2) -> e1 <> term_blame l.
  Proof.
    unfold not.
    intros.
    inversion H.

    Case "E[e1'] ~~> E[e2']".
    rewrite -> H0 in H2.
    destruct E; try (inversion H2).
      SCase "E = []".
      compute in H2.
      apply lemm_left_red_not_blame with (e1 := e0) (e2 := e3) (l := l).
      apply H1.
      apply H2.

    Case "E[blame l'] --> blame l'".
    rewrite -> H0 in H2.
    assert (size_term (E@[term_blame l0]) = size_term (term_blame l)).
      rewrite H2. reflexivity.
    assert (size_term(E@[term_blame l0]) =
            size_term (term_blame l0) + size_ectx(E)).
      apply lemm_ectx_plug_size.
    rewrite -> H4 in H5.
    inversion H5.
    assert (0 < size_ectx E).
      apply lemm_ectx_size_pos.
      apply H1.
    omega.
  Qed.

  Lemma lemm_red_blame_stuck :
    forall E l e, (E@[term_blame l] ~~> e) -> False.
  Proof.
    intros.
    induction E; inversion H; destruct E; discriminate.
  Qed.

  Lemma lemm_red_val_stuck :
    forall v e, (!v ~~> e) -> False.
  Proof.
    intros.
    inversion H.
  Qed.

  Lemma lemm_ectx_plug_blame :
    forall E e l, E@[e] = term_blame l -> E = ectx_hole.
  Proof.
    intros.
    destruct E; try reflexivity; try discriminate.
  Qed.

  Lemma lemm_ectx_plug_val :
    forall E e v, E@[e] = !v -> E = ectx_hole.
  Proof.
    intros.
    destruct E; try reflexivity; try discriminate.
  Qed.

  Lemma lemm_eval_val_stuck :
    forall v e, (!v --> e) -> False.
  Proof.
    intros.
    inversion H.

    Case "red".
    assert (E = ectx_hole).
      apply lemm_ectx_plug_val with (e := e1) (v := v).
      apply H0.
    rewrite -> H3 in H0.
    assert (e1 = !v). apply H0.
    rewrite -> H4 in H1.
    apply lemm_red_val_stuck with (v := v) (e := e2).
    apply H1.

    Case "blame".
    assert (E = ectx_hole).
      apply lemm_ectx_plug_val with (e := term_blame l) (v := v).
      apply H0.
    contradiction.
  Qed.

  Lemma lemm_subst_left_term_is_val :
    forall e e1 e2 v1,
      (e1 --> e2) ->
      substO e [e1] = !v1 ->
      exists v, e = !v.
  Proof.
    intros.
    induction e; try discriminate.

    Case "e = x".
    destruct v.
      SCase "substitued".
      assert (e1 = !v1). apply H0.
      exfalso.
      rewrite -> H1 in H.
      apply lemm_eval_val_stuck with (v := v1) (e := e2).
      apply H.
      SCase "not substituted".
      inversion H0.
      destruct v; discriminate.

    Case "e = v".
    exists v.
    reflexivity.
  Qed.

  Lemma lemm_coterm_right_val :
    forall e e1 e2 v2,
      (closed_term e1) ->
      (closed_term e2) ->
      (e1 --> e2) ->
      substO e [e2] = !v2 ->
      exists v', (substO e [e1] -->* !(subst_valO v' [e1])) /\
                 (v2 = subst_valO v' [e2]).
  Proof.
    unfold substO. unfold subst_valO. unfold closed_term. simpl.
    intros.
    destruct e; try discriminate.
      Case "e = x".
      destruct v.
        SCase "x[e2/x] = v2".
        assert (e2 = !v2). auto.
        exists v2.
        apply conj.
          SSCase "(1)".
          apply rt_step.
          rewrite H3 in H0.
          assert (!(subst_val v2 (0, [e1])) = !v2).
            simpl in H0.
            apply H0.
          rewrite H4.
          rewrite H3 in H1.
          assumption.

          SSCase "(2)".
          assert (subst !v2 (0, [e2]) = !v2).
            rewrite H3 in H0.
            apply H0.
          simpl in H4.
          injection H4; intros.
          rewrite H5.
          reflexivity.

        SCase "y[e2/x] = y".
        simpl in H2.
        destruct v; discriminate.

      Case "e = v".
      exists v.
      split.
        SCase "(1)".
        apply rt_refl.
        SCase "(2)".
        simpl in H2.
        injection H2; intros.
        rewrite H3.
        reflexivity.
  Qed.

  Lemma lemm_eval_blame_stuck :
    forall l e, (term_blame l --> e) -> False.
  Proof.
    intros.
    remember (term_blame l) as e'.
    destruct H.

    Case "blame l ~~> e".
    remember H as H0.
    destruct H.

      SCase "beta".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := substO e1 [!v2; fix_v]).
      apply H0'.

      SCase "tbeta".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e') (l := l). apply Heqe'.
      inversion H0.
      rewrite -> H in Heqe'.
      simpl in Heqe'.
      assert (H0' := H0).
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := tysubstO e [T]).
      assumption.

      SCase "cast_base".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with
              (e := !(val_cast (ty_base B) (ty_base B) l0) @ ! (v))
              (l := l).
        apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := !v).
      apply H0'.

      SCase "cast_fun".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := ! (val_fix
                   T11 T12
                   (! (val_fix T21' T12' (codomain_cast @ ((shift_n !v 4) @ term_var 0))) @
                      (domain_cast @ term_var 0)))).
      apply H0'.

      SCase "cast_tabs".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := ! (val_tabs (cast @ body))).
      apply H0'.

      SCase "cast_forget".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := ! (val_cast T1 T2 l0) @ ! (v)).
      apply H0'.

      SCase "cast_precheck".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := term_wcheck T1 e1 (! (val_cast T1 T2 l0) @ ! (v)) l0).
      apply H0'.

      SCase "cast_check".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e') (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := term_acheck T e (substO e [! (v)]) v l0).
      apply H0'.

      SCase "ok".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e0) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := !v).
      apply H0'.

      SCase "fail".
      assert (E = ectx_hole).
        apply lemm_ectx_plug_blame with (e := e0) (l := l). apply Heqe'.
      inversion H.
      assert (H0' := H0).
      rewrite -> H1 in Heqe'.
      unfold ectx_plug in Heqe'.
      rewrite -> Heqe' in H0'.
      apply lemm_red_blame_stuck with
        (E := ectx_hole)
        (l := l)
        (e := term_blame l0).
      apply H0'.

    Case "Eblame".
    assert (E = ectx_hole).
      apply lemm_ectx_plug_blame with (e := term_blame l0) (l := l). apply Heqe'.
    inversion H0.
    contradict H.
    apply H1.
  Qed.

  Lemma lemm_eval_multi_blame_refl :
    forall l e, (term_blame l -->* e) -> e = term_blame l.
  Proof.
    intros.
    remember (term_blame l) as e'.
    induction H.
    rewrite Heqe' in H.
    exfalso.
    apply lemm_eval_blame_stuck with (l := l) (e := y).
    auto.

    reflexivity.

    assert (y = x). auto.
    rewrite Heqe' in H1.
    assert (z = y). auto.
    rewrite -> H1 in H2.
    rewrite <- Heqe' in H2.
    auto.
  Qed.

  Lemma lemm_red_blame_exfalso :
    forall E1 E2 e e' l,
      E1@[e] = E2@[term_blame l] ->
      (e ~~> e') ->
      False.
  Proof.
    intros.
    generalize dependent E2.
    induction E1; intros; simpl in H0.
      Case "E1 = []".
      destruct E2; simpl in H;
        rewrite H in H0;
        inversion H0;
        try (unfold fix_v in H2);
        destruct E2; try discriminate.

      Case "E1 = app1".
      destruct E2; simpl in H; try discriminate.
        SCase "E2 = app1".
          injection H; intros. apply IHE1 with (E2 := E2). assumption.
        SCase "E2 = app2".
        injection H; intros.
        destruct E1;
          try (simpl in H2; rewrite H2 in H0; inversion H0);
          discriminate.

      Case "E1 = app2".
      destruct E2; simpl in H; try discriminate.
        SCase "E2 = app1".
        injection H; intros.
        destruct E2;
          try (simpl in H2; rewrite H2 in H0; inversion H0);
          discriminate.
        SCase "E2 = app2".
        injection H; intros. apply IHE1 with (E2 := E2). assumption.

      Case "E1 = tapp".
      destruct E2; simpl in H; try discriminate.
      injection H; intros. apply IHE1 with (E2 := E2). assumption.

      Case "E1 = wcheck".
      destruct E2; simpl in H; try discriminate.
      injection H; intros. apply IHE1 with (E2 := E2). assumption.

      Case "E1 = acheck".
      destruct E2; simpl in H; try discriminate.
      injection H; intros. apply IHE1 with (E2 := E2). assumption.
  Qed.

  Lemma lemm_ectx_blame_eq :
    forall E E' l l',
      E@[term_blame l] = E'@[term_blame l'] ->
      E = E'.
  Proof.
    intros.
    generalize dependent E'.
    induction E.

    Case "E = []".
    intros.
    assert (E' = ectx_hole).
      apply lemm_ectx_plug_blame with (e := term_blame l') (l := l).
      rewrite <- H.
      reflexivity.
    rewrite H0.
    reflexivity.

    Case "E = app1".
    intros.
    inversion H.
    destruct E'.
      SCase "E' = []".
      discriminate.
      SCase "E' = app1".
      inversion H.
      assert (E = E').
        apply IHE.
        apply H2.
      rewrite H0.
      reflexivity.
      SCase "E' = app2".
      inversion H1.
      assert (E = ectx_hole).
        apply lemm_ectx_plug_val with (e := term_blame l) (v := v).
        apply H2.
      rewrite -> H0 in H2.
      discriminate.
      SCase "E' = tapp".
      inversion H1.
      SCase "E' = wcheck".
      inversion H1.
      SCase "E' = acheck".
      inversion H1.

    Case "E = app2".
    intros.
    inversion H.
    destruct E'.
      SCase "E' = []".
      discriminate.
      SCase "E' = app1".
      inversion H1.
      assert (E' = ectx_hole).
        apply lemm_ectx_plug_val with (e := term_blame l') (v := v).
        rewrite H2.
        reflexivity.
      rewrite H0 in H2.
      discriminate.
      SCase "E' = app2".
      inversion H.
      assert (E = E').
        apply IHE.
        apply H3.
      rewrite -> H0.
      reflexivity.
      SCase "E' = tapp".
      inversion H1.
      SCase "E' = wcheck".
      inversion H1.
      SCase "E' = acheck".
      inversion H1.

    Case "E = tapp".
    intros.
    inversion H.
    destruct E'.
      SCase "E' = []".
      discriminate.
      SCase "E' = app1".
      inversion H1.
      SCase "E' = app2".
      inversion H.
      SCase "E' = tapp".
      inversion H1.
      assert (E = E').
        apply IHE.
        apply H2.
      rewrite H0.
      reflexivity.
      SCase "E' = wcheck".
      inversion H1.
      SCase "E' = acheck".
      inversion H1.

    Case "E = wcheck".
    intros.
    inversion H.
    destruct E'.
      SCase "E' = []".
      discriminate.
      SCase "E' = app1".
      inversion H1.
      SCase "E' = app2".
      inversion H.
      SCase "E' = tapp".
      inversion H1.
      SCase "E' = wcheck".
      inversion H1.
      assert (E = E').
        apply IHE.
        apply H4.
      rewrite H0.
      reflexivity.
      SCase "E' = acheck".
      inversion H1.

    Case "E = acheck".
    intros.
    inversion H.
    destruct E'.
      SCase "E' = []".
      discriminate.
      SCase "E' = app1".
      inversion H1.
      SCase "E' = app2".
      inversion H.
      SCase "E' = tapp".
      inversion H1.
      SCase "E' = wcheck".
      inversion H1.
      SCase "E' = acheck".
      inversion H1.
      assert (E = E').
        apply IHE.
        apply H4.
      rewrite H0.
      reflexivity.
  Qed.

  Lemma lemm_ectx_blame_eq_label :
    forall E l l',
      E@[term_blame l] = E@[term_blame l'] ->
      l = l'.
  Proof.
    intros.
    induction E; inversion H; try reflexivity; apply IHE; apply H1.
  Qed.

  Lemma lemm_eval_blame :
    forall E l e, (E@[term_blame l] --> e) -> e = term_blame l.
  Proof.
    intros.
    induction E.

    Case "E = []".
    unfold ectx_plug in H.
    exfalso.
    apply lemm_eval_blame_stuck with (l := l) (e := e).
    apply H.

    Case "E = app1".
    inversion H.
    exfalso.
    apply lemm_red_blame_exfalso with
      (E1 := E0)
      (E2 := ectx_app1 E t)
      (e := e1)
      (e' := e2)
      (l := l).
    apply H0.
    apply H1.
    assert (E0 = ectx_app1 E t).
      apply lemm_ectx_blame_eq with (l := l0) (l' := l).
      apply H0.
    rewrite -> H3 in H0.
    inversion H0.
    assert (l0 = l).
      apply lemm_ectx_blame_eq_label with (E := E).
      rewrite H5. reflexivity.
    rewrite H4. reflexivity.

    Case "E = app2".
    inversion H.
    exfalso.
    apply lemm_red_blame_exfalso with
      (E1 := E0)
      (E2 := ectx_app2 v E)
      (e := e1)
      (e' := e2)
      (l := l).
    apply H0.
    apply H1.
    assert (E0 = ectx_app2 v E).
      apply lemm_ectx_blame_eq with (l := l0) (l' := l).
      apply H0.
    rewrite -> H3 in H0.
    inversion H0.
    assert (l0 = l).
      apply lemm_ectx_blame_eq_label with (E := E).
      rewrite H5. reflexivity.
    rewrite H4. reflexivity.

    Case "E = tapp".
    inversion H.
    exfalso.
    apply lemm_red_blame_exfalso with
      (E1 := E0)
      (E2 := ectx_tapp E t)
      (e := e1)
      (e' := e2)
      (l := l).
    apply H0.
    apply H1.
    assert (E0 = ectx_tapp E t).
      apply lemm_ectx_blame_eq with (l := l0) (l' := l).
      apply H0.
    rewrite -> H3 in H0.
    inversion H0.
    assert (l0 = l).
      apply lemm_ectx_blame_eq_label with (E := E).
      rewrite H5. reflexivity.
    rewrite H4. reflexivity.

    Case "E = wcheck".
    inversion H.
    exfalso.
    apply lemm_red_blame_exfalso with
      (E1 := E0)
      (E2 := ectx_wcheck t t0 E l0)
      (e := e1)
      (e' := e2)
      (l := l).
    apply H0.
    apply H1.
    assert (E0 = ectx_wcheck t t0 E l0).
      apply lemm_ectx_blame_eq with (l := l1) (l' := l).
      rewrite -> H0.
      reflexivity.

    rewrite -> H3 in H0.
    inversion H0.
    assert (l1 = l).
      apply lemm_ectx_blame_eq_label with (E := E).
      rewrite H5. reflexivity.
    rewrite H4. reflexivity.

    Case "E = acheck".
    inversion H.
    exfalso.
    apply lemm_red_blame_exfalso with
      (E1 := E0)
      (E2 := ectx_acheck t t0 E v l0)
      (e := e1)
      (e' := e2)
      (l := l).
    apply H0.
    apply H1.
    assert (E0 = ectx_acheck t t0 E v l0).
      apply lemm_ectx_blame_eq with (l := l1) (l' := l).
      rewrite -> H0.
      reflexivity.

    rewrite -> H3 in H0.
    inversion H0.
    assert (l1 = l).
      apply lemm_ectx_blame_eq_label with (E := E).
      rewrite H5. reflexivity.
    rewrite H4. reflexivity.
  Qed.

  Lemma lemm_coterm_left_blame :
    forall e1 e2 e E1 l,
      (e1 --> e2) ->
      substO e [e1] = E1@[term_blame l] ->
      exists E2, substO e [e2] = E2@[term_blame l].
  Proof.
    intros.
    generalize dependent E1.
    induction e.

    Case "e = x".
    intros.
    destruct E1.
      SCase "E1 = []".
      unfold ectx_plug in H.
      apply lemm_subst_var in H0.
      inversion H0.
        SSCase "blame l = e1".
        assert (e1 <> term_blame l).
          apply lemm_left_eval_not_blame with (e2 := e2).
          apply H.
        rewrite <- H1 in H2.
        contradict H2.
        reflexivity.

        SSCase "blame l = y".
          discriminate.

      SCase "E1 = app1".
      exists ectx_hole.
      assert (ectx_hole@[term_blame l] = term_blame l).  reflexivity.
      rewrite -> H1.
      destruct v.
        SSCase "subst x e2 = e2".
        assert (e1 = ectx_app1 E1 t @[term_blame l]).
          apply H0.
        rewrite -> H2 in H.
        assert (e2 = term_blame l).
          apply lemm_eval_blame with (E := ectx_app1 E1 t).
          apply H.
        rewrite -> H3.
        reflexivity.
        SSCase "subst y e2 = y".
        destruct v; discriminate H0.

      SCase "E1 = app2".
      exists ectx_hole.
      assert (ectx_hole@[term_blame l] = term_blame l). reflexivity.
      rewrite -> H1.
      destruct v.
        SSCase "subst x e2 = e2".
        assert (e1 = (ectx_app2 v0 E1)@[term_blame l]).
          apply H0.
        rewrite -> H2 in H.
        assert (e2 = term_blame l).
          apply lemm_eval_blame with (E := ectx_app2 v0 E1).
          apply H.
        rewrite -> H3.
        reflexivity.
        SSCase "subst y e2 = y".
        destruct v; discriminate H0.

      SCase "E1 = tapp".
      exists ectx_hole.
      assert (ectx_hole@[term_blame l] = term_blame l). reflexivity.
      rewrite -> H1.
      destruct v.
        SSCase "subst x e2 = e2".
        assert (e1 = (ectx_tapp E1 t @[ term_blame l])). apply H0.
        rewrite -> H2 in H.
        assert (e2 = term_blame l).
          apply lemm_eval_blame with (E := ectx_tapp E1 t).
          apply H.
        rewrite -> H3.
        reflexivity.
        SSCase "subst y e2 = y".
        destruct v; discriminate H0.

      SCase "E1 = ectx_wcheck".
      exists ectx_hole.
      assert (ectx_hole@[term_blame l] = term_blame l). reflexivity.
      rewrite -> H1.
      destruct v.
        SSCase "subst x e2 = e2".
        assert (e1 = (ectx_wcheck t t0 E1 l0 @[ term_blame l])). apply H0.
        rewrite -> H2 in H.
        assert (e2 = term_blame l).
          apply lemm_eval_blame with (E := ectx_wcheck t t0 E1 l0).
          apply H.
        rewrite -> H3.
        reflexivity.
        SSCase "subst y e2 = y".
        destruct v; discriminate H0.

      SCase "E1 = ectx_acheck".
      exists ectx_hole.
      assert (ectx_hole@[term_blame l] = term_blame l). reflexivity.
      rewrite -> H1.
      destruct v.
        SSCase "subst x e2 = e2".
        assert (e1 = (ectx_acheck t t0 E1 v0 l0 @[ term_blame l])). apply H0.
        rewrite -> H2 in H.
        assert (e2 = term_blame l).
          apply lemm_eval_blame with (E := ectx_acheck t t0 E1 v0 l0).
          apply H.
        rewrite -> H3.
        reflexivity.
        SSCase "subst y e2 = y".
        destruct v; discriminate H0.

    Case "e = v".
    intros.
    assert (exists v', substO !(v) [e1] = !v').
      apply lemm_subst_val with (v := v) (s := [e1]).
      reflexivity.
    inversion H1.
    rewrite -> H2 in H0.
    assert (E1@[term_blame l] = !x). rewrite H0. reflexivity.
    assert (E1 = ectx_hole).
      apply lemm_ectx_plug_val with (e := term_blame l) (v := x).
      apply H3.
      rewrite -> H4 in H3.
    discriminate.

    Case "e = app".
    intros.
    destruct E1; try (inversion H0).
      SCase "E1 = app1".
      assert (exists E2', substO e3 [e2] = E2'@[term_blame l]).
        apply IHe1 with (E1 := E1).
        apply H2.
      inversion H1.
      exists (ectx_app1 x (substO e4 [e2])).
      assert (substO (e3 @ e4) [e2] = (substO e3 [e2]) @ (substO e4 [e2])).
        reflexivity.
      rewrite -> H5.
      rewrite -> H4.
      reflexivity.

      SCase "E1 = app2".
      assert (exists E2', substO e4 [e2] = E2'@[term_blame l]).
        apply IHe2 with (E1 := E1).
        apply H3.
        inversion H1.
      assert (exists v3, e3 = !v3).
        apply lemm_subst_left_term_is_val with (e1 := e1) (e2 := e2) (v1 := v).
        apply H.
        apply H2.
      inversion H5.
      exists (ectx_app2 (subst_valO x0 [e2]) x).
      assert (substO (e3 @ e4) [e2] = term_app (substO e3 [e2]) (substO e4 [e2])).
        reflexivity.
      rewrite -> H7.
      rewrite -> H6.
      rewrite -> H4.
      reflexivity.

    Case "e = tapp".
    intros.
    destruct E1; try (inversion H0).
      SCase "E1 = tapp".
      assert (exists E2', substO e [e2] = E2'@[term_blame l]).
        apply IHe with (E1 := E1).
        apply H2.
      inversion H1.
      exists (ectx_tapp x (subst_tyO t [e2])).
      assert (substO (e @@ t) [e2] = (substO e [e2]) @@ (subst_tyO t [e2])).
        reflexivity.
      rewrite -> H5.
      rewrite -> H4.
      reflexivity.

    Case "e = wcheck".
    intros.
    destruct E1; try (inversion H0).
    SCase "E1 = wcheck".
      assert (exists E2', substO e4 [e2] = E2'@[term_blame l]).
        apply IHe2 with (E1 := E1).
        apply H4.
      inversion H1.
      exists (ectx_wcheck (subst_tyO t [e2]) (subst e3 (shift_substO [e2])) x l1).
      assert (substO (term_wcheck t e3 e4 l1) [e2] =
              term_wcheck (subst_tyO t [e2])
                          (subst e3 (shift_substO [e2]))
                          (substO e4 [e2]) l1).
        reflexivity.
      rewrite -> H7.
      rewrite -> H6.
      reflexivity.

    Case "e = acheck".
    intros.
    destruct E1; try (inversion H0).
    SCase "E1 = acheck".
      assert (exists E2', substO e4 [e2] = E2'@[term_blame l]).
        apply IHe2 with (E1 := E1).
        apply H4.
      inversion H1.
      exists (ectx_acheck (subst_tyO t [e2])
                          (subst e3 (shift_substO [e2]))
                          x
                          (subst_valO v [e2])
                          l1).
      assert (substO (term_acheck t e3 e4 v l1) [e2] =
              term_acheck (subst_tyO t [e2])
                          (subst e3 (shift_substO [e2]))
                          (substO e4 [e2])
                          (subst_valO v [e2])
                          l1).
        reflexivity.
      rewrite -> H8.
      rewrite -> H7.
      reflexivity.

    Case "e = blame".
    intros.
    destruct E1; try (inversion H0).
      SCase "E1 = []".
      exists ectx_hole.
      reflexivity.
  Qed.

  Lemma lemm_ectx_compose :
    forall (E1 E2 : ectx) (e :term), E1@[E2@[e]] = (E1^^E2) @[e].
  Proof.
    intros.
    induction E1.
    SCase "E1 = []".
    reflexivity.
    SCase "E1 = app1".
    assert (ectx_app1 E1 t ^^ E2 = ectx_app1 (E1^^E2) t).
      reflexivity.
    rewrite -> H.
    assert (ectx_app1 E1 t @[E2@[e]] = term_app (E1@[E2@[e]]) t).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE1.
    reflexivity.
    SCase "E1 = app2".
    assert (ectx_app2 v E1 @[E2@[e]] = term_app !v (E1@[E2@[e]])).
      reflexivity.
    rewrite -> H.
    assert (ectx_app2 v E1 ^^ E2 = ectx_app2 v (E1^^E2)).
      reflexivity.
    rewrite -> H0.
    rewrite IHE1.
    reflexivity.
    SCase "E1 = tapp".
    assert (ectx_tapp E1 t @[E2@[e]] = term_tapp (E1@[E2@[e]]) t).
      reflexivity.
    rewrite -> H.
    assert (ectx_tapp E1 t ^^ E2 = ectx_tapp (E1^^E2) t).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE1.
    reflexivity.
    SCase "E1 = wcheck".
    assert (ectx_wcheck t t0 E1 l @[E2@[e]] = term_wcheck t t0 (E1@[E2@[e]]) l).
      reflexivity.
    rewrite -> H.
    assert (ectx_wcheck t t0 E1 l ^^ E2 = ectx_wcheck t t0 (E1^^E2)  l).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE1.
    reflexivity.
    SCase "E1 = acheck".
    assert (ectx_acheck t t0 E1 v l @[E2@[e]] =
            term_acheck t t0 (E1@[E2@[e]]) v l).
      reflexivity.
    rewrite -> H.
    assert (ectx_acheck t t0 E1 v l ^^ E2 = ectx_acheck t t0 (E1^^E2) v l).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE1.
    reflexivity.
  Qed.

  Lemma lemm_eval_ectx :
    forall e1 e2 E,
      (e1 -->* e2) ->
      (forall l, e2 <> term_blame l) ->
      E@[e1] -->* E@[e2].
  Proof.
    intros.
    induction H.

    Case "step".
    destruct H.
    SCase "red".
    assert (forall e, E@[E0@[e]] = (E^^E0)@[e]). apply lemm_ectx_compose.
    repeat (rewrite -> H1).
    apply rt_step. apply eval_red; repeat auto.
    rewrite -> lemm_ectx_compose.
    SCase "blame".
    assert (term_blame l <> term_blame l). auto.
    contradiction H1. reflexivity.

    Case "refl".
    apply rt_refl.

    Case "trans".
    remember y as y'.
    destruct y;
      try (apply rt_trans with (y := E@[y']));
      try (apply IHclos_refl_trans1;
           unfold not;
           intros;
           rewrite -> Heqy' in H2;
           discriminate);
      try (apply IHclos_refl_trans2;
           apply H0).

    SCase "y = blame".
    rewrite -> Heqy' in H1.
    assert (z = term_blame l). apply lemm_eval_multi_blame_refl. auto.
    assert (z <> term_blame l). auto.
    contradiction.
  Qed.

  Lemma lemm_eval_blame_eq :
    forall E l l', (E@[term_blame l] -->* term_blame l') -> l = l'.
  Proof.
    intros.
    remember (E@[term_blame l]) as e.
    remember (term_blame l') as e'.
    assert (clos_refl_trans_1n term eval e e').
      apply clos_rt_rt1n. auto.
    generalize dependent E.
    induction H0; intros.
    SSCase "refl".
      rewrite Heqe in Heqe'.
      destruct E; inversion Heqe'; try reflexivity.
    SSCase "1n".
      rewrite -> Heqe in H.
      rewrite -> Heqe' in H.
      rewrite -> Heqe in H0.
      assert (y = term_blame l).
        apply lemm_eval_blame with (E := E).
        auto.
        assert (forall l e, (term_blame l -->* e) -> e = term_blame l).
          intros.
          remember (term_blame l0) as e'.
          induction H3.
          exfalso.
          apply lemm_eval_blame_stuck with (l := l0) (e := y0).
            rewrite <- Heqe'0. auto.
          reflexivity.
          assert (y0 = x0). auto.
          rewrite Heqe'0 in H3.
          assert (z0 = y0). auto.
          rewrite -> Heqe'0.
          rewrite -> H4.
          rewrite -> H3.
          reflexivity.
        apply IHclos_refl_trans_1n with (E := ectx_hole).
        auto.
        apply clos_rt1n_rt. auto.
        auto.
  Qed.

  Lemma lemm_eval_blame_ectx :
    forall e l E, (e -->* term_blame l) -> (E@[e] -->* term_blame l).
  Proof.
    intros.
    remember (term_blame l) as e'.
    induction H.
    Case "one step".
    rewrite Heqe' in H.
    destruct E; rewrite -> Heqe'.
      SCase "E = []".
      compute.
      apply rt_step.
      auto.

      SCase "E = app1".
      assert (forall e E0, ectx_app1 E t @[E0@[e]] = ectx_app1 (E^^E0) t @[e]).
        intros.
        apply lemm_ectx_compose.
      inversion H.
        SSCase "red".
        rewrite H0.
        rewrite H1.
        apply rt_trans with (y := ectx_app1 (E ^^ E0) t @[term_blame l]).
        apply rt_step. apply eval_red.
        destruct E0; try discriminate.
        simpl in H1.
        rewrite H1 in H3.
        assumption.
        apply rt_step. apply eval_blame.
        discriminate.

        SSCase "blame".
        apply rt_step.
        rewrite -> H0.
        apply eval_blame.
        discriminate.

      SCase "E = app2".
      assert (forall e E0, ectx_app2 v E @[E0@[e]] = ectx_app2 v (E^^E0) @[e]).
        intros.
        apply lemm_ectx_compose.
      inversion H.
        SSCase "red".
        rewrite -> H0.
        apply rt_trans with (y := (ectx_app2 v (E^^E0)) @[e2]).
        apply rt_step. apply eval_red. assumption.
        rewrite -> H1.
        apply rt_step. rewrite <- H0. rewrite H1. apply eval_blame. discriminate.

        SSCase "blame".
        apply rt_step.
        rewrite -> H0.
        apply eval_blame.
        discriminate.

      SCase "E = tapp".
      assert (forall e E0, ectx_tapp E t @[E0@[e]] = ectx_tapp (E^^E0) t @[e]).
        intros. apply lemm_ectx_compose.
      inversion H.
       SSCase "red".
        rewrite -> H0.
        apply rt_trans with (y := ectx_tapp (E^^E0) t @[e2]).
        apply rt_step. apply eval_red. auto. auto. auto.
        apply rt_step.
        rewrite -> H1.
        rewrite <- H0.
        rewrite -> H1.
        apply eval_blame.
        discriminate.

        SSCase "blame".
        apply rt_step.
        rewrite -> H0.
        apply eval_blame.
        discriminate.

      SCase "E = wcheck".
      assert (forall e E0, (ectx_wcheck t t0 E l0) @[E0@[e]] =
                           (ectx_wcheck t t0 (E^^E0) l0) @[e]).
        intros. apply lemm_ectx_compose.
      inversion H.
        SSCase "red".
       rewrite -> H0.
        apply rt_trans with (y := ectx_wcheck t t0 (E^^E0) l0 @[e2]).
        apply rt_step. apply eval_red. auto. auto. auto.
        apply rt_step.
        rewrite -> H1.
        rewrite <- H0.
        rewrite -> H1.
        apply eval_blame.
        discriminate.

        SSCase "blame".
        apply rt_step.
        rewrite -> H0.
        apply eval_blame.
        discriminate.

      SCase "E = acheck".
      assert (forall e E0, (ectx_acheck t t0 E v l0) @[E0@[e]] =
                           (ectx_acheck t t0 (E^^E0) v l0) @[e]).
        intros. apply lemm_ectx_compose.
      inversion H.
        SSCase "red".
        rewrite -> H0.
        apply rt_trans with (y := ectx_acheck t t0 (E^^E0) v l0 @[e2]).
        apply rt_step. apply eval_red. auto. auto. auto.
        apply rt_step.
        rewrite -> H1.
        rewrite <- H0.
        rewrite -> H1.
        apply eval_blame.
        discriminate.

        SSCase "blame".
        apply rt_step.
        rewrite -> H0.
        apply eval_blame.
        discriminate.

    Case "refl".
    rewrite -> Heqe'.
    destruct E;
      try (apply rt_refl);
      try (apply rt_step; apply eval_blame; discriminate).

    Case "trans".
    assert (E@[y] -->* z). auto.
    rewrite -> Heqe'.
    rewrite -> Heqe' in H1.

    assert ((forall l, y <> term_blame l) -> E@[x] -->* term_blame l).
     intros.
     apply rt_trans with (y := E@[y]).
     apply lemm_eval_ectx.
     auto.
     apply H2.
     auto.

    remember y as y'.
    destruct y;
      try (apply H2; rewrite Heqy'; discriminate).

      SCase "y = blame".
      rewrite -> Heqy' in H1.
        SSCase "trans".
        assert (l0 = l). apply lemm_eval_blame_eq with (E := E). auto.
        rewrite -> H3 in Heqy'.
        rewrite -> Heqy' in IHclos_refl_trans1.
        auto.
  Qed.

  Lemma lemm_coterm_right_blame :
    forall e1 e2 e E2 l,
      (closed_term e1) ->
      (closed_term e2) ->
      (e1 --> e2) ->
      substO e [e2] = E2@[term_blame l] ->
      substO e [e1] -->* term_blame l.
  Proof.
    intros.
    generalize dependent E2.
    induction e.

    Case "e = x".
    intros.
    destruct v.
      SCase "x[e2/x] = E2[blame l]".
      assert (substO (term_var 0) [e1] = e1). reflexivity.
      assert (substO (term_var 0) [e2] = e2). reflexivity.
      rewrite -> H3.
      rewrite -> H4 in H2.
      rewrite -> H2 in H1.
      assert (E2@[term_blame l] -->* term_blame l).
        destruct E2;
          try (apply rt_refl);
          try (apply rt_step; apply eval_blame; unfold not; discriminate).
      apply rt_trans with (y := E2@[term_blame l]).
      apply rt_step. apply H1. apply H5.

      SCase "x[e2/y] = x".
      destruct v; inversion H2; destruct E2; discriminate.

    Case "e = v".
    intros.
    destruct E2; discriminate.

    Case "e = app".
    intros.
    destruct E2; try discriminate.
      SCase "E2 = app1".
      inversion H2.
      assert (substO e3 [e1] -->* term_blame l).
        apply IHe1 with (E2 := E2).
        auto.
      assert (substO (e3 @ e4) [e1] = (substO e3 [e1]) @ (substO e4 [e1])).
        reflexivity.
      rewrite -> H6.
      assert (substO e3 [e1] @ substO e4 [e1] =
              (ectx_app1 ectx_hole (substO e4 [e1])) @[substO e3 [e1]]).
        reflexivity.
      rewrite -> H7.
      apply lemm_eval_blame_ectx.
      auto.

      SCase "E2 = app2".
      inversion H2.
      assert (substO e4 [e1] -->* term_blame l).
        apply IHe2 with (E2 := E2).
        auto.
      assert (substO (e3 @ e4) [e1] = (substO e3 [e1]) @ (substO e4 [e1])).
        reflexivity.
      rewrite -> H6.
      assert (exists v', (substO e3 [e1] -->* !(subst_valO v' [e1])) /\
                         (v = subst_valO v' [e2])).
        apply lemm_coterm_right_val with (v2 := v); auto.
      inversion H7.
      inversion H8.
      assert (substO e3 [e1] @ substO e4 [e1] =
              (ectx_app1 ectx_hole (substO e4 [e1])) @[substO e3 [e1]]).
        reflexivity.
      rewrite -> H11.
      apply rt_trans with
        (y := ectx_app1 ectx_hole (substO e4 [e1]) @[ !(subst_valO x [e1]) ]).
      apply lemm_eval_ectx.
      auto. discriminate.
      assert (ectx_app1 ectx_hole (substO e4 [e1]) @[ !(subst_valO x [e1]) ] =
              (ectx_app2 (subst_valO x [e1]) ectx_hole) @[substO e4 [e1]]).
        reflexivity.
      rewrite -> H12.
      apply lemm_eval_blame_ectx.
      auto.

    Case "e = tapp".
    intros.
    destruct E2; try discriminate.
    assert (substO (e@@t) [e1] =
            (ectx_tapp ectx_hole (subst_tyO t [e1]) @[substO e [e1]])).
      reflexivity.
    rewrite -> H3.
    apply lemm_eval_blame_ectx.
    apply IHe with (E2 := E2).
    inversion H2.
    auto.

    Case "e = wcheck".
    intros.
    destruct E2; try discriminate.
    assert (substO (term_wcheck t e3 e4 l0) [e1] =
            (ectx_wcheck (subst_tyO t [e1])
                         (subst e3 (shift_substO [e1]))
                         ectx_hole
                         l0)
              @[substO e4 [e1]]).
      reflexivity.
    rewrite -> H3.
    apply lemm_eval_blame_ectx.
    apply IHe2 with (E2 := E2).
    inversion H2.
    auto.

    Case "e = acheck".
    intros.
    destruct E2; try discriminate.
    assert (substO (term_acheck t e3 e4 v l0) [e1] =
            (ectx_acheck (subst_tyO t [e1])
                         (subst e3 (shift_substO [e1]))
                         ectx_hole
                         (subst_valO v [e1])
                         l0)
              @[substO e4 [e1]]).
      reflexivity.
    rewrite -> H3.
    apply lemm_eval_blame_ectx.
    apply IHe2 with (E2 := E2).
    inversion H2.
    auto.

    Case "e = blame".
    intros.
    assert (forall e, substO (term_blame l0) [e] = term_blame l0).
      reflexivity.
    rewrite -> H3.
    rewrite -> H3 in H2.
    assert (term_blame l0 = term_blame l).
      destruct E2; try discriminate.
      auto.
    rewrite -> H4.
    apply rt_refl.
  Qed.

  Lemma lemm_subst_ectx :
    forall E e s, substO (E@[e]) s = (subst_ectxO E s) @[substO e s].
  Proof.
    intros.
    induction E.

    Case "E = []".
    reflexivity.

    Case "E = app1".
    assert (ectx_app1 E t @[e] = term_app (E@[e]) t).
      reflexivity.
    rewrite -> H.
    assert (substO (E@[e] @ t) s = (substO (E@[e]) s) @ (substO t s)).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    reflexivity.

    Case "E = app2".
    assert (ectx_app2 v E @[e] = term_app !v (E@[e])).
      reflexivity.
    rewrite -> H.
    assert (substO (!v @ (E@[e])) s = !(subst_valO v s) @ (substO (E@[e]) s)).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    reflexivity.

    Case "E = tapp".
    assert (ectx_tapp E t @[e]  = term_tapp (E@[e]) t).
      reflexivity.
    rewrite -> H.
    assert (substO (term_tapp (E@[e]) t) s =
            term_tapp (substO (E@[e]) s) (subst_tyO t s)).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    reflexivity.

    Case "E = wcheck".
    assert (ectx_wcheck t t0 E l @[e] = term_wcheck t t0 (E@[e]) l).
      reflexivity.
    rewrite -> H.
    assert (substO (term_wcheck t t0 (E@[e]) l) s =
            term_wcheck (subst_tyO t s)
                        (subst t0 (shift_substO s))
                        (substO (E@[e]) s)
                        l).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    reflexivity.

    Case "E = acheck".
    assert (ectx_acheck t t0 E v l @[e] = term_acheck t t0 (E@[e]) v l).
      reflexivity.
    rewrite -> H.
    assert (substO (term_acheck t t0 (E@[e]) v l) s =
            term_acheck (subst_tyO t s)
                        (subst t0 (shift_substO s))
                        (substO (E@[e]) s)
                        (subst_valO v s)
                        l).
      reflexivity.
    rewrite -> H0.
    rewrite -> IHE.
    reflexivity.
  Qed.

  Lemma lemm_subst_empty :
    (forall e n, subst e (n, []) = e) /\
    (forall v n, subst_val v (n, []) = v) /\
    (forall T n, subst_ty T (n, []) = T).
  Proof.
    apply term_val_ty_mind; try reflexivity.

    Case "e = x".
    intros.
    destruct v.
    simpl.
    destruct n; reflexivity.
    simpl.
    destruct (lt_dec (S v)).
    reflexivity.
    destruct (match n with | 0 => S v | S l => v - l end); reflexivity.
    Case "e = v".
    intros.
    remember (subst !v (n, [])) as x.
    rewrite <- (H n). rewrite -> Heqx. reflexivity.
    Case "e = app".
    intros.
    remember (subst (t@t0) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 n). rewrite -> Heqx. reflexivity.
    Case "e = tapp".
    intros.
    remember (subst (t@@t0) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 n). rewrite -> Heqx. reflexivity.
    Case "e = wcheck".
    intros.
    remember (subst (term_wcheck t t0 t1 l) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 (S n)). rewrite <- (H1 n). rewrite -> Heqx.
    simpl.
    assert (shift_subst (n, []) = (S n, [])). reflexivity.
    rewrite -> H2.
    reflexivity.
    Case "e = acheck".
    intros.
    remember (subst (term_acheck t t0 t1 v l) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 (S n)).
    rewrite <- (H1 n). rewrite <- (H2 n). rewrite -> Heqx.
    simpl.
    assert (shift_subst (n, []) = (S n, [])). reflexivity.
    rewrite -> H2.
    reflexivity.

    Case "v = fix".
    intros.
    remember (subst_val (val_fix t t0 t1) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 (S n)). rewrite <- (H1 (S (S n))).
    rewrite -> Heqx.
    reflexivity.
    Case "v = tabs".
    intros.
    remember (subst_val (val_tabs t) (n, [])) as x.
    rewrite <- (H n). rewrite -> Heqx. reflexivity.
    Case "v = cast".
    intros.
    remember (subst_val (val_cast t t0 l) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 n). rewrite -> Heqx. reflexivity.

    Case "T = tyfun".
    intros.
    remember (subst_ty (ty_fun t t0) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 (S n)). rewrite -> Heqx. reflexivity.
    Case "T = forall".
    intros.
    remember (subst_ty (ty_forall t) (n, [])) as x.
    rewrite <- (H n). rewrite -> Heqx. reflexivity.
    Case "T = refine".
    intros.
    remember (subst_ty (ty_refine t t0) (n, [])) as x.
    rewrite <- (H n). rewrite <- (H0 (S n)). rewrite -> Heqx. reflexivity.
  Qed.

  Lemma lemm_subst_shift_ident :
    (forall e n e1 e2, subst (shift (shift e n) n) (n, [e1; e2]) = e) /\
    (forall v n e1 e2, subst_val (shift_val (shift_val v n) n) (n, [e1; e2]) = v) /\
    (forall T n e1 e2, subst_ty (shift_ty (shift_ty T n) n) (n, [e1; e2]) = T).
  Proof.
    assert (forall x, x + 1 = S x) as Hsucc. intro. omega.
    assert (forall x, x + 2 = S (S x)) as Hssucc. intro. omega.

    apply term_val_ty_mind; intros; simpl;
      try (unfold shift_subst);
      try (unfold tyshift_subst);
      simpl;
      try (rewrite -> Hsucc);
      try (rewrite -> Hssucc);
      try (rewrite -> H);
      try (rewrite -> H0);
      try (rewrite -> H1);
      try (rewrite -> H2);
      try reflexivity.

    Case "e = x".
    repeat (rewrite -> Hsucc).
    destruct (le_gt_dec n v).
      SCase "n <= v".
      destruct (lt_dec n (S v)); try omega.
      destruct (le_gt_dec n (S v)); try omega.
      destruct (lt_dec (S (S v)) n); try omega.
      remember (S (S v) - n) as ssv_n.
      destruct ssv_n; try omega.
      destruct ssv_n; try omega.
      assert (S (S v) - 2 = v). omega.
      rewrite -> H.
      destruct ssv_n; reflexivity.

      SCase "n > v".
      destruct (le_gt_dec n v); try omega.
      destruct (lt_dec v n); try omega.
      reflexivity.
  Qed.

  Lemma lemm_tysubst_tyshift_ident :
    (forall e n T', tysubst (tyshift e n) (n, [T']) = e) /\
    (forall v n T', tysubst_val (tyshift_val v n) (n, [T']) = v) /\
    (forall T n T', tysubst_ty (tyshift_ty T n) (n, [T']) = T).
  Proof.
    apply term_val_ty_mind; intros; simpl;
      try (unfold shift_subst);
      try (unfold tyshift_subst);
      try (unfold shift_tysubst);
      try (unfold tyshift_tysubst);
      simpl;
      try (rewrite -> H);
      try (rewrite -> H0);
      try (rewrite -> H1);
      try (rewrite -> H2);
      try reflexivity.

    Case "e = x".
    destruct (le_gt_dec n t).
      SCase "n <= t".
      destruct (lt_dec (S t) n); try omega.
      remember (S t - n) as Stn.
      destruct Stn; try omega.
      destruct Stn; f_equal; omega.

      SCase "n > t".
      destruct (lt_dec t n); try omega.
      reflexivity.
  Qed.

  Lemma lemm_nth_map :
    forall A B x (f : A->B) l d d',
      x < length l ->
      f (nth x l d) = nth x (map f l) d'.
  Proof.
    intros.
    generalize dependent x.
    induction l.
    Case "l = nil".
    intros.
    destruct x; simpl in H; omega.
    Case "l = cons".
    intros.
    destruct x.
    reflexivity.
    simpl.
    simpl in H.
    assert (x < length l). omega.
    apply IHl.
    assumption.
  Qed.

  Lemma lemm_shift_swap :
    (forall e m n,
        n <= m ->
        shift (shift e m) n = shift (shift e n) (S m)) /\
    (forall v m n,
        n <= m ->
        shift_val (shift_val v m) n = shift_val (shift_val v n) (S m)) /\
    (forall T m n,
        n <= m ->
        shift_ty (shift_ty T m) n = shift_ty (shift_ty T n) (S m)).
  Proof.
    apply term_val_ty_mind; intros; simpl;
      try (rewrite -> H);
      try (rewrite -> H0);
      try (rewrite -> H1);
      try (rewrite -> H2);
      try reflexivity;
      try (repeat omega).

    Case "e = x".
    destruct (le_gt_dec m v).
      SCase "m <= v".
      destruct (le_gt_dec n v); try omega.
      destruct (le_gt_dec n (S v)); try omega.
      destruct (le_gt_dec m v); try omega.
      reflexivity.

      SCase "m > v".
      destruct (le_gt_dec n v).
        SSCase "n <= v".
        destruct (le_gt_dec m v); try omega; reflexivity.
        SSCase "n > v".
        destruct v.
          SSSCase "v = O". reflexivity.
          SSSCase "v = S v'".
          destruct (le_gt_dec m v); try omega.
          reflexivity.
  Qed.

  Lemma lemm_fun_shift_swap :
    forall m n,
        n <= m ->
        (fun e => shift (shift e m) n) = (fun e => shift (shift e n) (S m)).
  Proof.
    intros.
    apply functional_extensionality_dep.
    intro.
    apply lemm_shift_swap.
    omega.
  Qed.

  Lemma lemm_shift_swap3 :
    forall e l m n,
      m <= l ->
      n <= S l ->
      shift (shift (shift e l) m) n =
      shift (shift (shift e m) n) (S (S l)).
  Proof.
    intros.
    assert (shift (shift e l) m = shift (shift e m) (S l)).
      apply lemm_shift_swap.
      assumption.
    rewrite -> H1.
    apply lemm_shift_swap.
    assumption.
  Qed.

  Lemma lemm_fun_shift_swap3 :
    forall l m n,
      m <= l ->
      n <= S l ->
      (fun e => shift (shift (shift e l) m) n) =
      (fun e => shift (shift (shift e m) n) (S (S l))).
  Proof.
    intros.
    apply functional_extensionality_dep.
    intro.
    apply lemm_shift_swap3; assumption.
  Qed.

  Lemma lemm_tyshift_swap :
    (forall e m n,
        n <= m ->
        tyshift (tyshift e m) n = tyshift (tyshift e n) (S m)) /\
    (forall v m n,
        n <= m ->
        tyshift_val (tyshift_val v m) n = tyshift_val (tyshift_val v n) (S m)) /\
    (forall T m n,
        n <= m ->
        tyshift_ty (tyshift_ty T m) n = tyshift_ty (tyshift_ty T n) (S m)).
  Proof.
    apply term_val_ty_mind; intros; simpl;
      try (rewrite -> H);
      try (rewrite -> H0);
      try (rewrite -> H1);
      try (rewrite -> H2);
      try reflexivity;
      try (repeat omega).

    Case "T = A".
    destruct (le_gt_dec m t).
      SCase "m <= A".
      destruct (le_gt_dec n t); try omega.
      destruct (le_gt_dec n (S t)); try omega.
      destruct (le_gt_dec m t); try omega.
      reflexivity.

      SCase "m > A".
      destruct (le_gt_dec n t).
        SSCase "n <= A".
        destruct (le_gt_dec m t); try omega; reflexivity.
        SSCase "n > A".
        destruct t.
          SSSCase "A = O". reflexivity.
          SSSCase "A = S A'".
          destruct (le_gt_dec m t); try omega.
          reflexivity.
  Qed.

  Lemma lemm_fun_tyshift_swap :
    forall m n, n <= m ->
                (fun e => tyshift (tyshift e m) n) =
                (fun e => tyshift (tyshift e n) (S m)).
  Proof.
    intros.
    apply functional_extensionality_dep.
    intro.
    apply lemm_tyshift_swap.
    omega.
  Qed.

  Lemma lemm_shift_tyshift :
    (forall e m n,
        shift (tyshift e n) m = tyshift (shift e m) n) /\
    (forall v m n,
        shift_val (tyshift_val v n) m = tyshift_val (shift_val v m) n) /\
    (forall T m n,
        shift_ty (tyshift_ty T n) m = tyshift_ty (shift_ty T m) n).
  Proof.
    apply term_val_ty_mind;
      intros;
      simpl;
      try (rewrite -> H);
      try (rewrite -> H0);
      try (rewrite -> H1);
      try (rewrite -> H2);
      reflexivity.
  Qed.

  Lemma lemm_fun_shift_tyshift :
    forall m n,
      (fun x => shift (tyshift x n) m) = (fun x => tyshift (shift x m) n).
  Proof.
    intros.
    apply functional_extensionality_dep.
    intro.
    apply lemm_shift_tyshift.
  Qed.

  Lemma lemm_fun_shift_tyshift3 :
    forall l m n,
      (fun x => shift (shift (tyshift x n) m) l) =
      (fun x => tyshift (shift (shift x m) l) n).
  Proof.
    intros.
    apply functional_extensionality_dep.
    intro.
    destruct lemm_shift_tyshift.
    rewrite <- H.
    rewrite <- H.
    reflexivity.
  Qed.

  Lemma lemm_subst_shift_eq_shift_subst :
    (forall e m n l,
        m <= n ->
        subst (shift e m) (S n, map (fun e' => shift e' m) l) =
        shift (subst e (n, l)) m) /\
    (forall v m n l,
        m <= n ->
        subst_val (shift_val v m) (S n, map (fun e' => shift e' m) l) =
        shift_val (subst_val v (n, l)) m) /\
    (forall T m n l,
        m <= n ->
        subst_ty (shift_ty T m) (S n, map (fun e' => shift e' m) l) =
        shift_ty (subst_ty T (n, l)) m).
  Proof.
    apply term_val_ty_mind; intros; simpl;
      try (unfold shift_subst);
      try (unfold tyshift_subst);
      simpl;
      try rewrite -> H;
      try rewrite -> H0;
      try rewrite -> H1;
      try rewrite -> H2;
      try reflexivity;
      try assumption.

    Case "e = x".

    destruct (le_gt_dec m v).
      SCase "m <= v".
      destruct (lt_dec v n).
        SSCase "v < n".
        destruct (lt_dec (S v) (S n)); try omega.
        simpl.
        destruct (le_gt_dec m v); try reflexivity; try omega.
        SSCase "v >= n".
        destruct (lt_dec (S v) (S n)); try omega.
        assert (S v - S n = v - n). omega.
        rewrite -> H0.
        assert (length (map (fun e' : term => shift e' m) l) = length l).
          apply map_length.
        rewrite -> H1.
        simpl.

        remember (fun e' => shift e' m) as f.
        destruct (lt_dec (v - n) (length l)).
          SSSCase "v - n < length l".
          assert (forall d d', f (nth (v - n) l d) = nth (v - n) (map f l) d').
            intros.
            apply lemm_nth_map.
            assumption.
          destruct (length l).
            SSSSCase "length l = O".
            rewrite <- (H2 (term_var (v - 0)) (term_var (S v))).
            rewrite -> Heqf.
            reflexivity.

            SSSSCase "length l > 0".
            rewrite <- (H2 (term_var (v - S n2)) (term_var (v - n2))).
            rewrite -> Heqf.
            reflexivity.

          SSSCase "v - n >= length l".
          rewrite -> nth_overflow; try omega.
          rewrite -> nth_overflow; try omega.
          destruct (length l).
            SSSSCase "length l = O".
            destruct (le_gt_dec m (v - 0)); try omega.
            assert (v - 0 = v). omega.
            rewrite -> H2.
            simpl.
            destruct (le_gt_dec m v); try omega.
            reflexivity.

            SSSSCase "length l > O".
            simpl.
            destruct (le_gt_dec m (v - S n3)); try omega.
              assert (v - n3 = S (v - S n3)). omega.
              rewrite <- H2.
              reflexivity.

      SCase "m > n".
      destruct (lt_dec v n).
        SSCase "v < n".
        destruct (lt_dec v (S n)); try omega.
        simpl.
        destruct (le_gt_dec m v); try omega.
        reflexivity.

        SSCase "v >= n".
        destruct (lt_dec v (S n)); omega.

    Case "e = wcheck".
    assert (subst (shift t0 (1 + m))
                   (shift_subst (S n, map (fun e' : term => shift e' m) l0)) =
            shift (subst t0 (shift_subst (n, l0))) (1 + m)).
      unfold shift_subst.
      simpl.
      rewrite -> map_map.
      rewrite -> lemm_fun_shift_swap.
      rewrite <- H0.
      rewrite -> map_map.
      reflexivity.
      omega.
      omega.

    Case "e = wcheck".
    unfold shift_subst.
    simpl.
    rewrite <- H0.
    rewrite -> map_map.
    rewrite -> map_map.
    rewrite -> lemm_fun_shift_swap.
    reflexivity.
    omega.
    omega.

    Case "e = acheck".
    unfold shift_subst.
    simpl.
    rewrite <- H0.
    rewrite -> map_map.
    rewrite -> map_map.
    rewrite -> lemm_fun_shift_swap.
    reflexivity.
    omega.
    omega.

    Case "v = fix".
    unfold shift_subst.
    simpl.
    rewrite <- H0.
    rewrite <- H1.
    f_equal.
    f_equal.
    rewrite -> map_map.
    rewrite -> map_map.
    rewrite -> lemm_fun_shift_swap.
    reflexivity.
    omega.
    f_equal.
    f_equal.
    rewrite -> map_map.
    rewrite -> map_map.
    rewrite -> map_map.
    rewrite -> map_map.
    f_equal.
    apply lemm_fun_shift_swap3; omega.
    omega.
    omega.

    f_equal.
    rewrite -> map_map.
    generalize lemm_shift_tyshift; intros.
    inversion H1.
    rewrite <- lemm_fun_shift_tyshift.
    assert (map (fun x => shift (tyshift x 0) m) l =
            map (fun x => shift x m) (map (fun x => tyshift x 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H4.
    rewrite -> H; try reflexivity; omega.

    Case "T = fun".
    f_equal.
    unfold shift_subst.
    simpl.
    rewrite -> map_map.
    rewrite <- H0.
    f_equal.
    f_equal.
    rewrite -> map_map.
    f_equal.
    apply lemm_fun_shift_swap; omega.
    omega.

    Case "T = forall".
    f_equal.
    unfold shift_subst.
    simpl.
    rewrite -> map_map.
    rewrite <- lemm_fun_shift_tyshift.
    assert (map (fun x => shift (tyshift x 0) m) l =
            map (fun x => shift x m) (map (fun x => tyshift x 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H1.
    rewrite -> H; try omega.
    reflexivity.

    Case "T = refine".
    f_equal.
    unfold shift_subst.
    simpl.
    rewrite -> map_map.
    rewrite -> lemm_fun_shift_swap; try omega.
    assert (map (fun x => shift (shift x 0) (S m)) l =
            map (fun x => shift x (S m)) (map (fun x => shift x 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H2.
    rewrite -> H0; try omega.
    reflexivity.
  Qed.


  Lemma lemm_tyshift_subst :
    (forall e m n l,
        tyshift (subst e (m, l)) n =
        subst (tyshift e n) (m, map (fun e' => tyshift e' n) l)) /\
    (forall v m n l,
        tyshift_val (subst_val v (m, l)) n =
        subst_val (tyshift_val v n) (m, map (fun e' => tyshift e' n) l)) /\
    (forall T m n l,
        tyshift_ty (subst_ty T (m, l)) n =
        subst_ty (tyshift_ty T n) (m, map (fun e' => tyshift e' n) l)).
  Proof.
    apply term_val_ty_mind; intros;
      simpl;
      try rewrite -> H;
      try rewrite -> H0;
      try rewrite -> H1;
      try unfold shift_subst;
      simpl;
      try reflexivity.

    Case "e = x".
    destruct (lt_dec v m).
      SCase "v < m".
      reflexivity.
      SCase "v >= m".
      simpl.
      rewrite map_length.
      destruct (lt_dec (v - m) (length l)).
        SSCase "v - m < length l".
        remember (fun x => tyshift x n) as f.
        assert (tyshift (nth (v - m) l (term_var (v - Datatypes.length l))) n =
                f (nth (v - m) l (term_var (v - Datatypes.length l)))).
          rewrite -> Heqf. reflexivity.
        rewrite -> H.
        rewrite lemm_nth_map with (d' := term_var (v - Datatypes.length l)).
        reflexivity.
        assumption.
        SSCase "v - m >= length l".
        rewrite -> nth_overflow; try omega.
        rewrite -> nth_overflow.
        reflexivity.
        rewrite map_length.
        omega.

    Case "e = wcheck".
    rewrite -> H0.
    f_equal. f_equal.
    repeat (rewrite map_map).
    rewrite <- lemm_fun_shift_tyshift.
    reflexivity.

    rewrite -> H0.
    f_equal. f_equal.
    repeat (rewrite map_map).
    rewrite <- lemm_fun_shift_tyshift.
    reflexivity.

    rewrite -> H2.
    reflexivity.

    Case "v = fix".
    f_equal.
    rewrite H0.
    f_equal. f_equal.
    repeat (rewrite map_map).
    rewrite -> lemm_fun_shift_tyshift.
    reflexivity.

    repeat (rewrite map_map).
    rewrite -> lemm_fun_shift_tyshift3.
    assert (map (fun x => tyshift (shift (shift x 0) 0) n) l =
            map (fun x => tyshift x n) (map (fun x => shift (shift x 0) 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H2.
    rewrite -> H1.
    reflexivity.

    Case "v = tab".
    f_equal.
    unfold tyshift_subst.
    simpl.
    rewrite H.
    repeat (rewrite map_map).
    rewrite <- lemm_fun_tyshift_swap; try reflexivity; omega.

    Case "T = fun".
    f_equal.
    repeat (rewrite map_map).
    rewrite lemm_fun_shift_tyshift.
    assert (map (fun x => tyshift (shift x 0) n) l =
            map (fun x => tyshift x n) (map (fun x => shift x 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H1.
    rewrite H0.
    reflexivity.

    Case "T = forall".
    f_equal.
    unfold tyshift_subst.
    simpl.
    rewrite map_map.
    rewrite lemm_fun_tyshift_swap; try omega.
    assert (map (fun x => tyshift (tyshift x 0) (S n)) l =
            map (fun x => tyshift x (S n)) (map (fun x => tyshift x 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H.
    rewrite H0.
    reflexivity.

    Case "T = refine".
    f_equal.
    repeat (rewrite map_map).
    rewrite lemm_fun_shift_tyshift.
    assert (map (fun x => tyshift (shift x 0) n) l =
            map (fun x => tyshift x n) (map (fun x => shift x 0) l)).
      symmetry.
      apply map_map.
    rewrite -> H1.
    rewrite H0.
    reflexivity.
  Qed.

  Lemma lemm_subst_compose_aux1 :
    forall x, x - 1 - 2 = x - 3.
  Proof.
    intro. omega.
  Qed.

  Lemma lemm_subst_compose_aux2 :
    forall x, x - 2 - 1 = x - 3.
  Proof.
    intro. omega.
  Qed.

  Lemma lemm_subst_compose :
    (forall e m n e1 e2 e3,
        m <= n ->
        subst (subst e (m, [e1; e2])) (n, [e3]) =
        subst (subst e (S (S n), [shift (shift e3 m) (S m)]))
              (m, [subst e1 (n, [e3]); subst e2 (n, [e3])])) /\
    (forall v m n e1 e2 e3,
        m <= n ->
        subst_val (subst_val v (m, [e1; e2])) (n, [e3]) =
        subst_val (subst_val v (S (S n), [shift (shift e3 m) (S m)]))
              (m, [subst e1 (n, [e3]); subst e2 (n, [e3])])) /\
    (forall T m n e1 e2 e3,
        m <= n ->
        subst_ty (subst_ty T (m, [e1; e2])) (n, [e3]) =
        subst_ty (subst_ty T (S (S n), [shift (shift e3 m) (S m)]))
              (m, [subst e1 (n, [e3]); subst e2 (n, [e3])])).
  Proof.
    destruct lemm_tyshift_subst as [ H_tyshift_subst H_tyshift_subst' ].
    destruct H_tyshift_subst' as [ H_tyshift_subst_val H_tyshift_subst_ty ].

    destruct lemm_shift_tyshift as [ H_shift_tyshift H_shift_tyshift' ].
    destruct H_shift_tyshift' as [ H_shift_tyshift_val H_shift_tyshift_ty ].

    assert (forall l m n, l - m - n = l - n - m) as H_l_mn. intros. omega.

    apply term_val_ty_mind; intros; simpl;
      try (rewrite <- H);
      try (rewrite <- H0); try omega;
      try (rewrite <- H1); try omega;
      try (rewrite <- H2); try omega;
      try reflexivity.

    Case "e = x".
    destruct (lt_dec v (S (S n))).
      SCase "v < S (S n)".
      destruct (lt_dec v m).
        SSCase "v < m".
        simpl.
        destruct (lt_dec v n); try omega.
        destruct (lt_dec v m); try omega.
        reflexivity.
        SSCase "v >= m".
        remember (v - m) as vm.
        destruct vm.
          SSSCase "v - m = 0".
          assert (m = v) as H_vm.  omega.
          rewrite <- H_vm.
          simpl.
          destruct (lt_dec m m); try omega.
          assert (m - m = 0). omega.
          rewrite H0.
          reflexivity.

          destruct vm.
          SSSCase "v - m = 1".
          assert (S m = v) as H_vm. omega.
          rewrite <- H_vm.
          simpl.
          destruct (lt_dec (S m) m); try omega.
          assert (match m with
                  | 0 => S m
                  | S l0 => m - l0
                  end = 1) as H_m1.
            destruct m; omega.
          rewrite H_m1.
          reflexivity.

          SSSCase "v - m >= 2".
          assert (match vm with
                  | 0 => term_var (v - 2)
                  | S _ => term_var (v - 2)
                  end = term_var (v - 2)) as H_vm.
            destruct vm; reflexivity.
            rewrite H_vm.
            simpl.
            destruct (lt_dec (v - 2) n); try omega.
            destruct (lt_dec v m); try omega.
            rewrite <- Heqvm.
            rewrite H_vm.
            reflexivity.

      SCase "v >= S (S n)".
      destruct (lt_dec v m); try omega.
      remember (v - m) as vm.
      destruct vm; try omega.
      destruct vm; try omega.
      destruct vm.
        SSCase "v - m = 2".
        simpl.
        assert (v - 2 - n = O). omega.
        destruct (lt_dec (v - 2) n); try omega.
        rewrite -> H0.
        assert (v - S (S n) = 0). omega.
        rewrite -> H1.
        assert (shift (shift e3 m) (S m) = shift (shift e3 m) m).
          symmetry.
          apply lemm_shift_swap; omega.
        rewrite -> H2.
        symmetry.
        apply lemm_subst_shift_ident.

        SSCase "v - n >= 3".
        simpl.
        destruct (lt_dec (v - 2) n); try omega.
        assert (v - 2 - n = v - S (S n)). omega.
        rewrite H0.
        destruct (v - S (S n)).
          SSSCase "v - S (S n) = 0".
          assert (shift (shift e3 m) (S m) = shift (shift e3 m) m).
            symmetry.
            apply lemm_shift_swap; omega.
          rewrite -> H1.
          symmetry.
          apply lemm_subst_shift_ident.

          SSSCase "v - S (S n) > 0".
          assert (match n3 with
                  | 0 => term_var (v - 2 - 1)
                  | S _ => term_var (v - 2 - 1)
                  end = term_var (v - 2 - 1)).
            destruct n3; reflexivity.
          rewrite H1.
          assert (match n3 with
                  | 0 => term_var (v - 1)
                  | S _ => term_var (v - 1)
                  end = term_var (v - 1)).
            destruct n3; reflexivity.
          rewrite H2.
          simpl.
          destruct (lt_dec (v - 1) m); try omega.
          remember (v - 1 - m) as v1m.
          destruct v1m; try omega.
          destruct v1m; try omega.
          destruct v1m; f_equal; apply H_l_mn.

    Case "e = wcheck".
    f_equal.
    unfold shift_subst.
    simpl.
    rewrite -> H0; try omega.
    destruct lemm_shift_swap.
    destruct lemm_subst_shift_eq_shift_subst.
    f_equal.
    f_equal.
    f_equal.
    f_equal.
    assert (shift (shift e3 0) (S m) = shift (shift e3 m) 0).
      rewrite <- H3; try reflexivity; omega.
    rewrite -> H7.
    rewrite <- H3; try reflexivity; omega.

    f_equal.
    f_equal.
    rewrite <- H5; try reflexivity; omega.

    f_equal.
    rewrite <- H5; try reflexivity; omega.

    Case "e = acheck".
    f_equal.
    unfold shift_subst.
    simpl.
    rewrite -> H0; try omega.
    destruct lemm_shift_swap.
    destruct lemm_subst_shift_eq_shift_subst.
    f_equal.
    f_equal.
    f_equal.
    f_equal.
    assert (shift (shift e3 0) (S m) = shift (shift e3 m) 0).
      rewrite <- H4; try reflexivity; omega.
    rewrite -> H8.
    rewrite <- H4; try reflexivity; omega.

    f_equal.
    f_equal.
    rewrite <- H6; try reflexivity; omega.

    f_equal.
    rewrite <- H6; try reflexivity; omega.

    Case "v = fix".
    unfold shift_subst.
    simpl.
    rewrite -> H0; try omega.
    rewrite -> H1; try omega.
    destruct lemm_shift_swap.
    destruct lemm_subst_shift_eq_shift_subst.
    f_equal.
    f_equal.
    f_equal.
    f_equal.
    f_equal.
    assert (shift (shift e3 0) (S m) = shift (shift e3 m) 0).
      rewrite <- H3; try reflexivity; omega.
    rewrite -> H7.
    rewrite <- H3; try reflexivity; omega.

    f_equal.
    f_equal.
    f_equal.
    rewrite <- H5; try reflexivity; omega.

    f_equal.
    rewrite <- H5; try reflexivity; omega.

    f_equal.
    f_equal.
    f_equal.
    f_equal.
    assert (shift (shift (shift e3 0) 0) (S (S m)) =
            shift (shift (shift e3 m) 0) 0).
      rewrite <- H3; try omega.
      rewrite <- H3; try omega.
      reflexivity.
    rewrite -> H7.
    rewrite <- H3; try omega.
    rewrite <- H3; try omega.
    reflexivity.

    f_equal.
    f_equal.
    f_equal.
    rewrite <- H5; try omega.
    rewrite <- H5; try omega.
    reflexivity.

    f_equal.
    rewrite <- H5; try omega.
    rewrite <- H5; try omega.
    reflexivity.

    Case "T = fun".
    unfold shift_subst.
    simpl.
    f_equal.
    unfold tyshift_subst.
    simpl.
    rewrite -> H; try omega.
    destruct lemm_shift_swap.
    destruct lemm_subst_shift_eq_shift_subst.
    f_equal.
    f_equal.
    f_equal.
    f_equal.
    assert (shift (shift e3 0) (S m) = shift (shift e3 m) 0).
      rewrite <- H1; try reflexivity; omega.
    destruct lemm_shift_tyshift.
    repeat (rewrite -> H6).
    reflexivity.

    f_equal.
    f_equal.
    destruct lemm_subst_shift_eq_shift_subst.
    rewrite -> H_tyshift_subst.
    reflexivity.

    f_equal.
    rewrite -> H_tyshift_subst.
    reflexivity.

    Case "T = fun".
    f_equal.
    unfold shift_subst.
    simpl.
    rewrite -> H0; try omega.
    f_equal.
    f_equal.
    f_equal.
    f_equal.
    assert (shift (shift e3 0) (S m) = shift (shift e3 m) 0).
      symmetry.
      apply lemm_shift_swap; omega.
    rewrite -> H2; try omega.
    symmetry.
    apply lemm_shift_swap; omega.


    f_equal. f_equal.
    destruct lemm_subst_shift_eq_shift_subst.
    rewrite <- H2; try reflexivity; omega.

    f_equal.
    destruct lemm_subst_shift_eq_shift_subst.
    rewrite <- H2; try reflexivity; omega.

    Case "T = forall".
    f_equal.
    unfold tyshift_subst. simpl.
    rewrite -> H; try omega.
    f_equal. f_equal. f_equal. f_equal.
    rewrite H_shift_tyshift.
    apply H_shift_tyshift.

    f_equal. f_equal.
    rewrite -> H_tyshift_subst.
    reflexivity.

    f_equal.
    rewrite -> H_tyshift_subst.
    reflexivity.

    Case "T = refine".
    unfold shift_subst.
    simpl.
    rewrite -> H0; try omega.
    destruct lemm_shift_swap.
    destruct lemm_subst_shift_eq_shift_subst.
    f_equal.

    f_equal. f_equal. f_equal. f_equal.
    assert (shift (shift e3 0) (S m) = shift (shift e3 m) 0).
      rewrite <- H2; try reflexivity; omega.
    rewrite -> H6.
    rewrite <- H2; try reflexivity; omega.

    f_equal. f_equal. f_equal. f_equal.
    rewrite <- H4; try reflexivity; omega.

    f_equal.
    rewrite <- H4; try reflexivity; omega.
  Qed.

  Lemma lemm_subst_tysubst :
    (forall e' m n T e,
        subst (tysubst e' (m, [T])) (n, [e]) =
        tysubst (subst e' (n, [tyshift e m]))
                (m, [subst_ty T (n, [e])])) /\
    (forall v' m n T e,
        subst_val (tysubst_val v' (m, [T])) (n, [e]) =
        tysubst_val (subst_val v' (n, [tyshift e m]))
                    (m, [subst_ty T (n, [e])])) /\
    (forall T' m n T e,
        subst_ty (tysubst_ty T' (m, [T])) (n, [e]) =
        tysubst_ty (subst_ty T' (n, [tyshift e m]))
                    (m, [subst_ty T (n, [e])])).
  Proof.
    destruct lemm_tyshift_subst as [ H_tyshift_subst H_tyshift_subst' ].
    destruct H_tyshift_subst' as [ H_tyshift_subst_val H_tyshift_subst_ty ].

    destruct lemm_shift_tyshift as [ H_shift_tyshift H_shift_tyshift' ].
    destruct H_shift_tyshift' as [ H_shift_tyshift_val H_shift_tyshift_ty ].

    destruct lemm_subst_shift_eq_shift_subst as [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].

    destruct lemm_tyshift_swap as [ H_tyshift_swap H_tyshift_swap' ].
    destruct H_tyshift_swap' as [ H_tyshift_swap_val H_tyshift_swap_ty ].

    apply term_val_ty_mind; intros; simpl;
      repeat (unfold shift_subst);
      repeat (unfold shift_tysubst);
      repeat (unfold tyshift_subst);
      repeat (unfold tyshift_tysubst);
      simpl;
      try (rewrite -> H_shift_tyshift);
      try (rewrite <- H_ss_eq_ss_ty; try omega);
      try (rewrite <- H);
      try (rewrite <- H0);
      try (rewrite <- H1);
      try (rewrite <- H2);
      try (rewrite -> H);
      try (rewrite -> H0);
      try (rewrite -> H1);
      try (rewrite <- H_tyshift_swap; try omega);
      try (rewrite -> H_tyshift_subst_ty);
      simpl;
      try reflexivity.

    Case "e = x".
    destruct (lt_dec v n).
      SCase "v < n".
      reflexivity.
      SSCase "v >= n".
      remember (v - n) as vn.
      destruct vn.
        SSSCase "v - n = 0".
        assert (n = v). omega.
        symmetry.
        apply lemm_tysubst_tyshift_ident.

        SSSCase "v - n = S vn".
        destruct vn; reflexivity.

    Case "v = fix".
    f_equal. f_equal. f_equal. f_equal. f_equal.
    repeat (rewrite <- H_shift_tyshift).
    reflexivity.

    f_equal. f_equal.
    rewrite <- H_ss_eq_ss_ty; try omega.
    reflexivity.

    Case "T = A".
    destruct (lt_dec t m); try reflexivity.
    remember (t - m) as tm.
    repeat (destruct tm; try reflexivity).
  Qed.

  Lemma lemm_coterm_app_aux :
    forall e1 e2 v1 v2 e,
      (substO (!v1 @ !v2) [e1] ~~> e) ->
      exists e', (e = substO e' [e1]) /\
                 (substO (!v1 @ !v2) [e2] ~~> substO e' [e2]).
  Proof.
    destruct lemm_subst_compose as [ H_subst_comp H_subst_comp' ].
    destruct lemm_shift_swap as [ H_shift_swap H_shift_swap' ].
    destruct H_shift_swap' as [ H_shift_swap_val H_shift_swap_ty ].
    destruct lemm_subst_shift_eq_shift_subst as
        [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].
    destruct lemm_tyshift_subst as [ H_tyshift_subst H_tyshift_subst' ].
    destruct H_tyshift_subst' as [ H_tyshift_subst_val H_tyshift_subst_ty ].

    intros e1 e2 v1 v2 e H_red_inv.
    unfold substO in H_subst_comp.
    simpl in H_subst_comp.

    inversion H_red_inv.
    Case "beta".
    destruct v1; try discriminate.
    unfold substO.
    simpl in H2.
    injection H0; intros.
    unfold fix_v.
    exists (subst t1 (O, [!v2; !(val_fix t t0 t1)])).
    apply conj.
      SCase "e = e'[e1/x]".
      rewrite -> H_subst_comp; try omega.
      f_equal.
      rewrite -> H.
      f_equal.
      unfold shift_subst.
      simpl.
      rewrite <- H_shift_swap; try reflexivity; omega.

      f_equal.
      f_equal.
      f_equal.
      simpl.
      simpl in H0.
      injection H0; intros.
      rewrite H, H3, H4.
      reflexivity.

      SCase "~~>".
      simpl.
      assert (! (val_fix (subst_ty t (0, [e2])) (subst_ty t0 (shift_subst (0, [e2])))
                         (subst t1 (shift_subst (shift_subst (0, [e2]))))) @
                ! (subst_val v2 (0, [e2]))
                ~~> (substO
                       (subst t1 (shift_subst (shift_subst (0, [e2]))))
                       [! (subst_val v2 (0, [e2]));
                        ! (val_fix (subst_ty t (0, [e2])) (subst_ty t0 (shift_subst (0, [e2])))
                         (subst t1 (shift_subst (shift_subst (0, [e2])))))])).
        apply red_beta.
      assert (substO (subst t1 (shift_subst (shift_subst (0, [e2]))))
                     [!(subst_val v2 (0, [e2]));
                      !(val_fix (subst_ty t (0, [e2]))
                                (subst_ty t0 (shift_subst (0, [e2])))
                                (subst t1
                                       (shift_subst (shift_subst (0, [e2])))))] =
              subst (subst t1 (0, [! (v2); ! (val_fix t t0 t1)])) (0, [e2])).
        unfold substO.
        unfold shift_subst.
        simpl.
        rewrite -> H_subst_comp; try omega.
        f_equal.
        f_equal.
        f_equal.
        f_equal.
        rewrite -> H_shift_swap; try reflexivity; omega.
      rewrite <- H6.
      apply H5.

    Case "cast_base".
    destruct v1; try discriminate.
    injection H0; intros.
    destruct t; try discriminate.
    destruct t0; try discriminate.
    injection H3; intros.
    injection H4; intros.
    rewrite <- H5.
    rewrite <- H6.
    exists (!v2).
    apply conj.
      SCase "e = e'[e1/x]".
      reflexivity.
      SCase "~~>".
      unfold substO.
      assert (subst (! (val_cast (ty_base B) (ty_base B) l0) @ ! (v2)) (0, [e2]) =
              !(val_cast (ty_base B) (ty_base B) l0) @ !(subst_val v2 (0, [e2]))).
        reflexivity.
      rewrite -> H7.
      simpl.
      apply red_cast_base.

    Case "cast_fun".
    destruct v1; try discriminate.
    injection H0; intros.
    destruct t0; try discriminate.
    destruct t; try discriminate.
    simpl in H3.
    inversion H3.
    simpl in H4.
    inversion H4.
    exists (!(val_fix
                t1 t2
                (!(val_fix (shift_ty_n t0_1 2)
                           (shift_ty (shift_ty t2 1) 0)
                           (!(val_cast (shift_ty (shift_ty_n t2 2) 3)
                                       (shift_ty (shift_ty (shift_ty t0_2 1) 1) 1)
                                       l)
                             @ ((shift_n !v2 4) @ (term_var 0))))
                  @ (!(val_cast (shift_ty_n t0_1 2) (shift_ty_n t1 2) l) @
                      (term_var 0))))).
    split.
      SCase "e = e'[e1/x]".
      unfold domain_cast. unfold codomain_cast.
      unfold T11'. unfold T12'. unfold T21'. unfold T22'. unfold T12''.
      rewrite -> H6. rewrite -> H9.
      unfold shift_ty_n. unfold shift_n. unfold shift_subst. unfold substO.
      simpl.
      repeat f_equal.

      repeat (rewrite <- H_ss_eq_ss_ty; try omega; simpl).
      reflexivity.

      repeat (rewrite <- H_ss_eq_ss_ty; try omega; simpl).
      unfold shift_subst; simpl.
      f_equal. f_equal. f_equal. f_equal.
      rewrite <- H_shift_swap; try reflexivity; omega.

      repeat (rewrite <- H_ss_eq_ss_ty; try omega; simpl).
      f_equal.
      f_equal.
      f_equal.
      repeat (rewrite <- H_shift_swap; try omega).
      reflexivity.

      rewrite -> H7.
      unfold shift_subst. simpl.
      repeat (rewrite <- H_ss_eq_ss_ty; try omega; simpl).
      f_equal. f_equal. f_equal.
      assert (shift (shift e1 0) 1 = shift (shift e1 0) 0).
        rewrite <- H_shift_swap; try reflexivity; omega.
      rewrite -> H5.
      assert (shift (shift (shift e1 0) 0) 1 = shift (shift (shift e1 0) 0) 0).
        rewrite <- H_shift_swap; try reflexivity; omega.
      rewrite -> H10.
      rewrite <- H_shift_swap; try reflexivity; omega.

      unfold v'.
      unfold shift_n.
      simpl.
      f_equal.
      repeat (rewrite <- H_ss_eq_ss_val; try omega; simpl).
      reflexivity.

      repeat (rewrite <- H_ss_eq_ss_ty; try omega; simpl).
      reflexivity.

      rewrite -> H8.
      repeat (rewrite <- H_ss_eq_ss_ty; try omega; simpl).
      reflexivity.

      SSCase "~~>".
      unfold substO.
      simpl.
      unfold shift_subst.
      simpl.
      assert (subst_ty (shift_ty_n t0_1 2) (2, [shift (shift e2 0) 0]) =
              shift_ty_n (subst_ty t0_1 (0, [e2])) 2).
        unfold shift_ty_n.
        simpl.
        repeat (rewrite <- H_ss_eq_ss_ty; try omega).
        reflexivity.
      rewrite -> H5.
      assert (subst_ty (shift_ty (shift_ty t2 1) 0)
                       (3, [shift (shift (shift e2 0) 0) 0]) =
              shift_ty (shift_ty (subst_ty t2 (1, [shift e2 0])) 1) 0).
        repeat (rewrite <- H_ss_eq_ss_ty; try omega).
        simpl.
        assert (shift (shift e2 0) 0 = shift (shift e2 0) 1).
          apply H_shift_swap; omega.
        rewrite -> H10.
        reflexivity.
      rewrite -> H10.

      assert (shift_ty (shift_ty_n (subst_ty t2 (1, [shift e2 0])) 2) 3 =
              subst_ty (shift_ty (shift_ty_n t2 2) 3)
                       (4, [shift (shift (shift (shift e2 0) 0) 0) 0])).
        unfold shift_ty_n.
        simpl.
        repeat (rewrite <- H_ss_eq_ss_ty; try omega).
        simpl. f_equal. f_equal. f_equal.
        repeat (rewrite <- H_shift_swap; try omega).
        reflexivity.
      rewrite <- H11.
      assert (shift_ty
                (shift_ty (shift_ty (subst_ty t0_2 (1, [shift e2 0])) 1) 1) 1 =
              subst_ty (shift_ty (shift_ty (shift_ty t0_2 1) 1) 1)
                       (4, [shift (shift (shift (shift e2 0) 0) 0) 0])).
        repeat (rewrite <- H_ss_eq_ss_ty; try omega).
        simpl. f_equal. f_equal. f_equal.
        assert (shift (shift e2 0) 1 = shift (shift e2 0) 0).
          rewrite <- H_shift_swap; try reflexivity; omega.
        rewrite -> H12.
        assert (shift (shift (shift e2 0) 0) 1 = shift (shift (shift e2 0) 0) 0).
          rewrite <- H_shift_swap; try reflexivity; omega.
        rewrite -> H13.
        rewrite <- H_shift_swap; try reflexivity; omega.
      rewrite <- H12.
      assert (shift_n !(subst_val v2 (0, [e2])) 4 =
              !(subst_val
                  (shift_val (shift_val (shift_val (shift_val v2 0) 0) 0) 0)
                  (4, [shift (shift (shift (shift e2 0) 0) 0) 0]))).
        unfold shift_n.
        simpl.
        f_equal.
        repeat (rewrite <- H_ss_eq_ss_val; try omega).
        reflexivity.
      rewrite <- H13.
      assert (shift_ty_n (subst_ty t1 (0, [e2])) 2 =
              subst_ty (shift_ty_n t1 2) (2, [shift (shift e2 0) 0])).
        unfold shift_ty_n.
        simpl.
        repeat (rewrite <- H_ss_eq_ss_ty; try omega).
        reflexivity.
      rewrite <- H14.
      rewrite -> H.
      apply red_cast_fun.

    Case "cast_forall".
    destruct v1; try discriminate.
    injection H0; intros.
    destruct t0; try discriminate.
    destruct t; try discriminate.
    simpl in H3.
    inversion H3.
    simpl in H4.
    inversion H4.
    exists (!(val_tabs ((!(val_cast t t0 l0))
                          @ ((tyshift !v2 0) @@ (ty_var 0))))).
    split.
      SCase "e = e'[e1/x]".
      unfold cast. unfold body.
      unfold substO. rewrite -> H6. rewrite -> H7.
      unfold tyshift_subst. simpl. unfold tyshift_subst. simpl.
      f_equal. f_equal. f_equal. f_equal. f_equal.
      unfold tyshift_subst. simpl.
      assumption.

      f_equal. f_equal.
      rewrite -> H_tyshift_subst_val.
      reflexivity.

      SCase "~~>".
      simpl.
      unfold substO.
      repeat (unfold tyshift_subst; simpl).
      assert (tyshift !(subst_val v2 (0, [e2])) 0 =
              !(subst_val (tyshift_val v2 0) (0, [tyshift e2 0]))).
        assert (!(subst_val v2 (0, [e2])) = subst !v2 (0, [e2])).
          reflexivity.
        rewrite -> H5.
        rewrite -> H_tyshift_subst.
        reflexivity.
      rewrite <- H5.
      apply red_cast_forall.

    Case "cast_forget".
    destruct v1; try discriminate.
    injection H0; intros.
    destruct t0; try discriminate.
    simpl in H3.
    inversion H3.
    exists (!(val_cast t t0 l0) @ !v2).
    split.
      SCase "e = e'[e1/x]".
      unfold substO.
      simpl. f_equal.

      f_equal.
      rewrite H4. rewrite H. reflexivity.

      SCase "~~>".
      unfold substO. simpl.
      apply red_forget.

    Case "cast_precheck".
    destruct v1; try discriminate.
    injection H; intros.
    destruct t; try discriminate.
    simpl in H5.
    inversion H5.
    exists (term_wcheck t t1 (!(val_cast t t0 l0) @ !v2) l0).
    unfold substO. unfold shift_subst. simpl.
    split.
      SCase "e = e'[e1/x]".
      f_equal. f_equal. f_equal. f_equal.
      assumption. assumption. assumption.

      SCase "~~>".
      apply red_precheck.
      unfold not.
      intros.
      destruct t0; try discriminate.
      simpl in H4.
      unfold not in H2.
      apply H2 with (T := subst_ty t0 (0, [e1]))
                    (e := subst t2 (shift_subst (0, [e1]))).
      assumption.
  Qed.

  Lemma lemm_coterm_tapp_aux :
    forall e1 e2 v T e,
      (substO (!v @@ T) [e1] ~~> e) ->
      exists e', (e = substO e' [e1]) /\
                 (substO (!v @@ T) [e2] ~~> substO e' [e2]).
  Proof.
    destruct lemm_subst_tysubst as [ H_subst_tysubst H_subst_tysubst' ].
    destruct H_subst_tysubst' as [ H_subst_tysubst_val H_subst_tysubst_ty ].

    intros e1 e2 v T e H_red_inv.
    inversion H_red_inv.
    destruct v; try discriminate.
    inversion H0.
    exists (tysubst t (O, [T])).
    split.
      Case "e = e'[e1/x]".
      unfold tyshift_subst. unfold substO. unfold tysubstO.
      rewrite -> H_subst_tysubst.
      reflexivity.

      Case "~~>".
      unfold substO. simpl.
      unfold tyshift_subst. simpl.
      rewrite -> H_subst_tysubst.
      apply red_tbeta.
  Qed.

  Lemma lemm_subst_dummy :
    (forall e n e1 e2,
      subst e (n, [e1]) = subst (shift e (1 + n)) (n, [e1; e2])) /\
    (forall v n e1 e2,
        subst_val v (n, [e1]) = subst_val (shift_val v (1 + n)) (n, [e1; e2])) /\
    (forall T n e1 e2,
        subst_ty T (n, [e1]) = subst_ty (shift_ty T (1 + n)) (n, [e1; e2])).
  Proof.
    apply term_val_ty_mind; intros;
      repeat (unfold shift_subst; unfold tyshift_subst; simpl);
      try (rewrite <- H);
      try (rewrite <- H0);
      try (rewrite <- H1);
      try (rewrite <- H2);
      try reflexivity.

    Case "e = x".
    simpl.
    destruct (lt_dec v n).
      SCase "v < n".
      destruct v.
        SSCase "v = 0".
        destruct (lt_dec 0 n); try reflexivity; omega.
        SSCase "v = S _".
        destruct (le_gt_dec n v); try omega.
        simpl.
        destruct (lt_dec (S v) n); try reflexivity; omega.
      SCase "v >= n".
      destruct v.
        SSCase "v = 0".
        destruct (lt_dec 0 n); try omega.
        remember (0 - n) as nz.
        destruct nz; try reflexivity; omega.
        SSCase "v = S _".
        destruct (le_gt_dec n v).
          SSSCase "n <= v".
          simpl.
          destruct (lt_dec (S (S v)) n); try omega.
          destruct n; try reflexivity.
          remember (v - n) as vn.
          destruct vn; try omega.
          destruct n.
          destruct v; try omega.
          destruct v; destruct vn; reflexivity.
          remember (v - n) as vn'.
          destruct vn'; try omega.
          destruct vn.
            destruct vn'; try omega.
            destruct vn'; reflexivity.
          destruct vn'; try omega.
          destruct vn'; reflexivity.

          SSSCase "n > v".
          simpl.
          destruct (lt_dec (S v) n).
            SSSSCase "S v < n".
            destruct n; try omega.
            SSSSCase "S v >= n".
            destruct n; try omega.
            remember (v - n) as vn.
            destruct vn; try reflexivity.
            destruct vn; omega.
  Qed.

  Lemma lemm_coterm_wcheck_aux :
    forall e1 e2 T' e' v' l e,
      (substO (term_wcheck T' e' !v' l) [e1] ~~> e) ->
      exists e'', (e = substO e'' [e1]) /\
                 (substO (term_wcheck T' e' !v' l) [e2] ~~> substO e'' [e2]).
  Proof.
    destruct lemm_subst_compose as [ H_subst_comp H_subst_comp' ].
    destruct H_subst_comp' as [ H_subst_comp_val H_subst_comp_ty ].

    destruct lemm_subst_shift_eq_shift_subst as [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].

    destruct lemm_subst_dummy as [ H_subst_dummy H_subst_dummy' ].

    intros e1 e2 T' e' v' l e H_red_inv.
    inversion H_red_inv.
    exists (term_acheck T' e' (substO e' [!v']) v' l).
    split.
      Case "e = e'[e1/x]".
      unfold shift_subst. unfold substO.
      simpl.
      f_equal.
      assert (subst e' (0, [!v']) =
              subst (shift e' (S O)) (0, [!v'; term_var 0])).
        apply H_subst_dummy.
      rewrite H.
      rewrite -> H_subst_comp; try omega.
      assert (subst (shift e' 1) (2, [shift (shift e1 0) 1]) =
              shift (subst e' (1, [shift e1 0])) 1).
        assert ([shift (shift e1 0) 1] =
                map (fun x => shift x 1) [shift e1 0]).
          reflexivity.
        rewrite -> H5.
        rewrite H_ss_eq_ss; try reflexivity; omega.
      rewrite H5.
      rewrite <- H_subst_dummy.
      reflexivity.

      Case "~~>".
      repeat (unfold subst0; unfold shift_subst; simpl).
      cut (subst (subst e' (0, [!v'])) (0, [e2]) =
           subst (subst e' (1, [shift e2 0])) (O, [!(subst_val v' (0, [e2]))])).
        intro.
        unfold substO. simpl.
        rewrite H.
        apply red_check.
        assert (subst e' (0, [!v']) =
                subst (shift e' 1) (0, [!v'; term_var 0])).
          apply H_subst_dummy.
          rewrite H.
          rewrite H_subst_comp; try omega.
          assert ([shift (shift e2 0) 1] =
                  map (fun x => shift x 1) [shift e2 0]).
            reflexivity.
          rewrite H5.
          rewrite H_ss_eq_ss; try omega.
          rewrite <- H_subst_dummy.
          reflexivity.
  Qed.

  Lemma lemm_coterm_inv_bool :
    forall v v' n l,
      subst_val v (n, l) = v' ->
      (v' = tt \/ v' = ff) ->
      v = v'.
  Proof.
    intros.
    inversion H0;
      destruct v;
      try (rewrite H1 in H; discriminate);
      inversion H; reflexivity.
  Qed.

  Lemma lemm_coterm_acheck_aux :
    forall e1 e2 T' e' v1' v2' l e,
      (substO (term_acheck T' e' !v1' v2' l) [e1] ~~> e) ->
      exists e'', (e = substO e'' [e1]) /\
                 (substO (term_acheck T' e' !v1' v2' l) [e2] ~~> substO e'' [e2]).
  Proof.
    intros e1 e2 T' e' v1' v2' l e H_red_inv.
    inversion H_red_inv.
      Case "v1' = tt".
      exists (!v2').
      split.
        SCase "e = e'[e1/x]".
        reflexivity.
        SCase "~~>".
        repeat (unfold substO; unfold shift_subst; simpl).
        assert (v1' = tt).
          apply lemm_coterm_inv_bool with (n := 0) (l := [e1]).
          symmetry. assumption.
          apply or_introl. reflexivity.
        rewrite H.
        apply red_ok.
      Case "v1' = ff".
      exists (term_blame l).
      split.
        SCase "e = e'[e1/x]".
        reflexivity.
        SCase "~~>".
        repeat (unfold substO; unfold shift_subst; simpl).
        assert (v1' = ff).
          apply lemm_coterm_inv_bool with (n := 0) (l := [e1]).
          symmetry. assumption.
          apply or_intror. reflexivity.
        rewrite H.
        apply red_fail.
  Qed.

  Lemma lemm_ectx_hole_ident :
    forall x, x = ectx_hole @[x].
  Proof.
    reflexivity.
  Qed.

  Lemma lemm_left_red_var_stuck :
    forall H e1 e2 E x e,
      (e1 --> e2) ->
      (E <> ectx_hole) ->
      (substO (E @[term_var x]) [e1] ~~> e) ->
      H.
  Proof.
    intros.
    exfalso.
    unfold substO in H2.
    destruct E; simpl in H2.
      Case "E = []".
      unfold not in H1.
      apply H1.
      reflexivity.
      Case "E = app1".
      inversion H2; destruct E; try discriminate; simpl in H4.
        SCase "e1 = fix".
        destruct (x - 0).
          rewrite <- H4 in H0.
          apply (lemm_eval_val_stuck (val_fix T1 T2 e0) e2). assumption.
          destruct n; discriminate.
        SCase "e1 = cast B".
        destruct (x - 0).
          rewrite <- H4 in H0.
          apply (lemm_eval_val_stuck (val_cast (ty_base B) (ty_base B) l) e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast fun".
        destruct (x - 0).
          rewrite <- H4 in H0.
          apply (lemm_eval_val_stuck
                   (val_cast (ty_fun T11 T12) (ty_fun T21 T22) l) e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast forall".
        destruct (x - 0).
          rewrite <- H4 in H0.
          apply (lemm_eval_val_stuck
                   (val_cast (ty_forall T1) (ty_forall T2) l) e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast T refine".
        destruct (x - 0).
          rewrite <- H4 in H0.
          apply (lemm_eval_val_stuck (val_cast T1 (ty_refine T2 e0) l) e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast refine T".
        simpl in H3.
        destruct (x - 0).
          rewrite <- H3 in H0.
          apply (lemm_eval_val_stuck (val_cast (ty_refine T1 e0) T2 l) e2).
            assumption.
          destruct n; discriminate.
      Case "E = app2".
      inversion H2; destruct E; try discriminate; simpl in H5.
        SCase "e1 = fix".
        destruct (x - 0).
          rewrite <- H5 in H0.
          apply (lemm_eval_val_stuck v2 e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast B".
        destruct (x - 0).
          rewrite <- H5 in H0.
          apply (lemm_eval_val_stuck v0 e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast fun".
        destruct (x - 0).
          rewrite <- H5 in H0.
          apply (lemm_eval_val_stuck v0 e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast forall".
        destruct (x - 0).
          rewrite <- H5 in H0.
          apply (lemm_eval_val_stuck v0 e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast T refine".
        destruct (x - 0).
          rewrite <- H5 in H0.
          apply (lemm_eval_val_stuck v0 e2).
            assumption.
          destruct n; discriminate.
        SCase "e1 = cast refine T".
        simpl in H3.
        destruct (x - 0).
          rewrite <- H5 in H0.
          apply (lemm_eval_val_stuck v0 e2).
            assumption.
          destruct n; discriminate.

      Case "E = tapp".
      inversion H2; destruct E; try discriminate; simpl in H4.
      destruct (x - 0).
        rewrite <- H4 in H0.
        apply (lemm_eval_val_stuck (val_tabs e0) e2). assumption.
        destruct n; discriminate.

      Case "E = wcheck".
      inversion H2; destruct E; try discriminate.
      simpl in H6.
      destruct (x - 0).
        rewrite <- H6 in H0.
        apply (lemm_eval_val_stuck v e2). assumption.
        destruct n; discriminate.

      Case "E = acheck".
      inversion H2; destruct E; try discriminate; simpl in H6.
        SCase "tt".
        destruct (x - 0).
          rewrite <- H6 in H0.
          apply (lemm_eval_val_stuck tt e2). assumption.
          destruct n; discriminate.
        SCase "ff".
        destruct (x - 0).
          rewrite <- H6 in H0.
          apply (lemm_eval_val_stuck ff e2). assumption.
          destruct n; discriminate.
  Qed.

  Lemma lemm_left_red_ectx :
    forall e1 e2 e e' E,
      (e1 --> e2) ->
      (E <> ectx_hole) ->
      (substO (E@[e]) [e1] ~~> e') ->
      exists v, e = !v.
  Proof.
    intros.
    destruct e.
    Case "e = x".
    apply lemm_left_red_var_stuck with (e1 := e1) (e2 := e2)
                                       (E := E)
                                       (x := v) (e := e');
      try assumption.

    Case "e = v".
    exists v. reflexivity.

    destruct E;
      try (exfalso; unfold not in H0; apply H0; reflexivity);
      try (inversion H1; destruct E; try discriminate).
    destruct E;
      try (exfalso; unfold not in H0; apply H0; reflexivity);
      try (inversion H1; destruct E; try discriminate).
    destruct E;
      try (exfalso; unfold not in H0; apply H0; reflexivity);
      try (inversion H1; destruct E; try discriminate).
    destruct E;
      try (exfalso; unfold not in H0; apply H0; reflexivity);
      try (inversion H1; destruct E; try discriminate).
    destruct E;
      try (exfalso; unfold not in H0; apply H0; reflexivity);
      try (inversion H1; destruct E; try discriminate).
  Qed.

  Lemma lemm_red_ectx_decomp :
    forall E e e1 e2 E' e1' e2',
      e = E@[e1] -> (e1 ~~> e2) ->
      e = E'@[e1'] -> (e1' ~~> e2') ->
      E = E' /\ e1 = e1'.
  Proof.
    intros E.
    induction E; intros.
      Case "E = []".
      simpl in H.
      rewrite H1 in H.
      rewrite <- H in H0.
      assert (E' = ectx_hole).
        destruct E';
          try reflexivity;
          try simpl in H0;
          try simpl in H1;
          inversion H0;
          try (unfold fix_v in H4);
          destruct E';
          try discriminate;
          try (simpl in H3; rewrite <- H3 in H2);
          try (simpl in H4; rewrite <- H4 in H2);
          try (simpl in H5; rewrite <- H5 in H2);
          try (simpl in H6; rewrite <- H6 in H2);
          inversion H2.
      split. symmetry. assumption. rewrite H3 in H. symmetry. assumption.

      Case "E = app1".
      simpl in H.
      rewrite H1 in H.
      destruct E'; try discriminate.
        SCase "E' = []".
        simpl in H. simpl in H1.
        rewrite H in H2.
        destruct E;
          inversion H2;
          try (unfold fix_v in H4);
          try (rewrite <- H3 in H0);
          try (rewrite <- H4 in H0);
          inversion H0.

        SCase "E' = app1".
        injection H; intros.
        destruct e; try discriminate.
        injection H1; intros.
        assert (E = E' /\ e1 = e1').
          apply IHE with (e := e3) (e2 := e2) (e2' := e2');
            try rewrite H6; try assumption; reflexivity.
        rewrite H3.
        inversion H7.
        rewrite H8. rewrite H9.
        split; reflexivity.

        SCase "E' = app2".
        simpl in H. simpl in H1.
        injection H; intros.
        destruct E; try discriminate.
        simpl in H4.
        rewrite <- H4 in H0.
        inversion H0.

      Case "E = app2".
      simpl in H.
      rewrite H1 in H.
      destruct E'; try discriminate.
        SCase "E' = []".
        simpl in H. simpl in H1.
        rewrite H in H2.
        destruct e1;
          inversion H0;
          destruct E;
          inversion H2.

        SCase "E' = app1".
        simpl in H. simpl in H1.
        injection H; intros.
        destruct E'; try discriminate.
        simpl in H4.
        rewrite H4 in H2.
        inversion H2.

        SCase "E' = app2".
        simpl in H1. simpl in H.
        destruct e; try discriminate.
        injection H; intros.
        injection H1; intros.
        assert (E = E' /\ e1 = e1').
          apply IHE with (e := e4) (e2 := e2) (e2' := e2'); try assumption.
          rewrite -> H5. assumption.
        inversion H7.
        rewrite H8. rewrite H9. rewrite H4.
        split; reflexivity.

      Case "E = tapp".
      simpl in H.
      rewrite H1 in H.
      destruct E'; try discriminate.
        SCase "E' = []".
        simpl in H. simpl in H1.
        rewrite H in H2.
        destruct E;
          inversion H2;
          try (unfold fix_v in H4);
          try (rewrite <- H3 in H0);
          try (rewrite <- H4 in H0);
          inversion H0.

        SCase "E' = tapp".
        simpl in H1. simpl in H.
        destruct e; try discriminate.
        injection H; intros.
        injection H1; intros.
        assert (E = E' /\ e1 = e1').
          apply IHE with (e := e) (e2 := e2) (e2' := e2'); try assumption.
          rewrite -> H6. assumption.
        inversion H7.
        rewrite H8. rewrite H9. rewrite H3.
        split; reflexivity.

      Case "E = wcheck".
      simpl in H.
      rewrite H1 in H.
      destruct E'; try discriminate.
        SCase "E' = []".
        simpl in H. simpl in H1.
        rewrite H in H2.
        destruct E;
          inversion H2;
          try (unfold fix_v in H4);
          try (rewrite <- H6 in H0);
          inversion H0.

        SCase "E' = whceck".
        simpl in H1. simpl in H.
        destruct e; try discriminate.
        injection H; intros.
        injection H1; intros.
        assert (E = E' /\ e1 = e1').
          apply IHE with (e := e4) (e2 := e2) (e2' := e2'); try assumption.
          rewrite -> H8. assumption.
        inversion H11.
        rewrite H12. rewrite H13. rewrite H6. rewrite H3. rewrite H5.
        split; reflexivity.

      Case "E = acheck".
      simpl in H.
      rewrite H1 in H.
      destruct E'; try discriminate.
        SCase "E' = []".
        simpl in H. simpl in H1.
        rewrite H in H2.
        destruct E;
          inversion H2;
          try (unfold fix_v in H4);
          try (rewrite <- H6 in H0);
          inversion H0.

        SCase "E' = ahceck".
        simpl in H1. simpl in H.
        destruct e; try discriminate.
        injection H; intros.
        injection H1; intros.
        assert (E = E' /\ e1 = e1').
          apply IHE with (e := e4) (e2 := e2) (e2' := e2'); try assumption.
          rewrite -> H10. assumption.
        inversion H13.
        rewrite H14. rewrite H15. rewrite H7. rewrite H6. rewrite H3. rewrite H4.
        split; reflexivity.
  Qed.

  Lemma lemm_red_deterministic :
    forall e e1 e2, (e ~~> e1) -> (e ~~> e2) -> e1 = e2.
  Proof.
    intros.
    inversion H; inversion H0;
      try rewrite <- H1 in H3;
      try rewrite <- H1 in H4;
      try rewrite <- H2 in H4;
      try rewrite <- H3 in H4;
      try discriminate.

      Case "beta".
      unfold fix_v.
      unfold fix_v0.
      unfold e3 in H3. unfold e5 in H3.
      injection H3; intros.
      rewrite H5. rewrite H6. rewrite H7. rewrite H8.
      reflexivity.

      Case "tbeta".
      unfold e' in H3. unfold e'0 in H3.
      injection H3; intros.
      rewrite H5. rewrite H6.
      reflexivity.

      Case "cast_base".
      injection H3; intros.
      rewrite H5.
      reflexivity.

      Case "cast_fun".
      unfold e3 in H3. unfold e0 in H3.
      unfold codomain_cast. unfold domain_cast.
      unfold codomain_cast0. unfold domain_cast0.
      unfold v'. unfold v'0.
      unfold T11'. unfold T12'. unfold T21'. unfold T22'. unfold T12''.
      unfold T11'0. unfold T12'0. unfold T21'0. unfold T22'0. unfold T12''0.
      injection H3; intros.
      rewrite H5. rewrite H6. rewrite H7. rewrite H8. rewrite H9. rewrite H10.
      reflexivity.

      Case "cast_forall".
      unfold e3 in H3. unfold e0 in H3.
      unfold cast. unfold body.
      unfold cast0. unfold body0.
      injection H3; intros.
      rewrite H5. rewrite H6. rewrite H7. rewrite H8.
      reflexivity.

      Case "cast_forget".
      unfold e5 in H3. unfold e3 in H3.
      injection H3; intros.
      rewrite H5. rewrite H6. rewrite H8. rewrite H9.
      reflexivity.

      Case "cast_forget x cast_precheck".
      unfold e5 in H4. unfold e3 in H4.
      injection H4; intros.
      exfalso.
      unfold not in H3.
      apply H3 with (T := T2) (e := e0).
      assumption.

      Case "cast_precheck x cast_forget".
      unfold e5 in H4. unfold e3 in H4.
      injection H4; intros.
      exfalso.
      unfold not in H3.
      apply H1 with (T := T3) (e := e4).
      symmetry. assumption.

      Case "cast_precheck".
      rewrite <- H2 in H5. unfold e5 in H5. unfold e3 in H5.
      injection H5; intros.
      rewrite H7. rewrite H8. rewrite H9. rewrite H10. rewrite H11.
      reflexivity.

      Case "cast_check".
      unfold e'0 in H3. unfold e' in H3.
      injection H3; intros.
      rewrite H5. rewrite H6. rewrite H7. rewrite H8.
      reflexivity.

      Case "ok".
      unfold e5 in H3. unfold e3 in H3.
      injection H3; intros.
      rewrite H6.
      reflexivity.

      Case "fail".
      unfold e5 in H3. unfold e3 in H3.
      injection H3; intros.
      rewrite H5.
      reflexivity.
  Qed.

  Lemma lemm_eval_deterministic :
    forall e e1 e2, (e --> e1) -> (e --> e2) -> e1 = e2.
  Proof.
    intros.
    inversion H.
      Case "red".
      inversion H0.
        SCase "red".
        assert (E = E0 /\ e0 = e4).
          apply lemm_red_ectx_decomp with (e := e) (e2 := e3) (e2' := e5);
            auto.
        inversion H7.
        rewrite H8.
        rewrite H9 in H1.
        assert (e3 = e5).
          apply lemm_red_deterministic with (e := e4); assumption.
        rewrite H10.
        reflexivity.

        SCase "blame".
        inversion H.
          SSCase "e red e2".
          rewrite <- H5 in H8.
          exfalso.
          apply lemm_red_blame_exfalso with (E1 := E1) (E2 := E0)
                                            (e := e4) (e' := e5) (l := l);
            assumption.

          SSCase "e blame e2".
          rewrite <- H5 in H2.
          exfalso.
          apply lemm_red_blame_exfalso with (E1 := E) (E2 := E0)
                                            (e := e0) (e' := e3) (l := l);
            assumption.

      Case "blame".
      inversion H0.
        SCase "e red e2".
        rewrite <- H2 in H5.
        exfalso.
        apply lemm_red_blame_exfalso with (E1 := E0) (E2 := E)
                                          (e := e0) (e' := e3) (l := l);
          assumption.

        SCase "e blame e2".
        rewrite <- H2 in H5.
        assert (E0 = E).
          apply lemm_ectx_blame_eq with (E := E0) (E' := E) (l := l0) (l' := l);
            assumption.
        rewrite H7 in H5.
        f_equal.
        symmetry.
        apply lemm_ectx_blame_eq_label with (E := E) (l := l0) (l' := l);
          assumption.
  Qed.

  Lemma lemm_coterm_left_red :
    forall e1 e2 e e1',
      (e1 --> e2) ->
      (closed_term e2) ->
      (substO e [e1] ~~> e1') ->
      exists e', forall E',
        (E'@[substO e [e2]] -->* E'@[substO e' [e2]]) /\
        (E'@[e1'] -->* E'@[substO e' [e1]]).
  Proof.
    destruct lemm_subst_compose as [ H_subst_comp H_subst_comp' ].
    destruct H_subst_comp' as [ H_subst_comp_val H_subst_comp_ty ].

    destruct lemm_subst_shift_eq_shift_subst as [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].

    destruct lemm_subst_dummy as [ H_subst_dummy H_subst_dummy' ].

    intros e1 e2 e e1' H_eval H_right_closed H_subst_red.

    destruct e.
      Case "e = x".
      unfold substO in H_subst_red.
      simpl in H_subst_red.
      remember (v - 0) as v0.
      destruct v0.
        SCase "x = 0".
        assert (0 = v) as H_v0. omega.
        rewrite <- H_v0. unfold substO. simpl.
        assert (e1 --> e1') as H_eval_e1_e1'.
          assert (e1 = ectx_hole@[e1]). reflexivity.
          assert (e1' = ectx_hole@[e1']). reflexivity.
          rewrite H. rewrite H0.
          apply eval_red.
          assumption.
        assert (e2 = e1').
          apply lemm_eval_deterministic with (e := e1); assumption.
        rewrite <- H.
        exists e2.
        unfold closed_term in H_right_closed.
        rewrite (H_right_closed 0 e2).
        rewrite (H_right_closed 0 e1).
        intro E'.
        split; apply rt_refl.

        SCase "x = S _".
        destruct v0; inversion H_subst_red.

      Case "e = v".
      exfalso.
      apply (lemm_red_val_stuck (subst_val v (O, [e1])) e1').
      assumption.

      Case "e = app".
      assert (exists v3, e3 = !v3) as H_left_val.
        apply lemm_left_red_ectx with (e1 := e1) (e2 := e2)
                                      (e' := e1')
                                      (E := ectx_app1 ectx_hole e4);
          try assumption; try discriminate.
      inversion H_left_val as [ lval H_left_val' ].
      rewrite -> H_left_val' in H_subst_red.
      assert (exists v4, e4 = !v4) as H_right_val.
        apply lemm_left_red_ectx with (e1 := e1) (e2 := e2)
                                      (e' := e1')
                                      (E := ectx_app2 lval ectx_hole);
          try assumption; try discriminate.
      inversion H_right_val as [ rval H_right_val' ].
      rewrite -> H_right_val' in H_subst_red.
      assert (exists e', (e1' = substO e' [e1]) /\
                         (substO (!lval @ !rval) [e2] ~~> substO e' [e2]))
        as H_exists.
        apply lemm_coterm_app_aux. assumption.
      inversion H_exists as [ e' H_exists' ].
      inversion H_exists' as [ H_exists1 H_exists2 ].
      exists e'.
      split.
        rewrite H_left_val'. rewrite H_right_val'.
        apply rt_step.
        rewrite lemm_ectx_hole_ident.
        assert (substO (!lval @ !rval) [e2] =
                ectx_hole @[substO (!lval @ !rval) [e2]]) as H_ectx_hole.
          apply lemm_ectx_hole_ident.
        rewrite H_ectx_hole.
        simpl.
        apply eval_red.
        assumption.

        rewrite H_exists1.
        apply rt_refl.

      Case "e = tapp".
      assert (exists v, e = !v) as H_left_val.
        apply lemm_left_red_ectx with (e1 := e1) (e2 := e2)
                                      (e' := e1')
                                      (E := ectx_tapp ectx_hole t);
          try assumption; try discriminate.
      inversion H_left_val as [ v H_left_val' ].
      rewrite H_left_val' in H_subst_red.
      assert (exists e', (e1' = substO e' [e1]) /\
                         (substO (!v @@ t) [e2] ~~> substO e' [e2]))
        as H_exists.
        apply lemm_coterm_tapp_aux. assumption.
      inversion H_exists as [ e' H_exists' ].
      inversion H_exists' as [ H_exists1 H_exists2 ].
      exists e'.
      split.
        apply rt_step.
        rewrite lemm_ectx_hole_ident.
        assert (substO (!v @@ t) [e2] =
                ectx_hole @[substO (!v @@ t) [e2]]) as H_ectx_hole.
          apply lemm_ectx_hole_ident.
        rewrite H_left_val'.
        rewrite H_ectx_hole.
        simpl.
        apply eval_red.
        assumption.

        rewrite H_exists1.
        apply rt_refl.

      Case "e = wcheck".
      assert (exists v4, e4 = !v4) as H_val.
        apply lemm_left_red_ectx with (e1 := e1) (e2 := e2) (e' := e1')
                                      (E := ectx_wcheck t e3 ectx_hole l);
          try assumption; try discriminate.
      inversion H_val as [ v4 H_val' ].
      rewrite H_val' in H_subst_red.
      assert (exists e', (e1' = substO e' [e1]) /\
                         (substO (term_wcheck t e3 !v4 l) [e2] ~~>
                                 substO e' [e2])) as H_exists.
        apply lemm_coterm_wcheck_aux; assumption.
      inversion H_exists as [ e' H_exists' ].
      inversion H_exists' as [ H_exists1 H_exists2 ].
      exists e'.
      repeat (unfold substO; unfold shift_subst; simpl).
      split.
        apply rt_step.
        rewrite lemm_ectx_hole_ident.
        assert (term_wcheck (subst_ty t (0, [e2]))
                              (subst e3 (1, [shift e2 0]))
                              (subst e4 (0, [e2])) l =
                  ectx_hole @[term_wcheck (subst_ty t (0, [e2]))
                              (subst e3 (1, [shift e2 0]))
                              (subst e4 (0, [e2])) l]) as H_ectx_hole.
            reflexivity.
          rewrite H_ectx_hole.
          simpl.
          apply eval_red.
          repeat (unfold substO in H7; unfold shift_subst in H7; simpl in H7).
          rewrite H_val'. assumption.

          rewrite H_exists1.
          apply rt_refl.

      Case "e = acheck".
      assert (exists v4, e4 = !v4) as H_val.
        apply lemm_left_red_ectx with (e1 := e1) (e2 := e2) (e' := e1')
                                      (E := ectx_acheck t e3 ectx_hole v l);
            try assumption; try discriminate.
      inversion H_val as [ v4 H_val' ].
      rewrite H_val' in H_subst_red.
      assert (exists e', (e1' = substO e' [e1]) /\
                         (substO (term_acheck t e3 !v4 v l) [e2] ~~>
                                 substO e' [e2])) as H_exists.
        apply lemm_coterm_acheck_aux; assumption.
      inversion H_exists as [ e' H_exists' ].
      inversion H_exists' as [ H_exists1 H_exists2 ].
      exists e'.
      repeat (unfold substO; unfold shift_subst; simpl).
      split.
        apply rt_step.
        rewrite lemm_ectx_hole_ident.
        assert (term_acheck (subst_ty t (0, [e2]))
                              (subst e3 (1, [shift e2 0]))
                              (subst e4 (0, [e2]))
                              (subst_val v (0, [e2]))
                              l =
                  ectx_hole @[term_acheck (subst_ty t (0, [e2]))
                              (subst e3 (1, [shift e2 0]))
                              (subst e4 (0, [e2]))
                              (subst_val v (0, [e2]))
                              l]) as H_ectx_hole.
            reflexivity.
        rewrite H_ectx_hole.
        simpl.
        apply eval_red.
        repeat (unfold substO in H_exists2;
                unfold shift_subst in H_exists2;
                simpl in H_exists2).
        rewrite H_val'. assumption.

        rewrite H_exists1.
        apply rt_refl.

      Case "e = blame".
      exfalso.
      apply lemm_red_blame_stuck with (E := ectx_hole) (l := l) (e := e1').
      assumption.
  Qed.

  Lemma lemm_subst_var_eval :
    forall x n l e,
      (subst (term_var x) (n, l) --> e) ->
      In (subst (term_var x) (n, l)) l.
  Proof.
    intros.

    assert (n <= x).
      unfold subst in H. simpl in H.
      destruct (lt_dec x n); try omega.
      inversion H.
        destruct E; try discriminate.
        simpl in H0. simpl in H2.
        rewrite H0 in H1.
        inversion H1.

        destruct E; try discriminate.

    assert (x - n < length l).
      unfold subst in H. simpl in H.
      destruct (lt_dec x n). try omega.
      assert (~ x-n >= length l).
        unfold not. intro.
        assert (forall y, (term_var y --> e) -> False).
          intros. inversion H2.
          destruct E; try discriminate.
          simpl in H3. rewrite H3 in H4. inversion H4.
          destruct E; discriminate.
        apply H2 with (y := x - Datatypes.length l).
        assert (nth (x - n) l (term_var (x - Datatypes.length l)) =
                term_var (x - Datatypes.length l)).
          apply nth_overflow.
          omega.
        rewrite <- H3.
        assumption.
        omega.

    unfold subst. simpl.
    destruct (lt_dec x n); try omega.
    apply nth_In. assumption.
  Qed.

  Lemma lemm_substO_var_eval :
    forall x e e',
      (substO (term_var x) [e] --> e') ->
      x = O.
  Proof.
    intros.
    unfold substO in H. unfold subst in H. simpl in H.
    assert (x - 0 = x). omega.
    rewrite H0 in H.
    destruct x; try reflexivity.
    assert ((term_var (S x - 1) --> e') -> S x = 0).
      intro. exfalso.
      inversion H1.
      destruct E; try discriminate. simpl in H2. rewrite H2 in H3.
      inversion H3.
      destruct E; try discriminate.
    destruct x; apply H1; assumption.
  Qed.

  Lemma lemm_eval_subterm_blame :
    forall e l E,
      (e --> term_blame l) ->
      (E@[e] -->* term_blame l).
  Proof.
    intros.
    destruct E;
      try (apply rt_step; assumption);
      try (inversion H;
           try (rewrite lemm_ectx_compose;
                apply rt_step;
                apply eval_blame;
                discriminate)).
    Case "E = app1".
    rewrite lemm_ectx_compose.
    apply rt_trans with (y := (ectx_app1 E t ^^ E0)@[e2]).
      apply rt_step. apply eval_red. assumption.
      rewrite <- lemm_ectx_compose.
      repeat (rewrite H0).
      apply rt_step. apply eval_blame. discriminate.

    Case "E = app2".
    rewrite lemm_ectx_compose.
    apply rt_trans with (y := (ectx_app2 v E ^^ E0)@[e2]).
      apply rt_step. apply eval_red. assumption.
      rewrite <- lemm_ectx_compose.
      repeat (rewrite H0).
      apply rt_step. apply eval_blame. discriminate.

    Case "E = tapp".
    rewrite lemm_ectx_compose.
    apply rt_trans with (y := (ectx_tapp E t ^^ E0)@[e2]).
      apply rt_step. apply eval_red. assumption.
      rewrite <- lemm_ectx_compose.
      repeat (rewrite H0).
      apply rt_step. apply eval_blame. discriminate.

    Case "E = wcheck".
    rewrite lemm_ectx_compose.
    apply rt_trans with (y := (ectx_wcheck t t0 E l0 ^^ E0)@[e2]).
      apply rt_step. apply eval_red. assumption.
      rewrite <- lemm_ectx_compose.
      repeat (rewrite H0).
      apply rt_step. apply eval_blame. discriminate.

    Case "E = acheck".
    rewrite lemm_ectx_compose.
    apply rt_trans with (y := (ectx_acheck t t0 E v l0 ^^ E0)@[e2]).
      apply rt_step. apply eval_red. assumption.
      rewrite <- lemm_ectx_compose.
      repeat (rewrite H0).
      apply rt_step. apply eval_blame. discriminate.
  Qed.


  Lemma lemm_coterm_left_subterm_red :
    forall e1 e2 E e e1' e2',
      (closed_term e1) ->
      (closed_term e2) ->
      (e1 --> e2) ->
      (substO e [e1] --> E@[e2']) ->
      (substO e [e1] = E@[e1']) ->
      (e1' ~~> e2') ->
      exists e', forall E',
        (E'@[substO e [e2]] -->* E'@[substO e' [e2]]) /\
        (E'@[E@[e2']] -->* E'@[substO e' [e1]]).
  Proof.
    intros e1 e2 E.
    induction E;
      intros e e1' e2'
             H_left_closed H_right_closed
             H_eval H_subst_eval H_orig H_red.
      Case "E = []".
      simpl in H_orig.
      rewrite <- H_orig in H_red.
      simpl.
      destruct e;
        try (apply lemm_coterm_left_red; try assumption;
             intro; discriminate;
             simpl in H_orig;
             rewrite <- H_orig in H_red;
             assumption).

      Case "E = app1".
      destruct e; try discriminate.
        SCase "e = x".
        exists e2.
        assert (v = O) as H_v0.
          apply lemm_substO_var_eval with (e := e1) (e' := ectx_app1 E t @[ e2']).
          assumption.
        rewrite H_v0.
        assert (forall e, substO e2 [e] = e2).
          unfold closed_term in H_right_closed.
          apply H_right_closed.
        split.
          SSCase "red".
          rewrite H.
          apply rt_refl.

          SSCase "-->*".
          rewrite H_v0 in H_subst_eval.
          unfold substO in H_subst_eval.
          simpl in H_subst_eval.
          assert (e2 = E@[e2'] @ t).
            apply lemm_eval_deterministic with (e := e1); assumption.
          simpl.
          rewrite <- H0.
          rewrite H.
          apply rt_refl.

        SCase "e = app".
        assert (exists e', forall E',
                   (E'@[substO e3 [e2]] -->* E'@[substO e' [e2]]) /\
                   (E'@[E @[ e2']] -->* E'@[substO e' [e1]])).
          unfold substO in H_orig.
          injection H_orig; intros.
          apply IHE with (e1' := e1'); try assumption.
          unfold substO.
          rewrite H0.
          apply eval_red; assumption.
        inversion H.
        exists (x @ e4).
        split.
          SSCase "red".
          destruct (H0 (E' ^^ (ectx_app1 ectx_hole (subst e4 (0, [e2]))))).
          rewrite <- lemm_ectx_compose in H1.
          rewrite <- lemm_ectx_compose in H1.
          simpl in H1.
          assumption.

          SSCase "-->*".
          injection H_orig; intros.
          destruct (H0 (E' ^^ (ectx_app1 ectx_hole t))).
          rewrite <- lemm_ectx_compose in H4.
          rewrite <- lemm_ectx_compose in H4.
          simpl in H4.
          unfold substO.
          simpl.
          rewrite H1.
          assumption.

      Case "E = app2".
      destruct e; try discriminate.
        SCase "e = x".
        exists e2.
        assert (v0 = O) as H_v0.
          apply lemm_substO_var_eval with (e := e1) (e' := ectx_app2 v E @[ e2']).
          assumption.
        rewrite H_v0.
        assert (forall e, substO e2 [e] = e2).
          unfold closed_term in H_right_closed.
          apply H_right_closed.
        split.
          SSCase "red".
          rewrite H.
          apply rt_refl.

          SSCase "-->*".
          rewrite H_v0 in H_subst_eval.
          unfold substO in H_subst_eval.
          simpl in H_subst_eval.
          assert (e2 = !v @ (E@[e2'])).
            apply lemm_eval_deterministic with (e := e1); assumption.
          simpl.
          rewrite <- H0.
          rewrite H.
          apply rt_refl.

        SCase "e = app".
        unfold substO in H_orig.
        injection H_orig; intros.
        destruct e3; try discriminate.
          simpl in H0.
          destruct (v0 - 0).
            exfalso.
            rewrite H0 in H_eval. apply (lemm_eval_val_stuck v e2). assumption.
            destruct n; discriminate.

            assert (exists e', forall E',
                         (E'@[substO e4 [e2]] -->* E'@[substO e' [e2]]) /\
                         (E'@[E @[ e2']] -->* E'@[substO e' [e1]])).
            injection H_orig; intros.
            apply IHE with (e1' := e1'); try assumption.
            unfold substO.
            rewrite H.
            apply eval_red; assumption.
            inversion H1.
            exists (!v0 @ x).
            split.
              SSCase "red".
              destruct
                (H2 (E' ^^ (ectx_app2 (subst_val v0 (0, [e2])) ectx_hole))).
              rewrite <- lemm_ectx_compose in H3.
              rewrite <- lemm_ectx_compose in H3.
              unfold substO in H3.
              simpl in H3.
              unfold substO.
              simpl.
              assumption.

              SSCase "-->*".
              injection H_orig; intros.
              destruct
                (H2 (E' ^^ (ectx_app2 (subst_val v0 (0, [e1])) ectx_hole))).
              rewrite <- lemm_ectx_compose in H6.
              rewrite <- lemm_ectx_compose in H6.
              simpl in H6.
              unfold substO.
              simpl.
              rewrite <- H4.
              assumption.

      Case "E = tapp".
      destruct e; try discriminate.
        SCase "e = x".
        exists e2.
        assert (v = O) as H_v0.
          apply lemm_substO_var_eval with (e := e1) (e' := ectx_tapp E t @[ e2']).
          assumption.
        rewrite H_v0.
        assert (forall e, substO e2 [e] = e2).
          unfold closed_term in H_right_closed.
          apply H_right_closed.
        split.
          SSCase "red".
          rewrite H.
          apply rt_refl.

          SSCase "-->*".
          rewrite H_v0 in H_subst_eval.
          unfold substO in H_subst_eval.
          simpl in H_subst_eval.
          assert (e2 = E@[e2'] @@ t).
            apply lemm_eval_deterministic with (e := e1); assumption.
          simpl.
          rewrite <- H0.
          rewrite H.
          apply rt_refl.

        SCase "e = tapp".
        injection H_orig; intros.
        assert (exists e', forall E',
                   (E'@[substO e [e2]] -->* E'@[substO e' [e2]]) /\
                   (E'@[E @[ e2']] -->* E'@[substO e' [e1]])).
          apply IHE with (e1' := e1'); try assumption.
          unfold substO.
          rewrite H0.
          apply eval_red; assumption.
        inversion H1.
        exists (x @@ t0).
        split.
          SSCase "red".
          destruct (H2 (E' ^^ (ectx_tapp ectx_hole (subst_ty t0 (0, [e2]))))).
          rewrite <- lemm_ectx_compose in H3.
          rewrite <- lemm_ectx_compose in H3.
          unfold substO in H3.
          simpl in H3.
          unfold substO.
          simpl.
          assumption.

          SSCase "-->*".
          destruct (H2 (E' ^^ (ectx_tapp ectx_hole (subst_ty t0 (0, [e1]))))).
          rewrite <- lemm_ectx_compose in H4.
          rewrite <- lemm_ectx_compose in H4.
          simpl in H4.
          unfold substO.
          simpl.
          rewrite <- H.
          assumption.

      Case "E = wcheck".
      destruct e; try discriminate.
        SCase "e = x".
        exists e2.
        assert (v = O) as H_v0.
          apply lemm_substO_var_eval with (e := e1)
                                          (e' := ectx_wcheck t t0 E l @[ e2']).
          assumption.
        rewrite H_v0.
        assert (forall e, substO e2 [e] = e2).
          unfold closed_term in H_right_closed.
          apply H_right_closed.
        split.
          SSCase "red".
          rewrite H.
          apply rt_refl.

          SSCase "-->*".
          rewrite H_v0 in H_subst_eval.
          unfold substO in H_subst_eval.
          simpl in H_subst_eval.
          assert (e2 = term_wcheck t t0 (E @[ e2']) l).
            apply lemm_eval_deterministic with (e := e1); assumption.
          simpl.
          rewrite <- H0.
          rewrite H.
          apply rt_refl.

        SCase "e = wcheck".
        injection H_orig; intros.
        assert (exists e', forall E',
                   (E'@[substO e4 [e2]] -->* E'@[substO e' [e2]]) /\
                   (E'@[E @[ e2']] -->* E'@[substO e' [e1]])).
          apply IHE with (e1' := e1'); try assumption.
          unfold substO.
          rewrite H0.
          apply eval_red; assumption.
        inversion H3.
        exists (term_wcheck t1 e3 x l0).
        split.
          SSCase "red".
          destruct
            (H4 (E' ^^ (ectx_wcheck (subst_ty t1 (O, [e2]))
                                    (subst e3 (shift_subst (O, [e2])))
                                    ectx_hole
                                    l0))).
          rewrite <- lemm_ectx_compose in H5.
          rewrite <- lemm_ectx_compose in H5.
          unfold substO in H5.
          simpl in H5.
          unfold substO.
          simpl.
          assumption.

          SSCase "-->*".
          destruct
            (H4 (E' ^^ (ectx_wcheck (subst_ty t1 (O, [e1]))
                                    (subst e3 (shift_subst (O, [e1])))
                                    ectx_hole
                                    l0))).
          rewrite <- lemm_ectx_compose in H6.
          rewrite <- lemm_ectx_compose in H6.
          simpl in H6.
          unfold substO.
          simpl.
          rewrite <- H.
          rewrite <- H1.
          rewrite <- H2.
          assumption.

      Case "E = acheck".
      destruct e; try discriminate.
        SCase "e = x".
        exists e2.
        assert (v0 = O) as H_v0.
          apply lemm_substO_var_eval with (e := e1)
                                          (e' := ectx_acheck t t0 E v l @[ e2']).
          assumption.
        rewrite H_v0.
        assert (forall e, substO e2 [e] = e2).
          unfold closed_term in H_right_closed.
          apply H_right_closed.
        split.
          SSCase "red".
          rewrite H.
          apply rt_refl.

          SSCase "-->*".
          rewrite H_v0 in H_subst_eval.
          unfold substO in H_subst_eval.
          simpl in H_subst_eval.
          assert (e2 = term_acheck t t0 (E @[ e2']) v l).
            apply lemm_eval_deterministic with (e := e1); assumption.
          simpl.
          rewrite <- H0.
          rewrite H.
          apply rt_refl.

        SCase "e = acheck".
        injection H_orig; intros.
        assert (exists e', forall E',
                   (E'@[substO e4 [e2]] -->* E'@[substO e' [e2]]) /\
                   (E'@[E @[ e2']] -->* E'@[substO e' [e1]])).
          apply IHE with (e1' := e1'); try assumption.
          unfold substO.
          rewrite H1.
          apply eval_red; assumption.
        inversion H4.
        exists (term_acheck t1 e3 x v0 l0).
        split.
          SSCase "red".
          destruct
            (H5 (E' ^^ (ectx_acheck (subst_ty t1 (O, [e2]))
                                    (subst e3 (shift_subst (O, [e2])))
                                    ectx_hole
                                    (subst_val v0 (O, [e2]))
                                    l0))).
          rewrite <- lemm_ectx_compose in H6.
          rewrite <- lemm_ectx_compose in H6.
          unfold substO in H6.
          simpl in H6.
          unfold substO.
          simpl.
          apply H6.

          SSCase "-->*".
          destruct
            (H5 (E' ^^ (ectx_acheck (subst_ty t1 (O, [e1]))
                                    (subst e3 (shift_subst (O, [e1])))
                                    ectx_hole
                                    (subst_val v0 (O, [e1]))
                                    l0))).
          rewrite <- lemm_ectx_compose in H7.
          rewrite <- lemm_ectx_compose in H7.
          simpl in H7.
          unfold substO.
          simpl.
          rewrite <- H.
          rewrite <- H0.
          rewrite <- H2.
          rewrite <- H3.
          assumption.
  Qed.

  Lemma lemm_coterm_left_eval :
    forall e1 e2 e e1',
      (closed_term e1) ->
      (closed_term e2) ->
      (e1 --> e2) ->
      (substO e [e1] --> e1') ->
      exists e',
        (substO e [e2] -->* substO e' [e2]) /\
        (e1' -->* substO e' [e1]).
  Proof.
    destruct lemm_subst_compose as [ H_subst_comp H_subst_comp' ].
    destruct H_subst_comp' as [ H_subst_comp_val H_subst_comp_ty ].

    destruct lemm_subst_shift_eq_shift_subst as [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].

    destruct lemm_subst_dummy as [ H_subst_dummy H_subst_dummy' ].

    intros e1 e2 e e1' H_left_closed H_right_closed H_eval H_subst_eval.

    inversion H_subst_eval.
    Case "E_red".
    assert (exists e', forall E',
                 (E'@[substO e [e2]] -->* E'@[substO e' [e2]]) /\
                 (E'@[E@[e3]] -->* E'@[substO e' [e1]])).
      apply lemm_coterm_left_subterm_red with (e1' := e0); try assumption.
      rewrite <- H1 in H_subst_eval. assumption.
      symmetry. assumption.
    inversion H2.
    exists x.
    apply (H3 ectx_hole).

    Case "E_blame".
    exists (term_blame l).
    split.
      assert (exists E', substO e [e2] = E'@[term_blame l]).
        apply lemm_coterm_left_blame with (e1 := e1) (E1 := E).
        assumption.
        symmetry. assumption.
      inversion H2.
      rewrite H3.
      destruct x;
        try (apply rt_refl);
        apply rt_step;
        apply eval_blame;
        discriminate.

      apply rt_refl.
  Qed.

  Lemma lemm_shift_map :
    forall e n, [shift e n] = map (fun x => shift x n) [e].
  Proof.
    intros. reflexivity.
  Qed.

  Lemma lemm_tyshift_map :
    forall e n, [tyshift e n] = map (fun x => tyshift x n) [e].
  Proof.
    intros. reflexivity.
  Qed.

  Lemma lemm_red_closed :
    forall e1 e2,
      (closed_term e1) ->
      (e1 ~~> e2) ->
      closed_term e2.
  Proof.
    destruct lemm_subst_compose as [ H_subst_comp H_subst_comp' ].
    destruct H_subst_comp' as [ H_subst_comp_val H_subst_comp_ty ].

    destruct lemm_shift_swap as [ H_shift_swap H_shift_swap' ].
    destruct H_shift_swap' as [ H_shift_swap_val H_shift_swap_ty ].

    destruct lemm_subst_tysubst as [ H_subst_tysubst H_subst_tysubst' ].

    destruct lemm_subst_shift_eq_shift_subst as [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].

    destruct lemm_tyshift_subst as [ H_tyshift_subst H_tyshift_subst' ].
    destruct H_tyshift_subst' as [ H_tyshift_subst_val H_tyshift_subst_ty ].

    assert (forall e, shift (shift (shift (shift e 0) 0) 0) 0 =
                      shift (shift (shift (shift e 0) 0) 0) 1) as H_e_shift4.
      intros; apply H_shift_swap; omega.

    assert (forall e, shift (shift (shift e 0) 0) 0 =
                      shift (shift (shift e 0) 0) 1) as H_e_shift3.
      intros; apply H_shift_swap; omega.

    assert (forall e, shift (shift e 0) 0 = shift (shift e 0) 1) as H_e_shift2.
      intros; apply H_shift_swap; omega.

    intros.
    unfold closed_term in H.
    inversion H0;
      unfold closed_term; intros;
      unfold substO; unfold tysubstO;
      simpl.

    Case "beta".
    rewrite <- H1 in H.
    unfold e in H.
    assert (subst (subst e0 (0, [!v2; fix_v])) (n, [e']) =
            subst (subst e0 (S (S n), [shift (shift e' 0) 1]))
                  (0, [subst !v2 (n, [e']); subst fix_v (n, [e'])])).
      apply H_subst_comp; omega.
    rewrite H3.
    pose proof (H n e') as H'.
    simpl in H'.
    injection H'; intros.
    simpl.
    rewrite H4. rewrite H5. rewrite H6. rewrite H7.
    unfold shift_subst in H5. simpl in H5.
    rewrite <- H_shift_swap; try omega.
    rewrite H5.
    reflexivity.

    Case "tbeta".
    pose proof (H n e'0) as H'.
    rewrite <- H1 in H'.
    injection H'; intros.
    unfold tyshift_subst in H4. simpl in H4.
    rewrite H_subst_tysubst.
    rewrite H4.
    rewrite H3.
    reflexivity.

    Case "cast_beta".
    pose proof (H n e') as H'.
    rewrite <- H1 in H'.
    injection H'; intros.
    f_equal.
    assumption.

    Case "cast_fun".
    unfold shift_subst. simpl.
    pose proof (H n e') as H'.
    rewrite <- H1 in H'.
    injection H'; intros.
    unfold shift_subst in H4. simpl in H4.
    unfold shift_subst in H6. simpl in H6.
    rewrite H6.
    rewrite H7.
    unfold codomain_cast. unfold domain_cast. unfold v'.
    unfold T11'. unfold T12'. unfold T21'. unfold T22'. unfold T12''.
    unfold shift_ty_n. unfold shift_n. simpl.
    repeat f_equal.

    repeat (rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega).
    repeat f_equal.
    assumption.

    rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega.
    f_equal.
    rewrite H_shift_swap; try omega.
    rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega.
    f_equal.
    assumption.

    repeat (rewrite <- H_shift_swap_ty; try omega).
    repeat (rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega).
    repeat f_equal.
    rewrite H_shift_swap; try omega.
    rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega.
    rewrite H6. reflexivity.

    rewrite H_e_shift4. rewrite H_e_shift3. rewrite H_e_shift2.
    repeat (rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega).
    repeat f_equal.
    assumption.

    repeat (rewrite lemm_shift_map; rewrite H_ss_eq_ss_val; try omega).
    rewrite H3.
    reflexivity.

    repeat (rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega).
    rewrite H5.
    reflexivity.

    repeat (rewrite lemm_shift_map; rewrite H_ss_eq_ss_ty; try omega).
    rewrite H7.
    reflexivity.

    Case "cast_forall".
    pose proof (H n e') as H'.
    rewrite <- H1 in H'.
    unfold e in H'.
    injection H'; intros.
    unfold tyshift_subst in H4. unfold tyshift_subst in H5. simpl.
    unfold cast.
    unfold body.
    unfold tyshift_subst. simpl.
    repeat f_equal; try assumption.

    rewrite lemm_tyshift_map.
    rewrite <- H_tyshift_subst_val.
    rewrite H3.
    reflexivity.

    Case "cast_forget".
    pose proof (H n e') as H'.
    rewrite <- H1 in H'.
    unfold e in H'.
    injection H'; intros.
    rewrite H3. rewrite H5. rewrite H6.
    reflexivity.

    Case "cast_precheck".
    pose proof (H n e') as H'.
    rewrite <- H2 in H'.
    unfold e in H'.
    injection H'; intros.
    rewrite H4. rewrite H5. rewrite H6. rewrite H7.
    reflexivity.

    Case "check".
    pose proof (H n e'0) as H'.
    rewrite <- H1 in H'.
    unfold e' in H'.
    injection H'; intros.
    unfold shift_subst in H4. simpl in H4.
    unfold shift_subst. simpl.
    rewrite H3. rewrite H4. rewrite H5.
    f_equal.
    assert (subst e (0, [!v]) =
            subst (shift e 1) (0, [!v; term_val (val_const (c_bool true))])).
      apply lemm_subst_dummy.
    rewrite H6.
    rewrite H_subst_comp; try omega.
    repeat f_equal.

    rewrite lemm_shift_map.
    rewrite H_ss_eq_ss; try omega.
    rewrite H4.
    reflexivity.

    simpl.
    rewrite H3.
    reflexivity.

    Case "ok".
    pose proof (H n e') as H'.
    rewrite <- H1 in H'.
    unfold e0 in H'.
    injection H'; intros.
    rewrite H3.
    reflexivity.

    Case "fail".
    reflexivity.
  Qed.


  Lemma lemm_ectx_closed :
    forall E e,
      closed_term (E@[e]) <-> (closed_ectx E /\ closed_term e).
  Proof.
    intros.
    unfold closed_term.
    unfold closed_ectx.
    split.
    Case "->".
    intro.
    induction E.
      SCase "E = []".
      split.
        SSCase "closed E".
        intros.
        reflexivity.
        SSCase "clolsed e".
        assumption.
      SCase "E = app1".
      assert (forall n e', subst (E@[e]) (n, [e']) = E@[e]) as H_E_closed.
          intros n0 e'0.
          pose proof (H n0 e'0) as H'.
          injection H'; intros.
          assumption.
      split.
        SSCase "closed E".
        intros.
        pose proof (H n e') as H'.
        injection H'; intros.
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        simpl.
        f_equal.
        apply (H2 n e').
        apply H0.

        SSCase "closed e".
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        assumption.

      SCase "E = app2".
      assert (forall n e', subst (E@[e]) (n, [e']) = E@[e]) as H_E_closed.
        intros n0 e'0.
        pose proof (H n0 e'0) as H'.
        injection H'; intros.
        assumption.
      split.
        SSCase "closed E".
        intros.
        pose proof (H n e') as H'.
        injection H'; intros.
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        simpl.
        f_equal.
        apply H1.
        apply (H2 n e').

        SSCase "closed e".
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        assumption.

      SCase "E = tapp".
      assert (forall n e', subst (E@[e]) (n, [e']) = E@[e]) as H_E_closed.
        intros n0 e'0.
        pose proof (H n0 e'0) as H'.
        injection H'; intros.
        assumption.
      split.
        SSCase "closed E".
        intros.
        pose proof (H n e') as H'.
        injection H'; intros.
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        simpl.
        f_equal.
        apply (H2 n e').
        apply H0.

        SSCase "closed e".
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        assumption.

      SCase "E = wcheck".
      assert (forall n e', subst (E@[e]) (n, [e']) = E@[e]) as H_E_closed.
        intros n0 e'0.
        pose proof (H n0 e'0) as H'.
        injection H'; intros.
        assumption.
      split.
        SSCase "closed E".
        intros.
        pose proof (H n e') as H'.
        injection H'; intros.
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        simpl.
        f_equal.
        apply H2.
        apply H1.
        apply (H3 n e').

        SSCase "closed e".
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        assumption.

      SCase "E = acheck".
      assert (forall n e', subst (E@[e]) (n, [e']) = E@[e]) as H_E_closed.
        intros n0 e'0.
        pose proof (H n0 e'0) as H'.
        injection H'; intros.
        assumption.
      split.
        SSCase "closed E".
        intros.
        pose proof (H n e') as H'.
        injection H'; intros.
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        simpl.
        f_equal; try assumption.
        apply (H4 n e').

        SSCase "closed e".
        pose proof (IHE H_E_closed) as IHE'.
        inversion IHE'.
        assumption.

    Case "<-".
    intros.
    induction E;
      inversion H as [ H_closed_E H_closed_e ];
      simpl;
      f_equal;
      try (pose proof (H_closed_E n e') as H_closed_E';
           injection H_closed_E'; intros;
           try f_equal; assumption);
      try (apply IHE; split; intros n0 e'0;
           try (pose proof (H_closed_E n0 e'0) as H_closed_E';
                injection H_closed_E'; intros);
           try (pose proof (H_closed_e n0 e'0) as H_closed_e';
                assumption)).

      SCase "E = []".
      apply H_closed_e.
  Qed.

  Lemma lemm_eval_closed :
    forall e1 e2,
      (closed_term e1) ->
      (e1 --> e2) ->
      closed_term e2.
  Proof.
    intros e1 e2 H_e1_closed H_eval.
    inversion H_eval.

    Case "red".
    rewrite <- H0 in H_e1_closed.
    pose proof (lemm_ectx_closed E e0) as H_E_e0_closed.
    inversion H_E_e0_closed.
    pose proof (H2 H_e1_closed) as H_E_e0_closed'.
    inversion H_E_e0_closed'.
    pose proof (lemm_red_closed e0 e3 H5 H) as H_e3_closed.
    apply lemm_ectx_closed.
    auto.

    Case "blame".
    unfold closed_term.
    intros.
    reflexivity.
  Qed.

  Lemma lemm_right_red_ectx :
    forall e1 e2 e e' E,
      (e1 --> e2) ->
      (E <> ectx_hole) ->
      (substO (E@[e]) [e2] ~~> e') ->
      (exists v, e = !v) \/ (e = term_var 0 /\ exists v, e2 = !v).
  Proof.
    intros e1 e2 e e' E H_eval H_E H_red.
    destruct e;
      try (destruct E;
           try (exfalso; unfold not in H_E; apply H_E; reflexivity);
           inversion H_red; destruct E; discriminate).

    Case "e = x".
    apply or_intror.
    unfold substO in H_red.
    induction E;
      simpl in H_red;
      try (try (destruct E);
           simpl in H_red;
           split;
           (
             try (remember (v - 0) as v_0;
                  destruct v_0;
                  try (f_equal; rewrite Heqv_0;
                       induction v; reflexivity; discriminate);
                  destruct v_0; inversion H_red);
             inversion H_red
           );
           (
             try (exfalso;
                  apply H_E;
                  reflexivity);
             try (remember (v - 0) as v_0;
                  destruct v_0);
             try (
                 try (exists v0; auto; destruct v_0; discriminate);
                 try (exists v1; auto; destruct v_0; discriminate);
                 try (exists v2; auto; destruct v_0; discriminate)
               )
           )
          ).

    exists (val_fix T1 T2 e0); auto.
    exists (val_cast (ty_base B) (ty_base B) l); auto.
    exists (val_cast (ty_fun T11 T12) (ty_fun T21 T22) l); auto.
    exists (val_cast (ty_forall T1) (ty_forall T2) l); auto.
    exists (val_cast T1 (ty_refine T2 e0) l); auto.
    exists (val_cast (ty_refine T1 e0) T2 l); auto.
    exists (val_tabs e); auto.
    destruct v_0; inversion H_red.
    exists tt; auto.
    exists ff; auto.

    Case "e = v".
    apply or_introl.
    exists v.
    reflexivity.
  Qed.

  Lemma lemm_coterm_right_red :
    forall e1 e2 e e2',
      (closed_term e1) ->
      (closed_term e2) ->
      (closed_term (substO e [e2])) ->
      (e1 --> e2) ->
      (substO e [e2] ~~> e2') ->
      exists e', forall E',
        (E'@[substO e [e1]] -->* E'@[substO e' [e1]]) /\
        (E'@[e2'] -->* E'@[substO e' [e2]]).
  Proof.
    destruct lemm_subst_compose as [ H_subst_comp H_subst_comp' ].
    destruct H_subst_comp' as [ H_subst_comp_val H_subst_comp_ty ].

    destruct lemm_subst_shift_eq_shift_subst as [ H_ss_eq_ss H_ss_eq_ss' ].
    destruct H_ss_eq_ss' as [ H_ss_eq_ss_val H_ss_eq_ss_ty ].

    destruct lemm_subst_dummy as [ H_subst_dummy H_subst_dummy' ].

    intros e1 e2 e e2' H_left_closed H_right_closed H_closed H_eval H_subst_red.

    destruct e.
      Case "e = x".
      unfold substO in H_subst_red.
      simpl in H_subst_red.
      remember (v - 0) as v0.
      destruct v0.
        SCase "x = 0".
        destruct H_eval.
          SSCase "e1 red e2".
          exists e2'.
          intro E'.
          pose proof (lemm_red_closed
                          (E@[e2]) e2' H_right_closed H_subst_red)
              as H_e2'_closed.
          unfold closed_term in H_e2'_closed; simpl in H_e2'_closed.
          split.
            pose proof (H_e2'_closed 0 (E@[e1])) as H_e2'_closed'.
            unfold substO; simpl.
            rewrite H_e2'_closed'.
            apply rt_trans with (y := E'@[E@[e2]]).
            rewrite <- Heqv0.
            apply rt_step;
              repeat (rewrite lemm_ectx_compose);
              apply eval_red;
              assumption.
            apply rt_step; apply eval_red; assumption.

            pose proof (H_e2'_closed 0 (E@[e2])) as H_e2'_closed'.
            unfold substO; simpl.
            rewrite H_e2'_closed'.
            apply rt_refl.

          SSCase "e1 blame e2".
          inversion H_subst_red.

        SCase "x > 0".
        destruct v0; inversion H_subst_red.

      Case "e = v".
      inversion H_subst_red.

      Case "e = app".
      assert ((exists v3, e3 = !v3) \/ (e3 = term_var 0 /\ exists v2, e2 = !v2))
        as H_fun.
        apply lemm_right_red_ectx with (e1 := e1) (e2 := e2)
                                       (e' := e2')
                                       (E := ectx_app1 ectx_hole e4);
          try assumption; try discriminate.
      inversion H_fun.
        SCase "e3 = v3".
        inversion H as [ v3 H_fun' ].
        assert ((exists v4, e4 = !v4) \/
                (e4 = term_var 0 /\ exists v2', e2 = !v2')) as H_arg.
          rewrite H_fun' in H_subst_red.
          apply lemm_right_red_ectx with (e1 := e1) (e2 := e2)
                                         (e' := e2')
                                         (E := ectx_app2 v3 ectx_hole);
          try assumption; try discriminate.
        rewrite H_fun'.
        inversion H_arg.
          SSCase "e4 = v4".
          inversion H0 as [ v4 H_arg' ].
          rewrite H_arg'.
          rewrite H_fun' in H_subst_red.
          rewrite H_arg' in H_subst_red.
          pose proof (lemm_coterm_app_aux e2 e1 v3 v4 e2' H_subst_red) as H_app.
          inversion H_app as [ e' H_app' ].
          inversion H_app' as [ H_app'1 H_app'2 ].
          exists e'.
          intro E'.
          split.
            apply rt_step; apply eval_red; assumption.
            rewrite H_app'1; apply rt_refl.

          SSCase "e4 = x4".
          inversion H0 as [ H_arg' H_right_val ].
          inversion H_right_val as [ v2 H_right_val' ].
          rewrite H_arg' in H_subst_red.
          rewrite H_right_val' in H_subst_red.
          rewrite H_fun' in H_subst_red.
          unfold substO in H_subst_red.
          simpl in H_subst_red.
          assert (subst !v2 (O, [!v2]) = !v2).
            rewrite H_right_val' in H_right_closed.
            unfold closed_term in H_right_closed.
            apply (H_right_closed 0 !v2).
          assert (!(subst_val v3 (0, [!v2])) @ !v2 =
                  substO (!v3 @ !v2) [e2]).
            unfold substO.
            simpl.
            rewrite H_right_val'.
            f_equal.
            simpl in H1.
            rewrite H1.
            reflexivity.
          rewrite H2 in H_subst_red.
          pose proof (lemm_coterm_app_aux e2 e1 v3 v2 e2' H_subst_red).
          inversion H3 as [ e' H4 ].
          inversion H4.
          exists e'.
          intro E'.
          split.
            rewrite H_arg'.
            unfold substO.
            simpl.
            apply rt_trans with (E'@[!(subst_val v3 (0, [e1])) @ e2]).
            SSSCase "rt_trans1".
              apply lemm_eval_ectx.
                apply (lemm_eval_ectx
                         e1 e2
                         (ectx_app2 (subst_val v3 (O, [e1])) ectx_hole)).
                  apply rt_step. assumption.
                  rewrite H_right_val'.
                  discriminate.
                discriminate.
            SSSCase "rt_trans2".
            apply rt_step. apply eval_red.
            unfold substO in H6.
            simpl in H6.
            rewrite H_right_val'.
            assert (!v2 = subst !v2 (O, [e1])).
              unfold closed_term in H_right_closed.
              rewrite H_right_val' in H_right_closed.
              symmetry.
              apply (H_right_closed 0 e1).
            rewrite H7.
            assumption.

            rewrite H5.
            apply rt_refl.

        SCase "e3 = x3".
        inversion H as [ H_fun' H_right_val ].
        inversion H_right_val as [ v2 H_right_val' ].
        rewrite H_fun' in H_subst_red.
        rewrite H_right_val' in H_subst_red.
        unfold closed_term in H_right_closed.
        pose proof (H_right_closed 0 e2) as H_right_closed'.
        rewrite H_right_val' in H_right_closed'.
        simpl in H_right_closed'.
        rewrite H_right_val'.
        unfold substO in H_subst_red.
        simpl in H_subst_red.
        rewrite H_fun'.

        assert (forall e, !v2 @ e = !(subst_val v2 (0, [!v2])) @ e).
          intro. f_equal. auto.

        assert ((exists v4, e4 = !v4) \/
                (e4 = term_var 0 /\ exists v2', e2 = !v2')) as H_arg.
          apply lemm_right_red_ectx with (e1 := e1) (e2 := e2)
                                         (e' := e2')
                                         (E := ectx_app2 v2 ectx_hole);
            try assumption; try discriminate.
          unfold substO in H_subst_red.
          simpl in H_subst_red.
          unfold substO.
          simpl.
          rewrite H_right_val'.
          rewrite H_right_closed'.
          assumption.

        inversion H_arg.
          SSCase "e4 = v4".
          inversion H1 as [ v4 H_arg' ].
          rewrite H_arg'.
          rewrite H_arg' in H_subst_red.
          unfold substO in H_subst_red.
          simpl in H_subst_red.
          rewrite H0 in H_subst_red.
          assert (substO (! (v2) @ ! (v4)) [e2] ~~> e2') as H_subst_red'.
            unfold substO. simpl.
            rewrite H_right_val'.
            assumption.

          pose proof (lemm_coterm_app_aux e2 e1 v2 v4 e2' H_subst_red') as H_app.
          inversion H_app as [ e' H_app' ].
          inversion H_app' as [ H_app'1 H_app'2 ].

          exists e'.
          intro E'.
          split.
            SSSCase "cond 1".
            unfold substO.
            simpl.
            apply rt_trans with
              (y := E'@[!(subst_val v2 (0, [e1])) @ !(subst_val v4 (0, [e1]))]).
              SSSSCase "trans1".
              apply lemm_eval_ectx.
                assert (forall e1 e2, e1 @ e2 = (ectx_app1 ectx_hole e2) @[e1]).
                  reflexivity.
                rewrite H2. rewrite H2.
                apply lemm_eval_ectx.
                  apply rt_step.
                  rewrite H_right_val' in H_eval.
                  pose proof (H_right_closed 0 e1) as H_right_closed''.
                  rewrite H_right_val' in H_right_closed''.
                  simpl in H_right_closed''.
                  rewrite H_right_closed''.
                  assumption.

                  discriminate.

                discriminate.

              SSSSCase "trans2".
              apply rt_step.
              apply eval_red.
              rewrite H_app'1 in H_subst_red.
              assumption.

            SSSCase "cond 2".
            rewrite H_app'1.
            rewrite H_right_val'.
            apply rt_refl.

          SSCase "e4 = x4".
          inversion H1 as [ H_arg' H_right_val'' ].
          rewrite H_arg'.
          rewrite H_arg' in H_subst_red.
          simpl in H_subst_red.
          unfold substO.
          simpl.
          assert (closed_term e2') as H_e2'_closed.
            apply lemm_red_closed with (e1 := !v2 @ !v2); try assumption.
            unfold closed_term.
            intros n e'.
            pose proof (H_right_closed n e').
            rewrite H_right_val' in H2.
            simpl in H2.
            simpl.
            rewrite H2.
            reflexivity.
          exists e2'.
          intro E'.
          split.
            SSSCase "cond 1".
            apply rt_trans with (y := E'@[e2 @ e1]).
              SSSSCase "trans 1".
              assert (forall e1 e2, e1 @ e2 = (ectx_app1 ectx_hole e2) @[e1]).
                reflexivity.
              rewrite H2. rewrite H2.
              rewrite lemm_ectx_compose. rewrite lemm_ectx_compose.
              apply lemm_eval_ectx.
              apply rt_step. assumption.
              rewrite H_right_val'. discriminate.

              apply rt_trans with (y := E'@[e2 @ e2]).
              SSSSCase "trans 2".
              rewrite H_right_val'.
              assert (forall v1 e2, !v1 @ e2 = (ectx_app2 v1 ectx_hole) @[e2]).
                reflexivity.
              rewrite H2. rewrite H2.
              rewrite lemm_ectx_compose. rewrite lemm_ectx_compose.
              apply lemm_eval_ectx.
              apply rt_step.
              rewrite H_right_val' in H_eval.
              assumption.
              discriminate.
              SSSSCase "trans 3".
              apply rt_step. apply eval_red.
              rewrite H_right_val'.
              unfold closed_term in H_e2'_closed.
              pose proof (H_e2'_closed 0 e1) as H_e2'_closed'.
              rewrite H_e2'_closed'.
              assumption.

              SSSCase "cond 2".
              unfold closed_term in H_e2'_closed.
              pose proof (H_e2'_closed 0 !v2) as H_e2'_closed'.
              rewrite H_e2'_closed'.
              apply rt_refl.


      Case "e = tapp".
      assert ((exists v, e = !v) \/ (e = term_var 0 /\ exists v2, e2 = !v2))
        as H_tabs.
        apply lemm_right_red_ectx with (e1 := e1) (e2 := e2)
                                       (e' := e2')
                                       (E := ectx_tapp ectx_hole t);
          try assumption; try discriminate.
      inversion H_tabs.
        SCase "e = v".
        inversion H as [ v H_tabs' ].
        rewrite H_tabs'.
        rewrite H_tabs' in H_subst_red.
        pose proof (lemm_coterm_tapp_aux e2 e1 v t e2' H_subst_red) as H_tapp.
        inversion H_tapp as [ e' H_tapp' ].
        inversion H_tapp' as [ H_tapp'1 H_tapp'2 ].
        exists e'.
        intro E'.
        split.
          SSCase "cond 1".
          apply rt_step.
          apply eval_red.
          assumption.
          SSCase "cond 2".
          rewrite H_tapp'1.
          apply rt_refl.

        SCase "e = x".
        inversion H as [ H_tabs' H_right_val ].
        inversion H_right_val as [ v2 H_right_val' ].
        rewrite H_tabs' in H_subst_red.
        unfold substO in H_subst_red.
        simpl in H_subst_red.
        assert (forall T, e2 @@ T = !v2 @@ T).
          rewrite H_right_val'. reflexivity.
        assert (forall T, !v2 @@ T = subst !v2 (0, [e2]) @@ T).
          intros; f_equal.
          unfold closed_term in H_right_closed.
          pose proof (H_right_closed 0 e2).
          rewrite H_right_val' in H1.
          rewrite H_right_val'.
          auto.
        assert (subst !v2 (0, [e2]) @@ subst_ty t (0, [e2]) =
                substO (!v2 @@ t) [e2]).
          reflexivity.
        rewrite H0 in H_subst_red.
        rewrite H1 in H_subst_red.
        rewrite H2 in H_subst_red.
        pose proof (lemm_coterm_tapp_aux e2 e1 v2 t e2' H_subst_red) as H_tapp.
        inversion H_tapp as [ e' H_tapp' ].
        inversion H_tapp' as [ H_tapp'1 H_tapp'2 ].
        exists e'.
        intro E'.
        split.
          SSCase "cond 1".
          rewrite H_tabs'.
          unfold substO.
          simpl.
          apply rt_trans with (y := E'@[e2 @@ subst_ty t (0, [e1])]).
          SSSCase "trans 1".
          assert (forall e T, e @@ T = (ectx_tapp ectx_hole T) @[e]).
            reflexivity.
          rewrite H3. rewrite H3.
          rewrite lemm_ectx_compose. rewrite lemm_ectx_compose.
          apply lemm_eval_ectx.
          apply rt_step. assumption.
          rewrite H_right_val'. discriminate.
          SSSCase "trans 2".
          apply rt_step.
          apply eval_red.
          rewrite H_right_val'.
          unfold substO in H_tapp'2.
          simpl in H_tapp'2.
          assert ((subst !v2 (0, [e1])) = !v2) as H_right_closed'.
            unfold closed_term in H_right_closed.
            rewrite H_right_val' in H_right_closed.
            apply H_right_closed.
          rewrite <- H_right_closed'.
          assumption.

          SSCase "cond 2".
          rewrite H_tapp'1.
          apply rt_refl.


      Case "e = wcheck".
      assert ((exists v4, e4 = !v4) \/ (e4 = term_var 0 /\ exists v2, e2 = !v2))
        as H_checked.
        apply lemm_right_red_ectx with (e1 := e1) (e2 := e2)
                                       (e' := e2')
                                       (E := ectx_wcheck t e3 ectx_hole l);
          try assumption; try discriminate.
      inversion H_checked.
        SCase "e4 = v4".
        inversion H as [ v4 H_checked' ].
        rewrite H_checked' in H_subst_red.
        pose proof (lemm_coterm_wcheck_aux e2 e1 t e3 v4 l e2' H_subst_red)
          as H_wcheck.
        inversion H_wcheck as [ e' H_wcheck' ].
        inversion H_wcheck' as [ H_wcheck'1 H_wcheck'2 ].
        exists e'.
        intro E'.
        split.
          SSCase "cond 1".
          apply rt_step.
          apply eval_red.
          rewrite H_checked'.
          assumption.
          SSCase "cond 2".
          rewrite H_wcheck'1.
          apply rt_refl.

        SCase "e = x".
        inversion H as [ H_checked' H_right_val ].
        inversion H_right_val as [ v2 H_right_val' ].
        rewrite H_checked' in H_subst_red.
        unfold substO in H_subst_red.
        simpl in H_subst_red.
        assert (forall T e l, term_wcheck T e e2 l = term_wcheck T e !v2 l).
          rewrite H_right_val'. reflexivity.
        assert (forall T e l, term_wcheck T e !v2 l =
                              term_wcheck T e (subst !v2 (0, [e2])) l).
          intros; f_equal.
          unfold closed_term in H_right_closed.
          pose proof (H_right_closed 0 e2).
          rewrite H_right_val' in H1.
          rewrite H_right_val'.
          auto.
        assert (term_wcheck (subst_ty t (0, [e2]))
                            (subst e3 (shift_subst (0, [e2])))
                            (subst !v2 (0, [e2]))
                            l =
                substO (term_wcheck t e3 !v2 l) [e2]).
          reflexivity.
        rewrite H0 in H_subst_red.
        rewrite H1 in H_subst_red.
        rewrite H2 in H_subst_red.
        pose proof (lemm_coterm_wcheck_aux e2 e1 t e3 v2 l e2' H_subst_red)
          as H_wcheck.
        inversion H_wcheck as [ e' H_wcheck' ].
        inversion H_wcheck' as [ H_wcheck'1 H_wcheck'2 ].
        exists e'.
        intro E'.
        split.
          SSCase "cond 1".
          rewrite H_checked'.
          unfold substO.
          simpl.
          apply rt_trans with (y := E' @[term_wcheck (subst_ty t (0, [e1]))
                                                     (subst e3 (shift_subst
                                                                  (0, [e1])))
                                                     e2
                                                     l]).
          SSSCase "trans 1".
          assert (forall T e1 e2 l, term_wcheck T e1 e2 l =
                                    (ectx_wcheck T e1 ectx_hole l) @[e2]).
            reflexivity.
          rewrite H3. rewrite H3.
          rewrite lemm_ectx_compose. rewrite lemm_ectx_compose.
          apply lemm_eval_ectx.
          apply rt_step. assumption.
          rewrite H_right_val'. discriminate.
          SSSCase "trans 2".
          apply rt_step.
          apply eval_red.
          rewrite H_right_val'.
          unfold substO in H_wcheck'2.
          simpl in H_wcheck'2.
          assert ((subst !v2 (0, [e1])) = !v2) as H_right_closed'.
            unfold closed_term in H_right_closed.
            rewrite H_right_val' in H_right_closed.
            apply H_right_closed.
          rewrite <- H_right_closed'.
          assumption.

          SSCase "cond 2".
          rewrite H_wcheck'1.
          apply rt_refl.

      Case "e = acheck".
      assert ((exists v4, e4 = !v4) \/ (e4 = term_var 0 /\ exists v2, e2 = !v2))
        as H_cond.
        apply lemm_right_red_ectx with (e1 := e1) (e2 := e2)
                                       (e' := e2')
                                       (E := ectx_acheck t e3 ectx_hole v l);
          try assumption; try discriminate.
      inversion H_cond.
        SCase "e4 = v4".
        inversion H as [ v4 H_cond' ].
        rewrite H_cond' in H_subst_red.
        pose proof (lemm_coterm_acheck_aux e2 e1 t e3 v4 v l e2' H_subst_red)
          as H_acheck.
        inversion H_acheck as [ e' H_acheck' ].
        inversion H_acheck' as [ H_acheck'1 H_acheck'2 ].
        exists e'.
        intro E'.
        split.
          SSCase "cond 1".
          apply rt_step.
          apply eval_red.
          rewrite H_cond'.
          assumption.
          SSCase "cond 2".
          rewrite H_acheck'1.
          apply rt_refl.

        SCase "e = x".
        inversion H as [ H_cond' H_right_val ].
        inversion H_right_val as [ v2 H_right_val' ].
        rewrite H_cond' in H_subst_red.
        unfold substO in H_subst_red.
        simpl in H_subst_red.
        assert (forall T e l, term_acheck T e e2 l = term_acheck T e !v2 l).
          rewrite H_right_val'. reflexivity.
        assert (forall T e l, term_acheck T e !v2 l =
                              term_acheck T e (subst !v2 (0, [e2])) l).
          intros; f_equal.
          unfold closed_term in H_right_closed.
          pose proof (H_right_closed 0 e2).
          rewrite H_right_val' in H1.
          rewrite H_right_val'.
          auto.
        assert (term_acheck (subst_ty t (0, [e2]))
                            (subst e3 (shift_subst (0, [e2])))
                            (subst !v2 (0, [e2]))
                            (subst_val v (0, [e2]))
                            l =
                substO (term_acheck t e3 !v2 v l) [e2]).
          reflexivity.
        rewrite H0 in H_subst_red.
        rewrite H1 in H_subst_red.
        rewrite H2 in H_subst_red.
        pose proof (lemm_coterm_acheck_aux e2 e1 t e3 v2 v l e2' H_subst_red)
          as H_acheck.
        inversion H_acheck as [ e' H_acheck' ].
        inversion H_acheck' as [ H_acheck'1 H_acheck'2 ].
        exists e'.
        intro E'.
        split.
          SSCase "cond 1".
          rewrite H_cond'.
          unfold substO.
          simpl.
          apply rt_trans with (y := E' @[term_acheck (subst_ty t (0, [e1]))
                                                     (subst e3 (shift_subst
                                                                  (0, [e1])))
                                                     e2
                                                     (subst_val v (0, [e1]))
                                                     l]).
          SSSCase "trans 1".
          assert (forall T e1 e2 v l, term_acheck T e1 e2 v l =
                                    (ectx_acheck T e1 ectx_hole v l) @[e2]).
            reflexivity.
          rewrite H3. rewrite H3.
          rewrite lemm_ectx_compose. rewrite lemm_ectx_compose.
          apply lemm_eval_ectx.
          apply rt_step. assumption.
          rewrite H_right_val'. discriminate.
          SSSCase "trans 2".
          apply rt_step.
          apply eval_red.
          rewrite H_right_val'.
          unfold substO in H_acheck'2.
          simpl in H_acheck'2.
          assert ((subst !v2 (0, [e1])) = !v2) as H_right_closed'.
            unfold closed_term in H_right_closed.
            rewrite H_right_val' in H_right_closed.
            apply H_right_closed.
          rewrite <- H_right_closed'.
          assumption.

          SSCase "cond 2".
          rewrite H_acheck'1.
          apply rt_refl.

      Case "e = blame".
      unfold substO in H_subst_red.
      simpl in H_subst_red.
      inversion H_subst_red.
  Qed.

  Lemma lemm_eval_var_stuck :
    forall H E e x,
      (E@[term_var x] --> e) -> H.
  Proof.
    intros H E e x H_eval.
    exfalso.
    generalize dependent H_eval.
    generalize dependent e.
    induction E; intros.

    Case "E = []".
    inversion H_eval.
      SCase "red".
      destruct E; try discriminate.
      simpl in H0.
      rewrite H0 in H1.
      inversion H1.

      SCase "blame".
      destruct E; discriminate.

    Case "E = app1".
    inversion H_eval.
      SCase "red".
      destruct E0; try discriminate.
        SSCase "E = []".
        simpl in H0. rewrite H0 in H1.
        destruct E; simpl in H1; inversion H1.
        SSCase "E = app1".
        injection H0; intros.
        apply IHE with (e := E0@[e2]).
        rewrite <- H4.
        apply eval_red. assumption.
        SSCase "E = app2".
        injection H0; intros.
        destruct E; discriminate.
      SCase "blame".
      destruct E0; try discriminate.
        SSCase "E0 = app1".
        injection H0; intros.
        apply IHE with (e := term_blame l).
        rewrite <- H4.
        apply eval_blame.
        unfold not. intro.
        rewrite H5 in H4.
        destruct E; discriminate.
        SSCase "E0 = app2".
        injection H0; intros.
        destruct E; discriminate.

    Case "E = app2".
    inversion H_eval.
      SCase "red".
      destruct E0; try discriminate.
        SSCase "E = []".
        simpl in H0. rewrite H0 in H1.
        destruct E; simpl in H1; inversion H1.
        SSCase "E = app1".
        injection H0; intros.
        destruct E0; try discriminate.
        simpl in H4. rewrite H4 in H1. inversion H1.
        SSCase "E = app2".
        injection H0; intros.
        apply IHE with (e := E0@[e2]).
        rewrite <- H3.
        apply eval_red. assumption.

      SCase "blame".
      destruct E0; try discriminate.
        SSCase "E0 = app1".
        injection H0; intros.
        destruct E0; discriminate.
        SSCase "E0 = app2".
        injection H0; intros.
        apply IHE with (e := term_blame l).
        rewrite <- H3.
        apply eval_blame.
        unfold not. intro.
        rewrite H5 in H3.
        destruct E; discriminate.

    Case "E = tapp".
    inversion H_eval.
      SCase "red".
      destruct E0; try discriminate.
        SSCase "E = []".
        simpl in H0. rewrite H0 in H1.
        destruct E; simpl in H1; inversion H1.
        SSCase "E = tapp".
        injection H0; intros.
        apply IHE with (e := E0@[e2]).
        rewrite <- H4.
        apply eval_red. assumption.

      SCase "blame".
      destruct E0; try discriminate.
        SSCase "E0 = tapp".
        injection H0; intros.
        apply IHE with (e := term_blame l).
        rewrite <- H4.
        apply eval_blame.
        unfold not. intro.
        rewrite H5 in H4.
        destruct E; discriminate.

    Case "E = wcheck".
    inversion H_eval.
      SCase "red".
      destruct E0; try discriminate.
        SSCase "E = []".
        simpl in H0. rewrite H0 in H1.
        destruct E; simpl in H1; inversion H1.
        SSCase "E = wcheck".
        injection H0; intros.
        apply IHE with (e := E0@[e2]).
        rewrite <- H4.
        apply eval_red. assumption.

      SCase "blame".
      destruct E0; try discriminate.
        SSCase "E0 = wcheck".
        injection H0; intros.
        apply IHE with (e := term_blame l0).
        rewrite <- H4.
        apply eval_blame.
        unfold not. intro.
        rewrite H7 in H4.
        destruct E; discriminate.

    Case "E = acheck".
    inversion H_eval.
      SCase "red".
      destruct E0; try discriminate.
        SSCase "E = []".
        simpl in H0. rewrite H0 in H1.
        destruct E; simpl in H1; inversion H1.
        SSCase "E = acheck".
        injection H0; intros.
        apply IHE with (e := E0@[e2]).
        rewrite <- H5.
        apply eval_red. assumption.

      SCase "blame".
      destruct E0; try discriminate.
        SSCase "E0 = acheck".
        injection H0; intros.
        apply IHE with (e := term_blame l0).
        rewrite <- H5.
        apply eval_blame.
        unfold not. intro.
        rewrite H8 in H5.
        destruct E; discriminate.
  Qed.

  Lemma lemm_coterm_right_subterm_red :
    forall e1 e2 E e e1' e2',
      (closed_term e1) ->
      (closed_term e2) ->
      (closed_term (substO e [e2])) ->
      (e1 --> e2) ->
      (substO e [e2] --> E@[e2']) ->
      (substO e [e2] = E@[e1']) ->
      (e1' ~~> e2') ->
      exists e', forall E',
        (E'@[substO e [e1]] -->* E'@[substO e' [e1]]) /\
        (E'@[E@[e2']] -->* E'@[substO e' [e2]]).
  Proof.
    intros e1 e2 E e e1' e2' H_left_closed H_right_closed.
    generalize dependent e2'.
    generalize dependent e1'.
    generalize dependent e.
    induction E;
      intros e e1' e2'
             H_closed H_eval H_subst_eval H_orig H_red.
      Case "E = []".
      simpl in H_orig.
      rewrite <- H_orig in H_red.
      simpl.
      destruct e.

        SCase "e = x".
        unfold substO.
        exists e2'.
        intro E'.
        unfold substO in H_subst_eval.
        simpl in H_subst_eval.
        remember (v - 0) as v_0.
        destruct v_0.
          SSCase "x = 0".
          pose proof (lemm_eval_closed e2 e2'
                                       H_right_closed
                                       H_subst_eval) as H_e2'_closed.
          unfold closed_term in H_e2'_closed.
          assert (0 = v) as H_v0. omega.
          rewrite <- H_v0 in H_orig. simpl in H_orig.
          rewrite <- H_v0 in H_red. simpl in H_red.
          rewrite <- H_v0. simpl.
          inversion H_eval.
            SSSCase "e1 red e2".
            split.
              SSSSCase "red".
              apply rt_trans with (y := E' @[e2]).
                rewrite <- H1.
                rewrite lemm_ectx_compose. rewrite lemm_ectx_compose.
                apply rt_step.
                apply eval_red.
                assumption.

                unfold substO in H_red. simpl in H_red.
                pose proof (H_e2'_closed 0 (E@[e0])) as H_e2'_closed'.
                rewrite H_e2'_closed.
                apply rt_step.
                apply eval_red.
                assumption.

              SSSSCase "-->*".
              pose proof (H_e2'_closed 0 (E@[e3])) as H_e2'_closed'.
              rewrite H_e2'_closed'.
              apply rt_refl.

            SSSCase "e1 blame e2".
            unfold substO in H_red.
            simpl in H_red.
            rewrite <- H1 in H_red.
            inversion H_red.

          SSCase "x = S _".
          destruct v_0;
            try (inversion H_subst_eval;
                 destruct E; try discriminate;
                 simpl in H;
                 rewrite H in H0;
                 inversion H0);
            destruct E; try discriminate.

        unfold substO in H_red.
        simpl in H_red.
        inversion H_red.

        SCase "e = app".
        unfold substO.
        pose proof (lemm_coterm_right_red e1 e2 (e3@e4) e2'
                                          H_left_closed H_right_closed H_closed
                                          H_eval H_red) as H_right_red.
        inversion H_right_red as [ e' H_right_red' ].
        exists e'.
        intro E'.
        apply (H_right_red' E').

        SCase "e = tapp".
        unfold substO.
        pose proof (lemm_coterm_right_red e1 e2 (e@@t) e2'
                                          H_left_closed H_right_closed H_closed
                                          H_eval H_red) as H_right_red.
        inversion H_right_red as [ e' H_right_red' ].
        exists e'.
        intro E'.
        apply (H_right_red' E').

        SCase "e = wcheck".
        unfold substO.
        pose proof (lemm_coterm_right_red e1 e2 (term_wcheck t e3 e4 l) e2'
                                          H_left_closed H_right_closed H_closed
                                          H_eval H_red) as H_right_red.
        inversion H_right_red as [ e' H_right_red' ].
        exists e'.
        intro E'.
        apply (H_right_red' E').

        SCase "e = acheck".
        unfold substO.
        pose proof (lemm_coterm_right_red e1 e2 (term_acheck t e3 e4 v l) e2'
                                          H_left_closed H_right_closed H_closed
                                          H_eval H_red) as H_right_red.
        inversion H_right_red as [ e' H_right_red' ].
        exists e'.
        intro E'.
        apply (H_right_red' E').

        SCase "e = blame".
        unfold substO in H_red.
        inversion H_red.

      Case "E = app1".
      unfold substO in H_orig.
      simpl in H_orig.
      destruct e; try discriminate.
        SCase "e = x".
        unfold substO in H_subst_eval.
        simpl in H_subst_eval.
        remember (v - 0) as v_0.
        destruct v_0.
          SSCase "x = 0".
          assert (0 = v) as H_v0. omega.
          rewrite <- H_v0.
          unfold substO. simpl.
          inversion H_subst_eval.
          exists (E@[e2'] @ t).
          intro E'.
          assert (closed_term (E@[e2']@t)) as H_closed_Ee2'.
            apply lemm_eval_closed with (e1 := e2); assumption.
            unfold closed_term in H_closed_Ee2'.
          split.
            SSSCase "cond 1".
            apply lemm_eval_ectx.
            rewrite H_closed_Ee2'.
            apply rt_trans with (y := e2); apply rt_step; assumption.
            discriminate.

            SSSCase "cond 2".
            rewrite H_closed_Ee2'.
            rewrite H.
            apply rt_refl.

          SSCase "x = S _".
          destruct v_0.
            SSSCase "x = 0".
            inversion H_subst_eval.
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := E@[e2']@t)
                                           (x := v - 1).
            assumption.
            SSSCase "x = S _".
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := E@[e2']@t)
                                           (x := v - 1).
            assumption.

        SCase "e = app".
        inversion H_subst_eval.
        assert (ectx_app1 E t = E0 /\ e1' = e0).
          apply lemm_red_ectx_decomp with (e := substO (e3@e4) [e2])
                                          (e2 := e2')
                                          (e2' := e5); auto.
        inversion H2.
        destruct E0; try discriminate.
        injection H3; intros.
        injection H_orig; intros.
        assert (exists e' : term,
                   forall E' : ectx,
                     (E'@[ substO e3 [e1]] -->* E'@[ substO e' [e1]]) /\
                     (E'@[ E @[ e2']] -->* E'@[ substO e' [e2]])).
          apply IHE with (e1' := e1'); auto.

          unfold closed_term.
          intros.
          unfold closed_term in H_closed.
          pose proof (H_closed n e').
          injection H9; intros.
          auto.

          unfold substO. rewrite H8. apply eval_red. assumption.

        inversion H9 as [ e' H9' ].
        exists (e' @ e4).
        intro E'.
        assert (forall e1 e2, e1@e2 = (ectx_app1 ectx_hole e2) @[e1]).
          reflexivity.
        split.
          SSCase "cond 1".
          unfold substO. simpl.
          repeat (rewrite H10; rewrite lemm_ectx_compose).
          apply (H9' (E' ^^ ectx_app1 ectx_hole (subst e4 (0, [e1])))).

          SSCase "cond 2".
          unfold substO. simpl.
          injection H; intros.
          rewrite <- H11.
          rewrite H5.
          rewrite H10. rewrite lemm_ectx_compose.
          rewrite H10.
          assert (forall e, E'@[ectx_app1 ectx_hole t0 @[e]] =
                            E'^^(ectx_app1 ectx_hole t0) @[e]).
            apply lemm_ectx_compose.
          rewrite H13.
          apply (H9' (E'^^(ectx_app1 ectx_hole t0))).


      Case "E = app2".
      unfold substO in H_orig.
      simpl in H_orig.
      destruct e; try discriminate.
        SCase "e = x".
        unfold substO in H_subst_eval.
        simpl in H_subst_eval.
        remember (v0 - 0) as v_0.
        destruct v_0.
          SSCase "x = 0".
          assert (0 = v0) as H_v0. omega.
          rewrite <- H_v0.
          unfold substO. simpl.
          inversion H_subst_eval.
          exists (!v @ (E@[e2'])).
          intro E'.
          assert (closed_term (!v @ (E@[e2']))) as H_closed_Ee2'.
            apply lemm_eval_closed with (e1 := e2); assumption.
            unfold closed_term in H_closed_Ee2'.
          split.
            SSSCase "cond 1".
            apply lemm_eval_ectx.
            rewrite H_closed_Ee2'.
            apply rt_trans with (y := e2); apply rt_step; assumption.
            discriminate.

            SSSCase "cond 2".
            rewrite H_closed_Ee2'.
            rewrite H.
            apply rt_refl.

          SSCase "x = S _".
          destruct v_0.
            SSSCase "x = 0".
            inversion H_subst_eval.
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := !v @ (E@[e2']))
                                           (x := v0 - 1).
            assumption.
            SSSCase "x = S _".
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := !v @ (E@[e2']))
                                           (x := v0 - 1).
            assumption.

        SCase "e = app".
        inversion H_subst_eval.
        assert (ectx_app2 v E = E0 /\ e1' = e0).
          apply lemm_red_ectx_decomp with (e := substO (e3@e4) [e2])
                                          (e2 := e2')
                                          (e2' := e5); auto.
        inversion H2.
        destruct E0; try discriminate.
        injection H3; intros.
        injection H_orig; intros.
        assert (exists e' : term,
                   forall E' : ectx,
                     (E'@[ substO e4 [e1]] -->* E'@[ substO e' [e1]]) /\
                     (E'@[ E @[ e2']] -->* E'@[ substO e' [e2]])).
          apply IHE with (e1' := e1'); auto.

          unfold closed_term.
          intros.
          unfold closed_term in H_closed.
          pose proof (H_closed n e').
          injection H9; intros.
          auto.

          unfold substO. rewrite H7. apply eval_red. assumption.

        inversion H9 as [ e' H9' ].
        assert (exists v3, (substO e3 [e1] -->* !(subst_valO v3 [e1])) /\
                           (v = subst_valO v3 [e2])) as H_v3.
          apply lemm_coterm_right_val; auto.
        inversion H_v3 as [ v3 H_v3' ].
        inversion H_v3'.
        exists (!v3 @ e').
        intro E'.
        assert (forall e1 e2, e1 @ e2 = (ectx_app1 ectx_hole e2) @[e1]).
          reflexivity.
        assert (forall v1 e2, !v1 @ e2 = (ectx_app2 v1 ectx_hole) @[e2]).
          reflexivity.
        split.
          SSCase "cond 1".
          apply rt_trans with (y := E'@[ substO (!v3 @ e4) [e1]]).
            SSSCase "trans 1".
            unfold substO. simpl.
            repeat (rewrite H12; rewrite lemm_ectx_compose).
            apply lemm_eval_ectx; auto.
            discriminate.

            SSSCase "trans 2".
            unfold substO. simpl.
            repeat (rewrite H13; rewrite lemm_ectx_compose).
            apply H9'.

            SSCase "cond 2".
            rewrite H11.
            unfold substO. unfold subst_valO. simpl.
            assert (forall e, E'@[!(subst_val v3 (0, [e2])) @ e] =
                              (E' ^^ (ectx_app2 (subst_val v3 (0, [e2])) ectx_hole) @[e])).
              intros.
              rewrite H13.
              rewrite lemm_ectx_compose.
              reflexivity.
            rewrite H14.
            rewrite H14.
            apply H9'.

      Case "E = tapp".
      unfold substO in H_orig.
      simpl in H_orig.
      destruct e; try discriminate.
        SCase "e = x".
        unfold substO in H_subst_eval.
        simpl in H_subst_eval.
        remember (v - 0) as v_0.
        destruct v_0.
          SSCase "x = 0".
          assert (0 = v) as H_v0. omega.
          rewrite <- H_v0.
          unfold substO. simpl.
          inversion H_subst_eval.
          exists (E@[e2'] @@ t).
          intro E'.
          assert (closed_term (E@[e2'] @@ t)) as H_closed_Ee2'.
            apply lemm_eval_closed with (e1 := e2); assumption.
            unfold closed_term in H_closed_Ee2'.
          split.
            SSSCase "cond 1".
            apply lemm_eval_ectx.
            rewrite H_closed_Ee2'.
            apply rt_trans with (y := e2); apply rt_step; assumption.
            discriminate.

            SSSCase "cond 2".
            rewrite H_closed_Ee2'.
            rewrite H.
            apply rt_refl.

          SSCase "x = S _".
          destruct v_0.
            SSSCase "x = 0".
            inversion H_subst_eval.
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := E@[e2'] @@ t)
                                           (x := v - 1).
            assumption.
            SSSCase "x = S _".
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := E@[e2'] @@ t)
                                           (x := v - 1).
            assumption.

        SCase "e = tapp".
        inversion H_subst_eval.
        assert (ectx_tapp E t = E0 /\ e1' = e0).
          apply lemm_red_ectx_decomp with (e := substO (e @@ t0) [e2])
                                          (e2 := e2')
                                          (e2' := e3); auto.
        inversion H2.
        destruct E0; try discriminate.
        injection H3; intros.
        injection H_orig; intros.
        assert (exists e' : term,
                   forall E' : ectx,
                     (E'@[ substO e [e1]] -->* E'@[ substO e' [e1]]) /\
                     (E'@[ E @[ e2']] -->* E'@[ substO e' [e2]])).
          apply IHE with (e1' := e1'); auto.

          unfold closed_term.
          intros.
          unfold closed_term in H_closed.
          pose proof (H_closed n e').
          injection H9; intros.
          auto.

          unfold substO. rewrite H8. apply eval_red. assumption.

        inversion H9 as [ e' H9' ].
        exists (e' @@ t0).
        intro E'.
        assert (forall e T, e @@ T = (ectx_tapp ectx_hole T) @[e]).
          reflexivity.
        split.
          SSCase "cond 1".
          unfold substO. simpl.
          repeat(rewrite H10; rewrite lemm_ectx_compose).
          apply H9'.

          SSCase "cond 2".
          rewrite H10.
          unfold substO. unfold subst_valO. simpl.
          repeat (rewrite H10).
          assert (forall E T e, E@[ (ectx_tapp ectx_hole T) @[e] ] =
                                (E ^^ (ectx_tapp ectx_hole T)) @[e]).
            intros.
            rewrite lemm_ectx_compose.
            reflexivity.
          repeat (rewrite H11).
          rewrite H7.
          apply H9'.

      Case "E = wcheck".
      unfold substO in H_orig.
      simpl in H_orig.
      destruct e; try discriminate.
        SCase "e = x".
        unfold substO in H_subst_eval.
        simpl in H_subst_eval.
        remember (v - 0) as v_0.
        destruct v_0.
          SSCase "x = 0".
          assert (0 = v) as H_v0. omega.
          rewrite <- H_v0.
          unfold substO. simpl.
          inversion H_subst_eval.
          exists (term_wcheck t t0 (E@[e2']) l).
          intro E'.
          assert (closed_term (term_wcheck t t0 (E@[e2']) l)) as H_closed_Ee2'.
            apply lemm_eval_closed with (e1 := e2); assumption.
            unfold closed_term in H_closed_Ee2'.
          split.
            SSSCase "cond 1".
            apply lemm_eval_ectx.
            rewrite H_closed_Ee2'.
            apply rt_trans with (y := e2); apply rt_step; assumption.
            discriminate.

            SSSCase "cond 2".
            rewrite H_closed_Ee2'.
            rewrite H.
            apply rt_refl.

          SSCase "x = S _".
          destruct v_0.
            SSSCase "x = 0".
            inversion H_subst_eval.
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := term_wcheck t t0 (E@[e2']) l)
                                           (x := v - 1).
            assumption.
            SSSCase "x = S _".
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := term_wcheck t t0 (E@[e2']) l)
                                           (x := v - 1).
            assumption.

        SCase "e = wcheck".
        inversion H_subst_eval.
        assert (ectx_wcheck t t0 E l = E0 /\ e1' = e0).
          apply lemm_red_ectx_decomp with
            (e := substO (term_wcheck t1 e3 e4 l0) [e2])
            (e2 := e2')
            (e2' := e5); auto.
        inversion H2.
        destruct E0; try discriminate.
        injection H; intros.
        injection H_orig; intros.
        assert (exists e' : term,
                   forall E' : ectx,
                     (E'@[ substO e4 [e1]] -->* E'@[ substO e' [e1]]) /\
                     (E'@[ E @[ e2']] -->* E'@[ substO e' [e2]])).
          apply IHE with (e1' := e1'); auto.

          unfold closed_term.
          intros.
          unfold closed_term in H_closed.
          pose proof (H_closed n e').
          injection H13; intros.
          auto.

          unfold substO. rewrite H10. apply eval_red. assumption.

        inversion H13 as [ e' H13' ].
        exists (term_wcheck t1 e3 e' l0).
        intro E'.
        assert (forall T e1 e2 l, term_wcheck T e1 e2 l =
                                  (ectx_wcheck T e1 ectx_hole l) @[e2]).
          reflexivity.
        split.
          SSCase "cond 1".
          unfold substO. simpl.
          repeat(rewrite H14; rewrite lemm_ectx_compose).
          apply H13'.

          SSCase "cond 2".
          rewrite <- H12.
          rewrite <- H11.
          rewrite H9.
          unfold substO. unfold subst_valO. simpl.
          repeat (rewrite H14).
          assert (forall E T e1 e2 l,
                     E@[ (ectx_wcheck T e1 ectx_hole l) @[e2] ] =
                     (E ^^ (ectx_wcheck T e1 ectx_hole l)) @[e2]).
            intros.
            rewrite lemm_ectx_compose.
            reflexivity.
          repeat (rewrite H15).
          apply H13'.

      Case "E = acheck".
      unfold substO in H_orig.
      simpl in H_orig.
      destruct e; try discriminate.
        SCase "e = x".
        unfold substO in H_subst_eval.
        simpl in H_subst_eval.
        remember (v0 - 0) as v_0.
        destruct v_0.
          SSCase "x = 0".
          assert (0 = v0) as H_v0. omega.
          rewrite <- H_v0.
          unfold substO. simpl.
          inversion H_subst_eval.
          exists (term_acheck t t0 (E@[e2']) v l).
          intro E'.
          assert (closed_term (term_acheck t t0 (E@[e2']) v l)) as H_closed_Ee2'.
            apply lemm_eval_closed with (e1 := e2); assumption.
            unfold closed_term in H_closed_Ee2'.
          split.
            SSSCase "cond 1".
            apply lemm_eval_ectx.
            rewrite H_closed_Ee2'.
            apply rt_trans with (y := e2); apply rt_step; assumption.
            discriminate.

            SSSCase "cond 2".
            rewrite H_closed_Ee2'.
            rewrite H.
            apply rt_refl.

          SSCase "x = S _".
          destruct v_0.
            SSSCase "x = 0".
            inversion H_subst_eval.
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := term_acheck t t0 (E@[e2']) v l)
                                           (x := v0 - 1).
            assumption.
            SSSCase "x = S _".
            apply lemm_eval_var_stuck with (E := ectx_hole)
                                           (e := term_acheck t t0 (E@[e2']) v l)
                                           (x := v0 - 1).
            assumption.

        SCase "e = acheck".
        inversion H_subst_eval.
        assert (ectx_acheck t t0 E v l = E0 /\ e1' = e0).
          apply lemm_red_ectx_decomp with
            (e := substO (term_acheck t1 e3 e4 v0 l0) [e2])
            (e2 := e2')
            (e2' := e5); auto.
        inversion H2.
        destruct E0; try discriminate.
        injection H; intros.
        injection H_orig; intros.
        assert (exists e' : term,
                   forall E' : ectx,
                     (E'@[ substO e4 [e1]] -->* E'@[ substO e' [e1]]) /\
                     (E'@[ E @[ e2']] -->* E'@[ substO e' [e2]])).
          apply IHE with (e1' := e1'); auto.

          unfold closed_term.
          intros.
          unfold closed_term in H_closed.
          pose proof (H_closed n e').
          injection H15; intros.
          auto.

          unfold substO. rewrite H12. apply eval_red. assumption.

        inversion H15 as [ e' H15' ].
        exists (term_acheck t1 e3 e' v0 l0).
        intro E'.
        assert (forall T e1 e2 v l, term_acheck T e1 e2 v l =
                                  (ectx_acheck T e1 ectx_hole v l) @[e2]).
          reflexivity.
        split.
          SSCase "cond 1".
          unfold substO. simpl.
          repeat(rewrite H16; rewrite lemm_ectx_compose).
          apply H15'.

          SSCase "cond 2".
          rewrite <- H14.
          rewrite <- H13.
          rewrite <- H11.
          rewrite H10.
          unfold substO. unfold subst_valO. simpl.
          repeat (rewrite H16).
          assert (forall E T e1 e2 v l,
                     E@[ (ectx_acheck T e1 ectx_hole v l) @[e2] ] =
                     (E ^^ (ectx_acheck T e1 ectx_hole v l)) @[e2]).
            intros.
            rewrite lemm_ectx_compose.
            reflexivity.
          repeat (rewrite H17).
          apply H15'.
  Qed.

  Lemma lemm_coterm_right_eval :
    forall e1 e2 e e2',
      (closed_term e1) ->
      (closed_term e2) ->
      (closed_term (substO e [e2])) ->
      (e1 --> e2) ->
      (substO e [e2] --> e2') ->
      exists e',
        (substO e [e1] -->* substO e' [e1]) /\
        (e2' -->* substO e' [e2]).
  Proof.
    intros e1 e2 e e2' H_left_closed H_right_closed H_closed H_eval H_subst_eval.
    inversion H_subst_eval.
    Case "red".
    rewrite <- H1 in H_subst_eval.
    symmetry in H.
    pose proof (lemm_coterm_right_subterm_red e1 e2 E e e0 e3
                                              H_left_closed H_right_closed
                                              H_closed
                                              H_eval
                                              H_subst_eval
                                              H H0) as H_exists.
    inversion H_exists as [ e' H_exists' ].
    exists e'.
    apply (H_exists' ectx_hole).

    Case "blame".
    exists (term_blame l).
    unfold substO. simpl.
    split.
      SCase "cond 1".
      symmetry in H.
      apply (lemm_coterm_right_blame e1 e2 e E l
                                     H_left_closed H_right_closed
                                     H_eval
                                     H).
      SCase "cond 2".
      apply rt_refl.
  Qed.

  Lemma nat_strong_ind :
    forall (P : nat -> Prop),
      (forall n, (forall k, k < n -> P k) -> P n) ->
      forall n, P n.
  Proof.
    intros.
    assert (forall k, k <= n -> P k).
      induction n; intros k H_kn.
      Case "n = O".
      apply H.
      intros k' H_kk'.
      omega.
      Case "n = S _".
      apply H.
      intros k' H_kk'.
      apply IHn.
      omega.

    apply H0.
    omega.
  Qed.

  Lemma lemm_coterm_aux' :
    forall n e1 e2 v,
      (e1 -->* e2) ->
      (clos_refl_trans_1n_length term eval e1 ! (v) n) ->
      exists n',
          n' <= n /\
          clos_refl_trans_1n_length term eval e2 ! (v) n'.
  Proof.
    intro n.
    induction n using nat_strong_ind; intros.
    pose proof (clos_rt_rt1n term eval e1 e2 H0) as H_eval1n.
    inversion H_eval1n.
      Case "refl".
      rewrite <- H2.
      exists n.
      auto.
      Case "step".
      inversion H1.
        SCase "refl".
        rewrite H5 in H2.
        exfalso.
        apply (lemm_eval_val_stuck v y H2).
        SCase "step".
        assert (y -->* e2).
          apply (clos_rt1n_rt term eval y e2 H3).
        assert (n0 < n). omega.
        assert (y = y0). apply lemm_eval_deterministic with (e := e1); auto.
        rewrite <- H11 in H6.
        pose proof (H n0 H10 y e2 v H9 H6).
        inversion H12.
        inversion H13.
        exists x.
        split; auto; omega.
  Qed.

  Lemma lemm_coterm_aux :
    forall n e1 e2 e3 v,
      clos_refl_trans_1n_length term eval e1 !v n ->
      (e1 --> e2) ->
      (e2 -->* e3) ->
      exists n',
        n' < n /\
        clos_refl_trans_1n_length term eval e3 !v n'.
  Proof.
    intros.
    inversion H.
    Case "n = 0".
    rewrite H2 in H0.
    exfalso.
    apply (lemm_eval_val_stuck v e2 H0).
    Case "n = S _".
    assert (y = e2).
      apply lemm_eval_deterministic with (e := e1); auto.
    rewrite H6 in H3.
    pose proof (lemm_coterm_aux' n0 e2 e3 v H1 H3).
    inversion H7 as [ n' H7' ].
    inversion H7'.
    exists n'.
    split; auto; omega.
  Qed.

  Lemma lemm_eval_multi_closed :
    forall e1 e2,
      (closed_term e1) ->
      (e1 -->* e2) ->
      closed_term e2.
  Proof.
    intros.
    induction H0.
    Case "step".
    apply lemm_eval_closed with (e1 := x); auto.
    Case "refl".
    auto.
    Case "step".
    apply IHclos_refl_trans2.
    apply IHclos_refl_trans1.
    auto.
  Qed.

  Lemma lemm_clos_rt1nlength_trans :
    forall A R e1 e2 m,
      clos_refl_trans_1n_length A R e1 e2 m ->
      forall e3 n,
        clos_refl_trans_1n_length A R e2 e3 n ->
        clos_refl_trans_1n_length A R e1 e3 (m + n).
  Proof.
    intros A R e1 e2 m H.
    induction H; intros.
    Case "refl".
    simpl.
    auto.
    Case "step".
    simpl.
    apply rt1n_length_trans with (y := y); auto.
  Qed.

  Lemma lemm_coterm :
    forall e1 e2 n e v1,
      (closed_term e1) ->
      (closed_term e2) ->
      (closed_term (substO e [e1])) ->
      ((e1 --> e2) \/ (e2 --> e1)) ->
      (clos_refl_trans_1n_length term eval (substO e [e1]) !v1 n) ->
      exists v,
        (exists n', clos_refl_trans_1n_length term eval
                                              (substO e [e2])
                                              (substO !v [e2])
                                              n') /\
        (v1 = subst_valO v [e1]).
  Proof.
    intros e1 e2.
    induction n using nat_strong_ind;
      intros e v1 H_left_closed H_right_closed H_closed H_eval H_multi_eval;
      inversion H_multi_eval.

    Case "rt_refl".
    destruct e; try discriminate.
      SCase "e = x".
      unfold substO in H1. simpl in H1.
      remember (v - 0) as v0.
      destruct v0.
        SSCase "x = 0".
        assert (0 = v) as H_v0. omega.
        rewrite <- H_v0.
        exists v1.
        unfold substO. simpl.
        unfold closed_term in H_left_closed.
        rewrite H1 in H_left_closed.
        split.
          SSSCase "cond 1".
          inversion H_eval.
            SSSSCase "e1 --> e2".
            rewrite H1 in H2.
            exfalso.
            apply (lemm_eval_val_stuck v1 e2 H2).
            SSSSCase "e2 --> e1".
            injection (H_left_closed 0 e2); intros.
            rewrite H3.
            rewrite H1 in H2.
            exists 1.
            apply rt1n_length_trans with (y := !v1); auto.
            apply rt1n_length_refl.

          SSSCase "cond 2".
          unfold subst_valO.
          injection (H_left_closed 0 e1); intros.
          auto.

        SSCase "x = S _".
        destruct v0; discriminate.

      SCase "e = v".
      exists v.
      split.
        SSCase "cond 1".
        exists 0. apply rt1n_length_refl.
        SSCase "cond 2".
        injection H1; intros.
        auto.

    Case "rt_trans".
    assert (exists e', (substO e [e2] -->* substO e' [e2]) /\
                       (y -->* substO e' [e1])).
      inversion H_eval.
        SCase "e1 --> e2".
        apply (lemm_coterm_left_eval e1 e2 e y
                                     H_left_closed
                                     H_right_closed
                                     H4
                                     H0).
        SCase "e2 --> e1".
        apply (lemm_coterm_right_eval e2 e1 e y
                                      H_right_closed
                                      H_left_closed
                                      H_closed
                                      H4
                                      H0).
      inversion H4 as [ e' H4' ].
      inversion H4'.
      pose proof (lemm_coterm_aux' n0 y (substO e' [e1]) v1 H6 H1).
      inversion H7 as [ n' H7' ].
      inversion H7'.
      assert (n' < n). omega.
      assert (closed_term y).
        apply lemm_eval_closed with (e1 := substO e [e1]); auto.
      assert (closed_term (substO e' [e1])).
        apply lemm_eval_multi_closed with (e1 := y); auto.
      pose proof (H n' H10 e' v1 H_left_closed H_right_closed H12 H_eval H9).
      inversion H13 as [ v H13' ].
      inversion H13' as [ H13'1 H13'2 ].
      inversion H13'1 as [ n'' H13'1' ].
      exists v.
      split.
        pose proof (clos_rt_rt1n term eval (substO e [e2]) (substO e' [e2]) H5).
        pose proof (clos_rt1n_length term eval
                                       (substO e [e2]) (substO e' [e2])
                                       H14).
        inversion H15.
        pose proof (lemm_clos_rt1nlength_trans
                      term eval (substO e [e2]) (substO e' [e2]) x H16
                      (substO !v [e2]) n'' H13'1').
        exists (x + n'').
        auto.

        auto.
  Qed.

  Lemma lemm_coterm_true :
    forall e1 e2 e,
      (closed_term e1) ->
      (closed_term e2) ->
      (closed_term (substO e [e1])) ->
      ((e1 --> e2) \/ (e2 --> e1)) ->
      (substO e [e1] -->* !(val_const (c_bool true))) ->
      (substO e [e2] -->* !(val_const (c_bool true))).
  Proof.
    intros e1 e2 e H_left_closed H_right_closed H_closed H_eval H_multi_eval.
    pose proof (clos_rt_rt1n term eval
                             (substO e [e1]) !(val_const (c_bool true))
                             H_multi_eval) as H_multi_eval_1n.
    pose proof (clos_rt1n_length term eval
                                 (substO e [e1]) !(val_const (c_bool true))
                                 H_multi_eval_1n) as H_multi_eval_length.
    inversion H_multi_eval_length as [ n H_multi_eval_length' ].
    pose proof (lemm_coterm e1 e2 n e (val_const (c_bool true))
                            H_left_closed H_right_closed H_closed H_eval
                            H_multi_eval_length') as H_coterm.
    inversion H_coterm as [ v H_coterm' ].
    inversion H_coterm' as [ H_coterm_eval H_coterm_val ].
    destruct v; try discriminate.
    injection H_coterm_val; intros H_const.
    rewrite <- H_const in H_coterm_eval.
    unfold substO in H_coterm_eval. simpl in H_coterm_eval.
    apply clos_rt1n_rt.
    inversion H_coterm_eval as [ n' H_coterm_eval' ].
    apply clos_length_rt1n with (n := n').
    auto.
  Qed.


End Coterm.
